;; Autogenerated file from EHMM state labels
(set! cmu_us_slt::phone_to_states '(
    ( aa aa_1 aa_2 aa_3  )
    ( ae ae_1 ae_2 ae_3  )
    ( ah ah_1 ah_2 ah_3  )
    ( ao ao_1 ao_2 ao_3  )
    ( aw aw_1 aw_2 aw_3  )
    ( ax ax_1 ax_2 ax_3  )
    ( ay ay_1 ay_2 ay_3  )
    ( b b_1 b_2 b_3  )
    ( ch ch_1 ch_2 ch_3  )
    ( d d_1 d_2 d_3  )
    ( dh dh_1 dh_2 dh_3  )
    ( eh eh_1 eh_2 eh_3  )
    ( er er_1 er_2 er_3  )
    ( ey ey_1 ey_2 ey_3  )
    ( f f_1 f_2 f_3  )
    ( g g_1 g_2 g_3  )
    ( hh hh_1 hh_2 hh_3  )
    ( ih ih_1 ih_2 ih_3  )
    ( iy iy_1 iy_2 iy_3  )
    ( jh jh_1 jh_2 jh_3  )
    ( k k_1 k_2 k_3  )
    ( l l_1 l_2 l_3  )
    ( m m_1 m_2 m_3  )
    ( n n_1 n_2 n_3  )
    ( ng ng_1 ng_2 ng_3  )
    ( ow ow_1 ow_2 ow_3  )
    ( oy oy_1 oy_2 oy_3  )
    ( p p_1 p_2 p_3  )
    ( pau pau_1 pau_2 pau_3  )
    ( r r_1 r_2 r_3  )
    ( s s_1 s_2 s_3  )
    ( sh sh_1 sh_2 sh_3  )
    ( ssil pau_5 )
    ( t t_1 t_2 t_3  )
    ( th th_1 th_2 th_3  )
    ( uh uh_1 uh_2 uh_3  )
    ( uw uw_1 uw_2 uw_3  )
    ( v v_1 v_2 v_3  )
    ( w w_1 w_2 w_3  )
    ( y y_1 y_2 y_3  )
    ( z z_1 z_2 z_3  )
    ( zh zh_1 zh_2 zh_3  )
))
(provide 'cmu_us_slt_statenames)
