;; Duration models autotrained by festvox
;; dur.S15.tree
(set! cmu_us_slt::phone_durs '
((pau 0.2 0.1)
 (oy_3 0.059012 0.040815)
 (oy_2 0.060988 0.035527)
 (oy_1 0.067901 0.042748)
 (uh_3 0.017545 0.027622)
 (uh_2 0.014576 0.018979)
 (uh_1 0.016576 0.021291)
 (ow_3 0.027323 0.026346)
 (ow_2 0.054379 0.044391)
 (ow_1 0.037709 0.036448)
 (zh_3 0.012222 0.021727)
 (zh_2 0.019028 0.018929)
 (zh_1 0.067639 0.038886)
 (v_3 0.026353 0.024296)
 (v_2 0.019882 0.014768)
 (v_1 0.01463 0.011822)
 (ay_3 0.053508 0.031052)
 (ay_2 0.052177 0.042235)
 (ay_1 0.034938 0.041109)
 (aw_3 0.051511 0.041822)
 (aw_2 0.050574 0.034838)
 (aw_1 0.058383 0.03858)
 (l_3 0.022871 0.027641)
 (l_2 0.044659 0.039082)
 (l_1 0.041087 0.042717)
 (g_3 0.026496 0.024494)
 (g_2 0.026911 0.019158)
 (g_1 0.016787 0.015853)
 (f_3 0.016071 0.012129)
 (f_2 0.064646 0.033903)
 (f_1 0.023921 0.021168)
 (ah_3 0.020331 0.019136)
 (ah_2 0.032948 0.032921)
 (ah_1 0.012666 0.013496)
 (z_3 0.042623 0.045634)
 (z_2 0.031889 0.021647)
 (z_1 0.02316 0.014951)
 (iy_3 0.021998 0.024223)
 (iy_2 0.056187 0.043912)
 (iy_1 0.024086 0.024388)
 (n_3 0.018083 0.022388)
 (n_2 0.036189 0.022035)
 (n_1 0.015271 0.018976)
 (s_3 0.019507 0.017653)
 (s_2 0.073062 0.036198)
 (s_1 0.022127 0.014674)
 (b_3 0.012139 0.008458)
 (b_2 0.02317 0.020256)
 (b_1 0.021735 0.019556)
 (ax_3 0.01263 0.015856)
 (ax_2 0.012441 0.014566)
 (ax_1 0.013972 0.017931)
 (ae_3 0.019836 0.027048)
 (ae_2 0.041702 0.043666)
 (ae_1 0.029703 0.033682)
 (hh_3 0.021679 0.023782)
 (hh_2 0.01786 0.023916)
 (hh_1 0.022058 0.028945)
 (d_3 0.016895 0.0179)
 (d_2 0.016964 0.016636)
 (d_1 0.012837 0.013812)
 (jh_3 0.051346 0.032162)
 (jh_2 0.021058 0.018385)
 (jh_1 0.032572 0.021808)
 (ch_3 0.033897 0.020887)
 (ch_2 0.053103 0.029445)
 (ch_1 0.039974 0.023378)
 (pau_5 0.029231 0.031578)
 (th_3 0.022288 0.016998)
 (th_2 0.048042 0.036804)
 (th_1 0.025472 0.018979)
 (r_3 0.019482 0.02163)
 (r_2 0.037302 0.03686)
 (r_1 0.025299 0.030355)
 (pau_2 0.026426 0.033818)
 (p_3 0.037075 0.026914)
 (p_2 0.038351 0.024055)
 (p_1 0.021806 0.014199)
 (aa_3 0.022357 0.021845)
 (aa_2 0.04858 0.051944)
 (aa_1 0.01502 0.015851)
 (sh_3 0.02215 0.012395)
 (sh_2 0.082117 0.026548)
 (sh_1 0.02755 0.014251)
 (ao_3 0.029907 0.047557)
 (ao_2 0.048119 0.057128)
 (ao_1 0.025708 0.029964)
 (t_3 0.013597 0.014554)
 (t_2 0.020422 0.020769)
 (t_1 0.022853 0.023236)
 (eh_3 0.028283 0.029115)
 (eh_2 0.032856 0.032646)
 (eh_1 0.016856 0.016591)
 (dh_3 0.01843 0.021958)
 (dh_2 0.010826 0.008511)
 (dh_1 0.021493 0.025807)
 (ng_3 0.025534 0.02559)
 (ng_2 0.048247 0.027264)
 (ng_1 0.032329 0.025612)
 (ih_3 0.018327 0.021515)
 (ih_2 0.02354 0.024523)
 (ih_1 0.0115 0.013632)
 (k_3 0.042208 0.024584)
 (k_2 0.033405 0.027742)
 (k_1 0.022006 0.013158)
 (ey_3 0.030913 0.035772)
 (ey_2 0.067314 0.045637)
 (ey_1 0.039282 0.045472)
 (m_3 0.021517 0.031293)
 (m_2 0.041899 0.027435)
 (m_1 0.014842 0.01834)
 (er_3 0.020806 0.027692)
 (er_2 0.03896 0.032525)
 (er_1 0.054182 0.050735)
 (w_3 0.019206 0.015882)
 (w_2 0.039164 0.035051)
 (w_1 0.018901 0.018738)
 (uw_3 0.034807 0.040831)
 (uw_2 0.028152 0.032449)
 (uw_1 0.029734 0.028615)
 (y_3 0.028885 0.026066)
 (y_2 0.032712 0.030707)
 (y_1 0.023154 0.027324)
 (pau_3 0.102875 0.059771)
 (pau_1 0.021917 0.028876))
)


(set! cmu_us_slt::zdur_tree '
((name is pau)
 ((p.R:SylStructure.parent.parent.pbreak is BB)
  ((0.0 2.0))
  ((0.0 0.0)))
((R:segstate.parent.R:SylStructure.parent.lisp_cg_break < 1.2)
 ((R:segstate.parent.nn.ph_vc is 0)
  ((n.name is pau_3)
   ((R:segstate.parent.p.ph_cvox is -)
    ((R:segstate.parent.pp.ph_vfront is 2)
     ((0.252211 -0.369703))
     ((R:segstate.parent.pp.ph_vlng is s)
      ((0.32399 -0.294242))
      ((0.63649 -0.024936))))
    ((R:segstate.parent.p.name is n)
     ((1.16161 2.76339))
     ((R:segstate.parent.p.name is l)
      ((R:segstate.parent.pp.ph_vheight is 2)
       ((0.797104 0.389154))
       ((0.60056 0.137322)))
      ((R:segstate.parent.p.ph_vrnd is +)
       ((1.4512 2.80206))
       ((R:segstate.parent.p.name is er)
        ((0.702154 0.638244))
        ((R:segstate.parent.p.name is z)
         ((0.66678 0.383812))
         ((R:segstate.parent.p.name is m)
          ((R:segstate.parent.pp.ph_vlng is s)
           ((1.0311 1.19418))
           ((1.26252 1.85755)))
          ((R:segstate.parent.pp.ph_cplace is a)
           ((R:segstate.parent.p.name is iy)
            ((0.696677 1.47688))
            ((1.1471 1.57031)))
           ((R:segstate.parent.p.name is r)
            ((0.970261 1.60156))
            ((R:segstate.parent.p.name is d)
             ((1.0928 1.67815))
             ((R:segstate.parent.p.ph_vlng is 0)
              ((1.10843 2.42766))
              ((0.790414 2.03151)))))))))))))
   ((R:segstate.parent.p.ph_ctype is f)
    ((n.name is 0)
     ((R:segstate.parent.p.name is s)
      ((R:segstate.parent.pp.ph_vlng is d)
       ((0.358691 -0.603998))
       ((R:segstate.parent.pp.ph_vc is +)
        ((0.136883 -0.718054))
        ((0.128817 -0.76528))))
      ((R:segstate.parent.p.name is z)
       ((R:segstate.parent.pp.ph_vc is +)
        ((R:segstate.parent.pp.ph_vfront is 2)
         ((0.532454 -0.718922))
         ((0.441607 -0.673534)))
        ((0.528456 -0.60192)))
       ((0.84812 -0.0651363))))
     ((R:segstate.parent.p.ph_cvox is +)
      ((R:segstate.parent.pp.ph_vheight is 2)
       ((0.347015 -0.345359))
       ((R:segstate.parent.pp.ph_vc is +)
        ((0.66054 0.187574))
        ((0.595957 -0.12591))))
      ((R:segstate.parent.p.name is s)
       ((R:segstate.parent.pp.ph_vlng is d)
        ((0.380489 -0.308804))
        ((R:segstate.parent.pp.ph_cplace is a)
         ((0.326071 -0.432676))
         ((R:segstate.parent.pp.ph_vheight is 2)
          ((0.199629 -0.488451))
          ((0.0894184 -0.562766)))))
       ((0.366718 -0.321561)))))
    ((R:segstate.parent.p.ph_cvox is -)
     ((n.name is 0)
      ((R:segstate.parent.p.name is t)
       ((R:segstate.parent.pp.ph_vc is +)
        ((R:segstate.parent.pp.ph_vlng is s)
         ((0.625271 -0.260579))
         ((R:segstate.parent.pp.ph_vheight is 3)
          ((0.784127 0.0839674))
          ((0.780856 0.362057))))
        ((0.701125 -0.475189)))
       ((R:segstate.parent.pp.ph_vfront is 1)
        ((1.16359 0.0961244))
        ((1.23419 0.695704))))
      ((R:segstate.parent.pp.ph_vfront is 1)
       ((R:segstate.parent.p.name is k)
        ((0.0774362 -0.54256))
        ((0.277866 -0.393457)))
       ((R:segstate.parent.pp.ph_cplace is a)
        ((1.17855 0.0144185))
        ((R:segstate.parent.pp.ph_vfront is 2)
         ((0.421792 -0.268652))
         ((0.42355 -0.217898))))))
     ((R:segstate.parent.p.ph_ctype is 0)
      ((n.name is pau_2)
       ((R:segstate.parent.p.ph_vrnd is +)
        ((1.60292 2.68099))
        ((R:segstate.parent.p.ph_vheight is 2)
         ((0.739418 0.127138))
         ((0.925631 1.22145))))
       ((R:segstate.parent.p.ph_vrnd is +)
        ((R:segstate.parent.p.name is uw)
         ((0.715912 -0.287683))
         ((0.875578 -0.0698038)))
        ((R:segstate.parent.p.ph_vheight is 2)
         ((0.99124 1.17455))
         ((R:segstate.parent.pp.ph_cplace is a)
          ((0.722828 0.335016))
          ((0.469882 0.185922))))))
      ((n.name is 0)
       ((R:segstate.parent.p.ph_ctype is s)
        ((R:segstate.parent.pp.ph_vfront is 1)
         ((0.812889 0.787376))
         ((R:segstate.parent.pp.ph_cplace is a)
          ((1.33376 0.979483))
          ((R:segstate.parent.pp.ph_vfront is 2)
           ((1.17374 1.14242))
           ((1.08136 1.43544)))))
        ((R:segstate.parent.p.name is m)
         ((R:segstate.parent.pp.ph_vlng is s)
          ((0.575255 -0.394878))
          ((1.09625 -0.121386)))
         ((R:segstate.parent.p.name is l)
          ((R:segstate.parent.pp.ph_vheight is 2)
           ((0.595048 -0.396054))
           ((0.732241 -0.140118)))
          ((R:segstate.parent.pp.ph_vlng is a)
           ((0.713299 0.1729))
           ((R:segstate.parent.p.name is er)
            ((0.526249 0.370771))
            ((R:segstate.parent.p.name is ng)
             ((1.4692 0.461882))
             ((R:segstate.parent.pp.ph_vlng is l)
              ((1.07401 0.802556))
              ((R:segstate.parent.pp.ph_vheight is 2)
               ((0.909816 1.10846))
               ((1.39686 1.18908))))))))))
       ((R:segstate.parent.p.ph_ctype is n)
        ((R:segstate.parent.pp.ph_vlng is s)
         ((R:segstate.parent.p.name is n)
          ((1.6267 0.55493))
          ((1.02471 1.39755)))
         ((R:segstate.parent.pp.ph_vlng is l)
          ((1.56785 0.433836))
          ((0.563778 -0.117693))))
        ((R:segstate.parent.pp.ph_vlng is a)
         ((1.25829 0.114638))
         ((R:segstate.parent.pp.ph_vfront is 1)
          ((R:segstate.parent.p.name is r)
           ((0.209981 -0.463624))
           ((0.890163 -0.0416511)))
          ((R:segstate.parent.p.name is er)
           ((0.313291 -0.305748))
           ((R:segstate.parent.p.name is d)
            ((R:segstate.parent.pp.ph_vc is +)
             ((0.204877 -0.455983))
             ((0.285114 -0.449149)))
            ((0.140939 -0.510095))))))))))))
  ((R:segstate.parent.p.name is ax)
   ((n.name is r_2)
    ((1.12287 2.2056))
    ((n.name is w_3)
     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
       <
       0.1)
      ((1.34208 0.669969))
      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
        <
        0.3)
       ((0.769072 1.84025))
       ((1.04272 1.13652))))
     ((n.name is d_2)
      ((R:segstate.parent.syl_final is 0)
       ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 4)
        ((0.320395 -0.341153))
        ((0.705784 0.023233)))
       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
         <
         1.2)
        ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.2)
         ((2.42244 3.11652))
         ((1.94835 1.4035)))
        ((1.12481 0.454725))))
      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
        <
        0.3)
       ((n.name is p_2)
        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
          <
          0.3)
         ((0.744551 1.17779))
         ((0.584339 0.518394)))
        ((n.name is z_2)
         ((1.78485 1.05946))
         ((n.name is pau_1)
          ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 2.4)
           ((1.73215 1.28187))
           ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 4.5)
            ((1.22274 0.483542))
            ((0.810798 -0.0759049))))
          ((R:segstate.parent.n.name is uw)
           ((R:segstate.parent.R:SylStructure.parent.asyl_out < 1.5)
            ((1.19528 0.951004))
            ((1.07586 0.42081)))
           ((R:segstate.parent.nn.ph_ctype is l)
            ((R:segstate.parent.n.name is t)
             ((0.382911 -0.78419))
             ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.2)
              ((R:segstate.parent.pos_in_syl is 0)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                 is
                 0)
                ((0.600958 -0.43926))
                ((0.938514 -0.206931)))
               ((1.12737 0.223071)))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                <
                0.2)
               ((0.411254 -0.693779))
               ((0.694599 -0.405233)))))
            ((n.name is z_3)
             ((0.93991 0.823064))
             ((n.name is n_3)
              ((R:segstate.parent.n.ph_cvox is +)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                 <
                 0.3)
                ((1.00095 0.295135))
                ((R:segstate.parent.n.ph_ctype is s)
                 ((0.77757 1.22742))
                 ((1.08195 0.740231))))
               ((R:segstate.parent.lisp_onset_stop is 0)
                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                  <
                  2.2)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                   <
                   0.2)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    3.2)
                   ((0.778836 -0.341381))
                   ((0.569943 -0.0673073)))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                    is
                    0)
                   ((0.688062 0.0285543))
                   ((0.950151 0.508386))))
                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                   <
                   6.1)
                  ((0.561502 -0.703752))
                  ((0.554935 -0.159853))))
                ((R:segstate.parent.nn.ph_ctype is 0)
                 ((1.34747 0.746203))
                 ((0.945227 0.384736)))))
              ((n.name is ay_1)
               ((2.30823 0.885536))
               ((n.name is k_2)
                ((0.683152 0.586428))
                ((R:segstate.parent.n.seg_onsetcoda is coda)
                 ((n.name is d_1)
                  ((0.474895 -0.436515))
                  ((n.name is s_3)
                   ((R:segstate.parent.n.name is t)
                    ((0.709464 -0.68315))
                    ((0.582387 -0.101165)))
                   ((R:segstate.parent.nn.ph_cplace is l)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                      <
                      0.3)
                     ((R:segstate.parent.syl_final is 0)
                      ((0.652473 -0.311178))
                      ((1.1328 0.261662)))
                     ((R:segstate.parent.n.ph_vc is +)
                      ((0.613469 1.33605))
                      ((0.966041 0.277611))))
                    ((R:segstate.parent.n.ph_vrnd is -)
                     ((R:segstate.parent.name is n)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                        <
                        0.3)
                       ((R:segstate.parent.nn.ph_ctype is f)
                        ((0.353794 -0.368204))
                        ((1.09818 0.109136)))
                       ((R:segstate.parent.R:SylStructure.parent.last_accent
                         <
                         1.4)
                        ((0.140157 -0.502528))
                        ((0.218594 -0.369329))))
                      ((n.name is l_2)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                         <
                         0.2)
                        ((0.908436 -0.103479))
                        ((0.339907 -0.586309)))
                       ((n.name is l_3)
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                          is
                          0)
                         ((0.910546 0.881702))
                         ((0.779422 -0.0168617)))
                        ((R:segstate.parent.pp.ph_cplace is p)
                         ((1.39129 0.468441))
                         ((n.name is g_3)
                          ((0.622839 0.457026))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((n.name is eh_1)
                            ((0.716724 -0.35996))
                            ((R:segstate.parent.R:SylStructure.parent.asyl_in
                              <
                              2.1)
                             ((n.name is ih_1)
                              ((0.536043 -0.207373))
                              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                <
                                1.2)
                               ((R:segstate.parent.R:SylStructure.parent.last_accent
                                 <
                                 1.6)
                                ((1.34359 0.613927))
                                ((R:segstate.parent.lisp_onset_stop
                                  is
                                  0)
                                 ((0.640528 -0.00125906))
                                 ((0.897629 0.469633))))
                               ((R:segstate.parent.R:SylStructure.parent.last_accent
                                 <
                                 1.4)
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  1.4)
                                 ((0.723798 -0.0286413))
                                 ((0.620348 -0.120063)))
                                ((0.809361 0.248287)))))
                             ((R:segstate.parent.R:SylStructure.parent.syl_in
                               <
                               9.5)
                              ((0.468023 -0.348594))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                <
                                7)
                               ((R:segstate.parent.ph_cplace is a)
                                ((0.658074 0.0527186))
                                ((1.20323 0.303619)))
                               ((0.598389 -0.251534))))))
                           ((R:segstate.parent.nn.ph_ctype is s)
                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                              <
                              1.4)
                             ((0.79172 0.314864))
                             ((0.92082 -0.0762342)))
                            ((R:segstate.parent.name is t)
                             ((0.321408 -0.55162))
                             ((R:segstate.parent.ph_cvox is +)
                              ((R:segstate.parent.R:SylStructure.parent.syl_out
                                <
                                9)
                               ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                 <
                                 1.3)
                                ((0.711154 -0.161919))
                                ((R:segstate.parent.R:SylStructure.parent.syl_in
                                  <
                                  11.5)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                   <
                                   7)
                                  ((0.508804 -0.310335))
                                  ((0.293043 -0.488067)))
                                 ((0.524785 -0.174266))))
                               ((0.590491 0.00580795)))
                              ((1.12645 0.124348)))))))))))
                     ((n.name is l_2)
                      ((R:segstate.parent.nn.ph_ctype is 0)
                       ((1.06851 1.08652))
                       ((1.20652 0.311071)))
                      ((R:segstate.parent.lisp_onset_nasal is 0)
                       ((R:segstate.parent.pos_in_syl is 0)
                        ((R:segstate.parent.ph_cplace is l)
                         ((0.569437 -0.244678))
                         ((R:segstate.parent.lisp_coda_glide is 0)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                            is
                            0)
                           ((1.01345 0.313262))
                           ((1.06046 0.288387)))
                          ((0.920509 0.734086))))
                        ((R:segstate.parent.lisp_coda_nasal is 0)
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           5.8)
                          ((R:segstate.parent.R:SylStructure.parent.syl_out
                            <
                            2.2)
                           ((0.66686 -0.315208))
                           ((R:segstate.parent.pp.ph_cplace is a)
                            ((0.848352 0.309102))
                            ((0.953439 -0.143061))))
                          ((R:segstate.parent.nn.ph_cplace is a)
                           ((0.670029 -0.142705))
                           ((0.45519 -0.531274))))
                         ((n.name is n_2)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                            <
                            0.2)
                           ((1.08472 -0.0017339))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             3.1)
                            ((0.522337 -0.139093))
                            ((0.176304 -0.468072))))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                            <
                            2.6)
                           ((0.512477 -0.00251282))
                           ((R:segstate.parent.nn.ph_ctype is 0)
                            ((1.02081 0.741914))
                            ((0.58253 0.116207)))))))
                       ((n.name is n_2)
                        ((1.40442 0.91672))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                          <
                          0.3)
                         ((1.27609 0.620419))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           4.2)
                          ((0.528697 0.128897))
                          ((0.558243 -0.183302)))))))))))
                 ((R:segstate.parent.name is d)
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    3.6)
                   ((0.931283 0.682093))
                   ((0.852991 0.260062)))
                  ((R:segstate.parent.name is l)
                   ((R:segstate.parent.pp.ph_cplace is v)
                    ((0.885068 -0.0989521))
                    ((n.name is l_2)
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                       is
                       0)
                      ((0.397838 -0.548268))
                      ((0.8029 0.0854423)))
                     ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                       <
                       2.3)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                        <
                        2.4)
                       ((0.573297 -0.314483))
                       ((0.393135 -0.586526)))
                      ((n.name is l_3)
                       ((0.214792 -0.826226))
                       ((0.14934 -0.565139))))))
                   ((n.name is r_1)
                    ((0.719854 0.535665))
                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                      <
                      3.6)
                     ((R:segstate.parent.n.name is l)
                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                        <
                        3.6)
                       ((1.88128 0.664326))
                       ((0.496521 -0.164567)))
                      ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                        <
                        1.2)
                       ((0.403703 -0.456079))
                       ((R:segstate.parent.name is p)
                        ((1.04624 -0.394679))
                        ((R:segstate.parent.nn.ph_vheight is 3)
                         ((R:segstate.parent.pos_in_syl is 2)
                          ((0.381673 -0.385671))
                          ((1.0374 -0.192904)))
                         ((R:segstate.parent.ph_cplace is a)
                          ((R:segstate.parent.nn.ph_vrnd is -)
                           ((R:segstate.parent.lisp_coda_nasal is 0)
                            ((0.710604 0.283766))
                            ((R:segstate.parent.n.name is t)
                             ((0.825295 0.185704))
                             ((R:segstate.parent.pp.ph_cplace is a)
                              ((0.837155 -0.0311006))
                              ((0.255374 -0.360695)))))
                           ((R:segstate.parent.lisp_onset_stop is 0)
                            ((0.500558 -0.305198))
                            ((0.362834 -0.160883))))
                          ((1.00636 0.360956)))))))
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       6.4)
                      ((0.380469 -0.54671))
                      ((0.34303 -0.302166)))))))))))))))))
       ((R:segstate.parent.name is t)
        ((n.name is t_2)
         ((0.768843 0.975784))
         ((R:segstate.parent.nn.ph_vc is +)
          ((1.75275 2.50407))
          ((0.90751 1.18069))))
        ((n.name is l_3)
         ((R:segstate.parent.R:SylStructure.parent.syl_codasize < 1.3)
          ((0.539474 2.21776))
          ((0.855104 1.59513)))
         ((R:segstate.parent.name is d)
          ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 2.1)
           ((0.821203 1.2549))
           ((0.800781 0.826595)))
          ((n.name is b_3)
           ((R:segstate.parent.n.ph_vrnd is -)
            ((0.830153 1.27883))
            ((0.77393 0.571034)))
           ((R:segstate.parent.name is l)
            ((n.name is l_2)
             ((R:segstate.parent.R:SylStructure.parent.position_type
               is
               initial)
              ((0.163973 -0.691165))
              ((0.54674 -0.476087)))
             ((R:segstate.parent.R:SylStructure.parent.syl_out < 6.1)
              ((0.29148 -0.370444))
              ((0.516084 -0.0187424))))
            ((n.name is hh_3)
             ((R:segstate.parent.R:SylStructure.parent.syl_out < 6.7)
              ((1.60274 1.39033))
              ((1.34302 0.410046)))
             ((R:segstate.parent.nn.ph_vc is +)
              ((R:segstate.parent.pp.ph_cplace is v)
               ((n.name is n_2)
                ((0.227067 -0.448266))
                ((R:segstate.parent.n.ph_ctype is f)
                 ((0.52717 0.266809))
                 ((R:segstate.parent.n.name is t)
                  ((0.348131 -0.145128))
                  ((0.579362 -0.0254538)))))
               ((R:segstate.parent.seg_onsetcoda is coda)
                ((R:segstate.parent.R:SylStructure.parent.syl_in < 5.7)
                 ((R:segstate.parent.nn.ph_vheight is 2)
                  ((R:segstate.parent.pp.ph_cplace is d)
                   ((0.779126 0.336996))
                   ((0.815254 -0.211453)))
                  ((1.52844 0.696019)))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((1.69651 0.38281))
                  ((R:segstate.parent.pp.ph_cplace is d)
                   ((1.66695 0.692251))
                   ((2.42224 1.66454)))))
                ((n.name is r_1)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                   <
                   0.3)
                  ((0.865929 0.206483))
                  ((0.422624 1.22454)))
                 ((n.name is f_3)
                  ((0.710446 0.6741))
                  ((R:segstate.parent.name is k)
                   ((n.name is k_3)
                    ((0.584829 -0.0213577))
                    ((0.976167 0.765858)))
                   ((R:segstate.parent.n.ph_ctype is r)
                    ((R:segstate.parent.nn.ph_vheight is 2)
                     ((0.724071 0.153626))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                       is
                       0)
                      ((0.64711 0.12632))
                      ((1.02698 0.703717))))
                    ((R:segstate.parent.name is f)
                     ((0.366501 -0.264399))
                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                       <
                       2.3)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                        <
                        0.3)
                       ((1.07493 0.261169))
                       ((0.455993 0.484864)))
                      ((R:segstate.parent.n.ph_cplace is a)
                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                         <
                         7.6)
                        ((0.548941 0.387488))
                        ((0.537904 0.095316)))
                       ((R:segstate.parent.lisp_onset_stop is 0)
                        ((0.541621 0.149107))
                        ((R:segstate.parent.n.name is p)
                         ((0.366546 -0.231316))
                         ((0.732661 -0.336387)))))))))))))
              ((n.name is ax_1)
               ((1.1215 0.974574))
               ((R:segstate.parent.R:SylStructure.parent.syl_out < 6)
                ((R:segstate.parent.n.ph_cvox is +)
                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                   <
                   1.2)
                  ((0.95605 -0.200846))
                  ((0.411701 -0.534808)))
                 ((n.name is s_3)
                  ((0.495994 0.388237))
                  ((n.name is f_2)
                   ((1.47483 0.551174))
                   ((n.name is m_3)
                    ((0.728256 0.369835))
                    ((R:segstate.parent.ph_cplace is p)
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                       is
                       0)
                      ((1.00275 0.390298))
                      ((0.825321 0.156531)))
                     ((R:segstate.parent.lisp_onset_stop is 0)
                      ((R:segstate.parent.ph_vlng is s)
                       ((1.22064 0.388444))
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         2.2)
                        ((R:segstate.parent.pp.ph_cplace is a)
                         ((R:segstate.parent.nn.ph_ctype is f)
                          ((0.418826 -0.60868))
                          ((0.504181 -0.240995)))
                         ((R:segstate.parent.n.ph_vrnd is +)
                          ((0.477776 -0.366135))
                          ((R:segstate.parent.ph_cplace is l)
                           ((0.362732 -0.345426))
                           ((R:segstate.parent.ph_cplace is a)
                            ((0.709151 -0.246366))
                            ((R:segstate.parent.ph_cplace is b)
                             ((0.602237 0.0716943))
                             ((1.17284 0.244062)))))))
                        ((R:segstate.parent.lisp_coda_glide is 0)
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           initial)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                            is
                            0)
                           ((0.845128 0.245181))
                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                             <
                             3.2)
                            ((0.656694 0.00150957))
                            ((R:segstate.parent.ph_cvox is +)
                             ((0.582612 -0.297345))
                             ((1.23091 -0.0577291)))))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                            <
                            0.2)
                           ((0.534946 -0.068474))
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             5.6)
                            ((0.617773 -0.209202))
                            ((0.402583 -0.472956)))))
                         ((1.31824 0.398536)))))
                      ((n.name is k_3)
                       ((0.442557 -0.393086))
                       ((R:segstate.parent.n.name is ae)
                        ((0.679067 0.630276))
                        ((R:segstate.parent.n.ph_vrnd is +)
                         ((0.710489 0.488495))
                         ((R:segstate.parent.name is b)
                          ((0.513603 -0.481611))
                          ((R:segstate.parent.ph_cplace is a)
                           ((0.696086 -0.118779))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                             <
                             0.3)
                            ((0.710632 0.0993291))
                            ((0.730836 0.414212))))))))))))))
                ((R:segstate.parent.n.name is uh)
                 ((2.02355 1.1235))
                 ((R:segstate.parent.ph_vrnd is -)
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    8.2)
                   ((2.27742 1.37633))
                   ((1.53048 0.26545)))
                  ((R:segstate.parent.name is b)
                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                     <
                     1.3)
                    ((0.874857 0.0722115))
                    ((0.399236 -0.537603)))
                   ((R:segstate.parent.name is r)
                    ((0.352071 -0.363279))
                    ((R:segstate.parent.n.ph_ctype is 0)
                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                       <
                       1.8)
                      ((R:segstate.parent.n.ph_vrnd is -)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                         is
                         0)
                        ((R:segstate.parent.n.name is ae)
                         ((0.458961 -0.169887))
                         ((R:segstate.parent.ph_cvox is +)
                          ((1.08435 0.409315))
                          ((0.759673 0.058968))))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                          is
                          0)
                         ((0.383833 -0.295916))
                         ((0.658246 0.112913))))
                       ((R:segstate.parent.lisp_onset_nasal is 0)
                        ((1.00639 0.401829))
                        ((0.932109 0.0631367))))
                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                        <
                        0.6)
                       ((R:segstate.parent.name is k)
                        ((0.554101 0.318155))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          3.4)
                         ((0.903513 0.210472))
                         ((R:segstate.parent.pp.ph_cplace is a)
                          ((0.890706 -0.183505))
                          ((0.94234 0.0551959)))))
                       ((R:segstate.parent.n.name is ih)
                        ((1.76035 0.932956))
                        ((R:segstate.parent.lisp_onset_stop is 0)
                         ((R:segstate.parent.pp.ph_cplace is d)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                            is
                            0)
                           ((0.81617 0.14208))
                           ((0.69262 -0.169142)))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            3.5)
                           ((1.1004 0.406754))
                           ((0.934141 0.356897))))
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           7.1)
                          ((0.855325 0.416782))
                          ((0.826823 0.708297)))))))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                       is
                       0)
                      ((0.599064 0.0464208))
                      ((R:segstate.parent.ph_cvox is -)
                       ((0.359162 -0.123912))
                       ((0.575368 -0.36029))))))))))))))))))))))
   ((R:segstate.parent.p.ph_cvox is 0)
    ((n.name is t_3)
     ((R:segstate.parent.pos_in_syl is 0)
      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
        is
        0)
       ((R:segstate.parent.nn.ph_ctype is n)
        ((0.966691 1.87353))
        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
          is
          0)
         ((0.815068 0.115099))
         ((1.12685 0.942655))))
       ((R:segstate.parent.R:SylStructure.parent.position_type
         is
         final)
        ((R:segstate.parent.n.name is ax)
         ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.4)
          ((0.0538318 -0.730512))
          ((0.135368 -0.685902)))
         ((R:segstate.parent.n.name is ih)
          ((R:segstate.parent.p.name is ey)
           ((0.110197 -0.67835))
           ((0.12212 -0.603172)))
          ((1.03854 -0.278895))))
        ((0.89827 0.286083))))
      ((R:segstate.parent.n.ph_ctype is s)
       ((R:segstate.parent.n.ph_cplace is l)
        ((2.54126e-06 -0.742549))
        ((R:segstate.parent.nn.ph_vfront is 2)
         ((0.252796 -0.61014))
         ((0.831283 -0.261062))))
       ((R:segstate.parent.n.name is hh)
        ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 1.6)
         ((0.691483 -0.35134))
         ((0.177378 -0.64625)))
        ((R:segstate.parent.nn.ph_vfront is 3)
         ((R:segstate.parent.n.ph_cvox is +)
          ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 6.6)
           ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 3.6)
            ((0.100836 -0.691865))
            ((0.0481487 -0.73292)))
           ((0.145972 -0.657581)))
          ((0.114971 -0.666524)))
         ((R:segstate.parent.pp.ph_cplace is d)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
            <
            0.2)
           ((0.120372 -0.652271))
           ((R:segstate.parent.R:SylStructure.parent.syl_in < 3.4)
            ((7.38713e-07 -0.742549))
            ((0.0946008 -0.700065))))
          ((R:segstate.parent.n.ph_ctype is 0)
           ((R:segstate.parent.nn.ph_ctype is r)
            ((3.00254e-06 -0.742549))
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              0.8)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
               <
               0.1)
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 1)
               ((0.128435 -0.677734))
               ((0.148226 -0.608803)))
              ((0.528016 -0.475056)))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
               is
               0)
              ((0.0746146 -0.717645))
              ((R:segstate.parent.n.name is pau)
               ((0.0982822 -0.694399))
               ((0.120372 -0.592084))))))
           ((R:segstate.parent.n.ph_cplace is d)
            ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 3)
             ((0.350056 -0.598104))
             ((0.0679691 -0.722487)))
            ((R:segstate.parent.R:SylStructure.parent.asyl_out < 3.5)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
               is
               0)
              ((0.46963 -0.418791))
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 4.5)
               ((0.115248 -0.667317))
               ((0.144505 -0.530129))))
             ((R:segstate.parent.R:SylStructure.parent.syl_out < 10.2)
              ((0.14291 -0.67835))
              ((0.200734 -0.614152)))))))))))
     ((n.name is pau_3)
      ((R:segstate.parent.pp.ph_vc is -)
       ((1.59894 0.721073))
       ((R:segstate.parent.nn.ph_ctype is s)
        ((R:segstate.parent.nn.ph_cplace is a)
         ((0.689708 -0.247745))
         ((0.596507 -0.178428)))
        ((R:segstate.parent.n.ph_vc is +)
         ((R:segstate.parent.nn.ph_ctype is n)
          ((R:segstate.parent.n.name is ae)
           ((0.185262 -0.455984))
           ((0.137693 -0.483994)))
          ((0.101854 -0.506488)))
         ((R:segstate.parent.nn.ph_ctype is n)
          ((R:segstate.parent.nn.ph_cplace is a)
           ((0.80457 -0.223051))
           ((0.768864 0.041834)))
          ((R:segstate.parent.n.ph_cplace is p)
           ((0.103588 -0.516588))
           ((R:segstate.parent.nn.ph_vheight is 3)
            ((R:segstate.parent.n.name is pau)
             ((R:segstate.parent.nn.ph_vfront is 2)
              ((0.366497 -0.425183))
              ((0.150578 -0.510336)))
             ((0.246348 -0.411727)))
            ((R:segstate.parent.nn.ph_vc is +)
             ((R:segstate.parent.n.name is pau)
              ((0.817398 -0.165494))
              ((R:segstate.parent.n.ph_cplace is a)
               ((0.259217 -0.442809))
               ((R:segstate.parent.n.ph_cplace is l)
                ((R:segstate.parent.nn.ph_vfront is 1)
                 ((0.12532 -0.528132))
                 ((0.462195 -0.380229)))
                ((R:segstate.parent.nn.ph_vfront is 1)
                 ((R:segstate.parent.nn.ph_vheight is 1)
                  ((R:segstate.parent.n.name is hh)
                   ((0.440885 -0.353174))
                   ((0.1901 -0.463738)))
                  ((0.365277 -0.338002)))
                 ((0.348676 -0.373552))))))
             ((R:segstate.parent.nn.ph_cplace is p)
              ((R:segstate.parent.nn.ph_ctype is r)
               ((0.856622 -0.203296))
               ((0.609973 -0.33189)))
              ((R:segstate.parent.nn.ph_ctype is r)
               ((0.317862 -0.382272))
               ((R:segstate.parent.nn.ph_cplace is a)
                ((0.274431 -0.472402))
                ((R:segstate.parent.nn.ph_cplace is g)
                 ((0.38048 -0.400615))
                 ((R:segstate.parent.nn.ph_cplace is d)
                  ((0.406701 -0.402539))
                  ((0.305213 -0.449344))))))))))))))
      ((R:segstate.parent.p.name is 0)
       ((R:segstate.parent.n.ph_cvox is -)
        ((R:segstate.parent.nn.ph_vheight is 2)
         ((R:segstate.parent.nn.ph_vfront is 2)
          ((1.0532 0.825762))
          ((0.654061 1.10212)))
         ((R:segstate.parent.nn.ph_vheight is 1)
          ((R:segstate.parent.n.ph_ctype is f)
           ((0.539434 0.422928))
           ((0.933465 0.727803)))
          ((R:segstate.parent.nn.ph_ctype is n)
           ((R:segstate.parent.nn.ph_cplace is a)
            ((0.68748 0.875425))
            ((0.48788 0.163492)))
           ((R:segstate.parent.nn.ph_ctype is r)
            ((R:segstate.parent.nn.ph_cplace is p)
             ((0.629599 0.0740092))
             ((0.70256 0.570472)))
            ((R:segstate.parent.nn.ph_cplace is a)
             ((0.729965 -0.777877))
             ((R:segstate.parent.nn.ph_ctype is f)
              ((R:segstate.parent.nn.ph_cplace is p)
               ((0.383952 0.225165))
               ((R:segstate.parent.nn.ph_cplace is g)
                ((0.582379 0.0898813))
                ((0.825472 -0.205842))))
              ((R:segstate.parent.nn.ph_vfront is 1)
               ((0.89131 0.35423))
               ((n.name is pau_1)
                ((R:segstate.parent.nn.ph_vfront is 2)
                 ((0.972314 -0.245825))
                 ((R:segstate.parent.nn.ph_ctype is 0)
                  ((0.719208 -0.517341))
                  ((0.68 -0.481671))))
                ((0.648774 -0.160899))))))))))
        ((R:segstate.parent.n.name is dh)
         ((R:segstate.parent.nn.ph_vfront is 1)
          ((0.577582 1.07827))
          ((0.775427 0.751282)))
         ((R:segstate.parent.n.ph_cplace is p)
          ((0.524638 1.03938))
          ((R:segstate.parent.nn.ph_ctype is s)
           ((0.667793 1.24553))
           ((R:segstate.parent.nn.ph_ctype is 0)
            ((R:segstate.parent.n.ph_cplace is l)
             ((R:segstate.parent.nn.ph_vheight is 2)
              ((0.164092 1.43673))
              ((0.43352 1.37399)))
             ((0.594906 1.19739)))
            ((n.name is ih_1)
             ((0.602785 1.38654))
             ((n.name is ae_1)
              ((0.107615 1.5501))
              ((0.088941 1.59949)))))))))
       ((n.name is d_1)
        ((R:segstate.parent.ph_vheight is 3)
         ((0.976472 0.247067))
         ((R:segstate.parent.name is l)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
            is
            0)
           ((0.226435 -0.559445))
           ((2.1744e-06 -0.646539)))
          ((R:segstate.parent.lisp_coda_nasal is 0)
           ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
             <
             0.3)
            ((0.428333 -0.481729))
            ((R:segstate.parent.pp.ph_cplace is a)
             ((0.54421 -0.314749))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
               is
               0)
              ((0.767932 -0.112047))
              ((0.447771 0.208372)))))
           ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.6)
            ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 0.3)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
               is
               0)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                is
                0)
               ((0.21005 -0.436917))
               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                 <
                 2.5)
                ((0.0819774 -0.547842))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                  <
                  1.2)
                 ((0.180965 -0.448113))
                 ((0.137844 -0.509931)))))
              ((0.0784461 -0.553922)))
             ((R:segstate.parent.p.ph_vheight is 2)
              ((R:segstate.parent.R:SylStructure.parent.last_accent
                <
                1.2)
               ((0.0576645 -0.569487))
               ((0.109291 -0.506208)))
              ((R:segstate.parent.nn.ph_ctype is s)
               ((0.385483 -0.204402))
               ((R:segstate.parent.pp.ph_cplace is a)
                ((0.713085 -0.222235))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                  <
                  0.2)
                 ((0.0800803 -0.552471))
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   6)
                  ((0.200942 -0.34509))
                  ((0.192358 -0.473967))))))))
            ((R:segstate.parent.nn.ph_ctype is f)
             ((0.597022 0.0598565))
             ((R:segstate.parent.nn.ph_ctype is s)
              ((0.721241 0.0437506))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                is
                0)
               ((0.302038 -0.386197))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                 is
                 0)
                ((0.0824597 -0.552007))
                ((0.182007 -0.468218))))))))))
        ((n.name is p_2)
         ((R:segstate.parent.R:SylStructure.parent.asyl_out < 2.1)
          ((R:segstate.parent.R:SylStructure.parent.syl_in < 4)
           ((1.00902 1.31842))
           ((R:segstate.parent.pp.ph_cplace is a)
            ((R:segstate.parent.p.ph_vheight is 1)
             ((0.54363 0.709135))
             ((0.642912 0.224946)))
            ((1.01581 0.966949))))
          ((R:segstate.parent.p.ph_vheight is 3)
           ((0.411209 0.224945))
           ((R:segstate.parent.n.seg_onsetcoda is coda)
            ((0.375072 0.447347))
            ((0.676969 0.687126)))))
         ((n.name is d_3)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
            <
            0.3)
           ((R:segstate.parent.syl_final is 0)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              0.7)
             ((0.572815 -0.241812))
             ((0.757546 0.483048)))
            ((R:segstate.parent.R:SylStructure.parent.position_type
              is
              final)
             ((0.666434 -0.118056))
             ((R:segstate.parent.R:SylStructure.parent.next_accent
               <
               1.8)
              ((0.243641 -0.547417))
              ((0.424603 -0.287118)))))
           ((R:segstate.parent.n.name is hh)
            ((0.541747 -0.301003))
            ((R:segstate.parent.n.name is iy)
             ((0.60039 -0.146681))
             ((R:segstate.parent.R:SylStructure.parent.last_accent
               <
               0.8)
              ((R:segstate.parent.p.ph_vheight is 1)
               ((R:segstate.parent.nn.ph_vc is +)
                ((0.769017 0.410501))
                ((0.601106 -0.118057)))
               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                 <
                 3.5)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                  <
                  0.3)
                 ((0.924795 0.372318))
                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                   <
                   4.5)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    4.2)
                   ((1.37279 1.36592))
                   ((0.981038 0.836641)))
                  ((0.828566 0.422937))))
                ((0.843037 0.270157))))
              ((R:segstate.parent.nn.ph_vheight is 3)
               ((0.737804 0.217175))
               ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.5)
                ((R:segstate.parent.R:SylStructure.parent.syl_out
                  <
                  5.8)
                 ((R:segstate.parent.p.ph_vlng is s)
                  ((1.67361 2.16261))
                  ((0.84454 0.839953)))
                 ((0.844098 0.772868)))
                ((R:segstate.parent.p.ph_vheight is 3)
                 ((0.834141 0.499746))
                 ((0.672553 0.99088)))))))))
          ((n.name is r_3)
           ((R:segstate.parent.p.name is ao)
            ((R:segstate.parent.nn.ph_vc is +)
             ((R:segstate.parent.n.name is dh)
              ((0.180199 -0.812508))
              ((R:segstate.parent.n.ph_cplace is a)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                 is
                 0)
                ((0.381969 -0.460919))
                ((0.23033 -0.73567)))
               ((R:segstate.parent.n.ph_cplace is l)
                ((0.30035 -0.59888))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                  is
                  0)
                 ((0.418065 -0.338774))
                 ((0.503267 0.00936112))))))
             ((R:segstate.parent.pp.ph_cplace is l)
              ((0.574572 -0.340889))
              ((R:segstate.parent.R:SylStructure.parent.last_accent
                <
                0.7)
               ((0.817635 0.63938))
               ((0.930545 0.0967877)))))
            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
              <
              0.3)
             ((R:segstate.parent.pos_in_syl is 2)
              ((R:segstate.parent.R:SylStructure.parent.syl_out < 8)
               ((0.883471 2.02314))
               ((1.15746 1.03071)))
              ((R:segstate.parent.n.ph_vrnd is -)
               ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.6)
                ((1.15101 0.480141))
                ((0.983023 0.734482)))
               ((1.65874 1.45362))))
             ((R:segstate.parent.pp.ph_cplace is d)
              ((R:segstate.parent.R:SylStructure.parent.last_accent
                <
                0.4)
               ((0.513676 -0.122138))
               ((0.457248 -0.443964)))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                <
                0.3)
               ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 2)
                ((1.07155 0.506227))
                ((R:segstate.parent.R:SylStructure.parent.syl_out < 4)
                 ((1.09598 0.998797))
                 ((0.960668 1.49072))))
               ((R:segstate.parent.p.ph_vheight is 1)
                ((R:segstate.parent.pp.ph_cplace is p)
                 ((1.45268 0.263104))
                 ((0.518903 -0.378966)))
                ((R:segstate.parent.pos_in_syl is 2)
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   2.4)
                  ((0.640354 0.635886))
                  ((0.743773 1.40112)))
                 ((R:segstate.parent.pp.ph_cplace is a)
                  ((0.902235 -0.0828004))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                    <
                    0.3)
                   ((0.663134 0.156092))
                   ((0.907542 0.758974))))))))))
           ((R:segstate.parent.name is dh)
            ((n.name is er_1)
             ((1.00702 1.87418))
             ((n.name is ax_1)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                <
                0.3)
               ((0.739694 0.0411388))
               ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 3)
                ((1.0761 0.925404))
                ((1.27617 2.22713))))
              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                <
                1.2)
               ((n.name is dh_3)
                ((R:segstate.parent.R:SylStructure.parent.last_accent
                  <
                  2.4)
                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                   <
                   4.6)
                  ((0.134776 -0.653606))
                  ((0.536289 -0.39079)))
                 ((8.43339e-06 -0.684522)))
                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                  <
                  3.7)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                   is
                   0)
                  ((1.43881 0.768316))
                  ((0.979739 0.00405235)))
                 ((0.409823 -0.495887))))
               ((R:segstate.parent.n.ph_ctype is f)
                ((R:segstate.parent.n.name is hh)
                 ((1.00638 0.447936))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                   is
                   0)
                  ((0.552998 -0.448441))
                  ((0.0646533 -0.634236))))
                ((n.name is pau_5)
                 ((0.201535 -0.526231))
                 ((n.name is dh_2)
                  ((R:segstate.parent.p.ph_vlng is s)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                     <
                     0.3)
                    ((0.990714 1.75967))
                    ((0.794366 1.00775)))
                   ((0.761489 -0.242372)))
                  ((R:segstate.parent.nn.ph_ctype is 0)
                   ((0.90909 -0.293998))
                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                     <
                     4.5)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                      <
                      0.2)
                     ((0.717679 0.368737))
                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                       <
                       3.2)
                      ((1.39736 0.700659))
                      ((1.81207 1.31448))))
                    ((1.02939 5.51257e-05))))))))))
            ((n.name is ax_1)
             ((R:segstate.parent.name is sh)
              ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                <
                3.2)
               ((0.933248 -0.20035))
               ((0.964815 0.514673)))
              ((R:segstate.parent.name is l)
               ((R:segstate.parent.R:SylStructure.parent.position_type
                 is
                 final)
                ((0.690946 0.178322))
                ((0.360309 -0.210758)))
               ((R:segstate.parent.p.name is ow)
                ((0.849075 0.193349))
                ((R:segstate.parent.R:SylStructure.parent.position_type
                  is
                  initial)
                 ((0.485508 0.167412))
                 ((R:segstate.parent.name is z)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    2.4)
                   ((0.365217 -0.576733))
                   ((0.135698 -0.761239)))
                  ((R:segstate.parent.lisp_coda_glide is 0)
                   ((R:segstate.parent.ph_cplace is v)
                    ((R:segstate.parent.R:SylStructure.parent.position_type
                      is
                      final)
                     ((0.255638 -0.515264))
                     ((0.260063 -0.764001)))
                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                      <
                      2.4)
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                       <
                       4.8)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                        <
                        0.3)
                       ((R:segstate.parent.ph_cplace is a)
                        ((R:segstate.parent.lisp_coda_stop is 0)
                         ((0.645154 -0.24013))
                         ((0.356823 -0.418287)))
                        ((0.329651 -0.586882)))
                       ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                         <
                         1.2)
                        ((0.984199 0.0609207))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                          <
                          0.3)
                         ((0.378105 -0.436082))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           3.2)
                          ((0.376856 -0.310713))
                          ((0.522946 -0.107976))))))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                        <
                        0.3)
                       ((0.239335 -0.651901))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                         <
                         0.3)
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                          <
                          0.1)
                         ((0.334932 -0.460681))
                         ((0.348447 -0.236154)))
                        ((0.211017 -0.526389)))))
                     ((R:segstate.parent.lisp_onset_nasal is 0)
                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                        <
                        0.7)
                       ((0.254038 -0.66164))
                       ((0.234167 -0.491452)))
                      ((0.448016 -0.365131)))))
                   ((R:segstate.parent.ph_cplace is a)
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      8)
                     ((0.221939 -0.505075))
                     ((0.04054 -0.637578)))
                    ((0.301013 -0.716211)))))))))
             ((n.name is l_2)
              ((R:segstate.parent.p.ph_vheight is 1)
               ((R:segstate.parent.lisp_onset_glide is 0)
                ((R:segstate.parent.lisp_onset_stop is 0)
                 ((1.3352 0.672167))
                 ((0.855201 -0.190693)))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                  <
                  0.3)
                 ((0.181452 -0.692119))
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   0.5)
                  ((0.254835 -0.552169))
                  ((0.685672 -0.228743)))))
               ((R:segstate.parent.pos_in_syl is 2)
                ((R:segstate.parent.p.name is ow)
                 ((1.27287 1.56396))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                    <
                    0.2)
                   ((0.798234 1.42734))
                   ((0.96441 0.721248)))
                  ((R:segstate.parent.n.ph_cplace is a)
                   ((0.679895 0.250791))
                   ((0.574885 0.646504)))))
                ((R:segstate.parent.p.name is aa)
                 ((0.584614 1.0014))
                 ((R:segstate.parent.R:SylStructure.parent.position_type
                   is
                   single)
                  ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                    <
                    0.2)
                   ((R:segstate.parent.n.seg_onsetcoda is coda)
                    ((0.660355 1.30892))
                    ((0.853147 0.48777)))
                   ((0.886725 -0.014299)))
                  ((R:segstate.parent.p.ph_vlng is s)
                   ((R:segstate.parent.nn.ph_cplace is 0)
                    ((0.753583 -0.0693394))
                    ((0.947577 0.432573)))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                     is
                     0)
                    ((R:segstate.parent.p.ph_vrnd is +)
                     ((0.753617 0.17939))
                     ((0.362998 -0.300857)))
                    ((0.496489 -0.366417))))))))
              ((n.name is d_2)
               ((R:segstate.parent.syl_final is 0)
                ((R:segstate.parent.p.name is iy)
                 ((0.254077 -0.467543))
                 ((R:segstate.parent.p.ph_vlng is s)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                    is
                    0)
                   ((0.264371 -0.386405))
                   ((0.495049 -0.139582)))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                    <
                    0.1)
                   ((0.980981 0.344846))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                     is
                     0)
                    ((0.474304 0.201853))
                    ((0.39052 -0.186349))))))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                  <
                  0.3)
                 ((R:segstate.parent.lisp_onset_stop is 0)
                  ((R:segstate.parent.p.ph_vlng is s)
                   ((0.37856 -0.300666))
                   ((1.2159 1.173)))
                  ((1.96352 2.29572)))
                 ((R:segstate.parent.lisp_onset_nasal is 0)
                  ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                    <
                    2.5)
                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                     <
                     0.8)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                      is
                      0)
                     ((1.88972 1.24261))
                     ((R:segstate.parent.nn.ph_vheight is 1)
                      ((0.80887 0.752846))
                      ((R:segstate.parent.p.ph_vheight is 3)
                       ((0.648599 0.367772))
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         7)
                        ((0.99333 0.277271))
                        ((0.659222 -0.120224))))))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      2.7)
                     ((R:segstate.parent.lisp_onset_stop is 0)
                      ((0.612918 0.247103))
                      ((0.730019 0.422072)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                       <
                       0.3)
                      ((0.337859 -0.332102))
                      ((R:segstate.parent.p.ph_vheight is 1)
                       ((0.561806 -0.225512))
                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                         <
                         1.4)
                        ((0.658492 0.135308))
                        ((0.636884 0.397939)))))))
                   ((1.11457 1.03303)))
                  ((0.331757 -0.346181)))))
               ((n.name is z_2)
                ((R:segstate.parent.p.ph_vlng is s)
                 ((R:segstate.parent.pos_in_syl is 0)
                  ((0.615718 -0.0539795))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    4)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                     is
                     0)
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      4.4)
                     ((0.965403 0.875528))
                     ((1.03618 1.39389)))
                    ((R:segstate.parent.n.ph_cvox is -)
                     ((R:segstate.parent.n.name is pau)
                      ((0.739969 0.687411))
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        3.2)
                       ((0.93 0.457495))
                       ((0.727263 0.102167))))
                     ((0.928271 0.869096))))
                   ((R:segstate.parent.nn.ph_ctype is s)
                    ((0.814368 0.950334))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                      is
                      0)
                     ((1.05583 -0.0137417))
                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                       <
                       1.2)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        1.4)
                       ((0.658504 0.0878655))
                       ((0.70009 0.401757)))
                      ((1.03568 0.651109)))))))
                 ((R:segstate.parent.lisp_onset_glide is 0)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    6.8)
                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                     <
                     3.6)
                    ((R:segstate.parent.p.ph_vheight is 2)
                     ((0.476874 0.0115932))
                     ((0.702825 -0.278242)))
                    ((R:segstate.parent.p.ph_vheight is 2)
                     ((0.480768 -0.300537))
                     ((0.410478 -0.638679))))
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     0.7)
                    ((0.565813 -0.190455))
                    ((1.27765 0.478394))))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    4.5)
                   ((R:segstate.parent.nn.ph_ctype is r)
                    ((0.578841 -0.190454))
                    ((R:segstate.parent.nn.ph_vc is +)
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                       <
                       4)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                        <
                        0.3)
                       ((0.53185 -0.0650472))
                       ((0.6712 -0.256961)))
                      ((R:segstate.parent.n.ph_cvox is +)
                       ((0.750859 0.397776))
                       ((0.610194 -0.0650447))))
                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                       <
                       0.9)
                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                        <
                        4)
                       ((0.770942 0.189953))
                       ((0.924257 0.145363)))
                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                        <
                        10)
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         5.4)
                        ((0.91032 0.412904))
                        ((0.842883 0.881101)))
                       ((0.798144 0.175872))))))
                   ((1.39963 0.658152)))))
                ((R:segstate.parent.seg_onsetcoda is coda)
                 ((R:segstate.parent.name is l)
                  ((R:segstate.parent.p.ph_vheight is 1)
                   ((n.name is l_3)
                    ((R:segstate.parent.p.ph_vlng is l)
                     ((1.09902 1.13883))
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       7.4)
                      ((0.928677 0.314858))
                      ((0.690088 -0.289792))))
                    ((R:segstate.parent.p.name is ih)
                     ((0.387264 -0.320937))
                     ((0.225349 -0.490316))))
                   ((R:segstate.parent.n.ph_vc is +)
                    ((0.844949 -0.178343))
                    ((R:segstate.parent.n.name is d)
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                       is
                       0)
                      ((0.650412 -0.23046))
                      ((0.503207 -0.540648)))
                     ((n.name is l_3)
                      ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                        <
                        0.1)
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                         <
                         4)
                        ((R:segstate.parent.pp.ph_cplace is a)
                         ((0.033033 -1.00624))
                         ((0.12782 -0.924143)))
                        ((0.232775 -0.808644)))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                         <
                         1.2)
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          3.2)
                         ((0.722018 -0.327107))
                         ((0.294821 -0.709139)))
                        ((R:segstate.parent.n.ph_cvox is +)
                         ((0.33439 -0.643749))
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           final)
                          ((0.0942615 -0.938003))
                          ((0.161828 -0.861241))))))
                      ((R:segstate.parent.n.name is pau)
                       ((0.495233 -0.272699))
                       ((R:segstate.parent.n.ph_cplace is l)
                        ((0.0611105 -0.623928))
                        ((R:segstate.parent.nn.ph_vrnd is +)
                         ((0.252809 -0.427567))
                         ((R:segstate.parent.n.ph_cplace is a)
                          ((R:segstate.parent.nn.ph_ctype is 0)
                           ((0.117396 -0.59664))
                           ((0.114405 -0.538004)))
                          ((R:segstate.parent.n.ph_cplace is b)
                           ((0.0905738 -0.572843))
                           ((0.272024 -0.385252)))))))))))
                  ((n.name is s_2)
                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                     <
                     1.1)
                    ((R:segstate.parent.p.ph_vheight is 2)
                     ((R:segstate.parent.nn.ph_vc is +)
                      ((0.77129 0.491095))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                        <
                        1.2)
                       ((0.372612 0.0467151))
                       ((0.556425 0.309369))))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                       is
                       0)
                      ((0.256388 -0.225124))
                      ((0.329184 0.0254208))))
                    ((R:segstate.parent.lisp_onset_glide is 0)
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       7)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        5.6)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                         is
                         0)
                        ((0.412068 0.344861))
                        ((0.315566 0.0920853)))
                       ((0.503442 0.612249)))
                      ((0.486291 0.644945)))
                     ((1.61842 0.986303))))
                   ((n.name is iy_2)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                      is
                      0)
                     ((0.88284 0.529522))
                     ((1.51347 1.45894)))
                    ((n.name is f_3)
                     ((R:segstate.parent.p.ph_vlng is s)
                      ((R:segstate.parent.n.seg_onsetcoda is coda)
                       ((0.664785 -0.677796))
                       ((0.518421 -1.134)))
                      ((0.518141 -0.361207)))
                     ((n.name is k_2)
                      ((R:segstate.parent.p.name is ay)
                       ((0.782565 0.921449))
                       ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                         <
                         3)
                        ((R:segstate.parent.n.name is s)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                           is
                           0)
                          ((0.657816 0.58643))
                          ((0.613868 0.199393)))
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           10)
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            3)
                           ((R:segstate.parent.nn.ph_ctype is 0)
                            ((0.958158 0.759539))
                            ((0.833359 -0.0257878)))
                           ((R:segstate.parent.p.ph_vheight is 3)
                            ((0.951112 -0.46912))
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              3.4)
                             ((0.64349 0.361659))
                             ((0.753207 -0.0533247)))))
                          ((0.71573 -0.297215))))
                        ((0.885792 0.965188))))
                      ((R:segstate.parent.name is k)
                       ((n.name is k_3)
                        ((R:segstate.parent.n.ph_ctype is s)
                         ((R:segstate.parent.R:SylStructure.parent.asyl_out
                           <
                           2.2)
                          ((1.19924 1.5354))
                          ((1.22228 0.614575)))
                         ((R:segstate.parent.n.ph_ctype is f)
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            7.6)
                           ((R:segstate.parent.lisp_onset_stop is 0)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                              is
                              0)
                             ((0.353087 -0.79665))
                             ((0.414535 -0.548742)))
                            ((0.538692 -0.30297)))
                           ((0.811549 -0.148487)))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                            <
                            0.2)
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             2.4)
                            ((0.511893 -0.528261))
                            ((0.762158 -0.242893)))
                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                             <
                             8)
                            ((0.648172 0.449033))
                            ((0.776576 0.00742883))))))
                        ((n.name is t_1)
                         ((R:segstate.parent.p.ph_vheight is 3)
                          ((0.589463 -0.984706))
                          ((0.217878 -1.30044)))
                         ((R:segstate.parent.pp.ph_cplace is a)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                            is
                            0)
                           ((R:segstate.parent.n.ph_ctype is 0)
                            ((0.393422 -0.959848))
                            ((0.749604 -0.549313)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                             <
                             0.3)
                            ((0.42513 -0.824257))
                            ((0.349603 -1.14911))))
                          ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                            <
                            2.2)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                             <
                             1.2)
                            ((0.728597 -0.374553))
                            ((0.460206 -0.699967)))
                           ((0.553265 -0.821997))))))
                       ((n.name is s_3)
                        ((R:segstate.parent.n.ph_ctype is 0)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                           <
                           0.2)
                          ((0.410365 0.175731))
                          ((0.501023 -0.248299)))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           6.6)
                          ((R:segstate.parent.p.name is ih)
                           ((0.686367 -0.25522))
                           ((R:segstate.parent.pp.ph_cplace is a)
                            ((0.774775 -0.406892))
                            ((R:segstate.parent.nn.ph_cplace is 0)
                             ((R:segstate.parent.nn.ph_vheight is 2)
                              ((0.652495 -1.30013))
                              ((0.492821 -0.994988)))
                             ((0.70868 -0.694264)))))
                          ((0.887716 -0.15941))))
                        ((n.name is m_3)
                         ((R:segstate.parent.n.ph_cvox is -)
                          ((R:segstate.parent.n.name is p)
                           ((R:segstate.parent.lisp_onset_stop is 0)
                            ((0.641307 0.325654))
                            ((0.550974 0.0272667)))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             4)
                            ((R:segstate.parent.nn.ph_vfront is 1)
                             ((0.605259 0.627618))
                             ((0.809673 1.12299)))
                            ((R:segstate.parent.nn.ph_vfront is 1)
                             ((0.672266 0.703518))
                             ((0.677762 0.33578)))))
                          ((R:segstate.parent.n.ph_ctype is f)
                           ((0.885184 0.494886))
                           ((R:segstate.parent.p.ph_vheight is 2)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                              is
                              0)
                             ((0.868741 0.0421555))
                             ((0.867406 -0.791464)))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                              <
                              1.2)
                             ((0.76323 0.247319))
                             ((0.896029 -0.167937))))))
                         ((n.name is ch_2)
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            0.5)
                           ((0.798852 0.772594))
                           ((0.364012 0.40213)))
                          ((n.name is t_2)
                           ((R:segstate.parent.n.name is dh)
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              4)
                             ((1.32931 2.0694))
                             ((1.19662 1.25892)))
                            ((R:segstate.parent.n.ph_cplace is a)
                             ((R:segstate.parent.n.name is d)
                              ((1.33806 1.74214))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                6.6)
                               ((R:segstate.parent.n.name is s)
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  2.1)
                                 ((0.55679 1.28308))
                                 ((0.684072 0.928305)))
                                ((1.19467 0.445301)))
                               ((0.749005 -0.0582288))))
                             ((R:segstate.parent.p.ph_vlng is d)
                              ((R:segstate.parent.R:SylStructure.parent.next_accent
                                <
                                0.7)
                               ((1.64503 0.840863))
                               ((R:segstate.parent.R:SylStructure.parent.position_type
                                 is
                                 final)
                                ((0.353311 -0.593497))
                                ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                  <
                                  0.2)
                                 ((0.816833 -0.151475))
                                 ((1.11675 0.320241)))))
                              ((R:segstate.parent.n.ph_cvox is -)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                 <
                                 3.6)
                                ((R:segstate.parent.pp.ph_cplace is l)
                                 ((0.773201 0.155161))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                   is
                                   0)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                    <
                                    0.2)
                                   ((0.337306 -0.460929))
                                   ((0.203441 -0.677295)))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    5)
                                   ((0.839823 0.0804453))
                                   ((0.618474 -0.225698)))))
                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                  <
                                  1.2)
                                 ((1.2195 0.0206713))
                                 ((1.24378 0.623184))))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                 is
                                 0)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                  <
                                  1.2)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                   <
                                   11)
                                  ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                    <
                                    2.5)
                                   ((0.560211 -0.55315))
                                   ((0.271889 -0.655079)))
                                  ((0.518336 -0.431525)))
                                 ((R:segstate.parent.p.ph_vheight is 1)
                                  ((1.29927 0.565804))
                                  ((R:segstate.parent.p.name is ae)
                                   ((0.683723 -0.270721))
                                   ((0.380647 -0.539701)))))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                  is
                                  0)
                                 ((0.0555601 -0.753989))
                                 ((R:segstate.parent.nn.ph_ctype is 0)
                                  ((0.270766 -0.68764))
                                  ((0.332202 -0.526252)))))))))
                           ((R:segstate.parent.name is t)
                            ((R:segstate.parent.n.ph_cplace is l)
                             ((R:segstate.parent.R:SylStructure.parent.syl_in
                               <
                               5.1)
                              ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                <
                                2.2)
                               ((R:segstate.parent.pp.ph_cplace is a)
                                ((0.250892 -0.418923))
                                ((0.197372 -0.495266)))
                               ((R:segstate.parent.pp.ph_cplace is 0)
                                ((0.157631 -0.554533))
                                ((0.0887037 -0.567794))))
                              ((6.27268e-06 -0.590698)))
                             ((R:segstate.parent.p.name is ae)
                              ((R:segstate.parent.pp.ph_cplace is 0)
                               ((0.279629 -0.39745))
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 1.1)
                                ((R:segstate.parent.n.ph_ctype is 0)
                                 ((0.146967 -0.514353))
                                 ((4.46116e-06 -0.590697)))
                                ((0.229849 -0.459821))))
                              ((R:segstate.parent.p.ph_vheight is 1)
                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                 <
                                 8.6)
                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                  <
                                  0.8)
                                 ((0.582146 -0.156741))
                                 ((R:segstate.parent.n.seg_onsetcoda
                                   is
                                   coda)
                                  ((0.219082 -0.467018))
                                  ((0.117345 -0.547756))))
                                ((0.0673757 -0.577484)))
                               ((R:segstate.parent.nn.ph_ctype is f)
                                ((0.145547 -0.514352))
                                ((R:segstate.parent.nn.ph_vfront is 2)
                                 ((R:segstate.parent.lisp_onset_stop
                                   is
                                   0)
                                  ((0.271493 -0.426391))
                                  ((0.177407 -0.544891)))
                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                   <
                                   2.4)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                    is
                                    0)
                                   ((R:segstate.parent.pp.ph_cplace
                                     is
                                     l)
                                    ((0.412077 -0.396519))
                                    ((0.181597 -0.53007)))
                                   ((R:segstate.parent.lisp_onset_stop
                                     is
                                     0)
                                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                                      <
                                      0.2)
                                     ((0.673168 -0.0925514))
                                     ((0.364975 -0.433237)))
                                    ((1.11499 0.0964022))))
                                  ((1.29459 0.0964015))))))))
                            ((n.name is g_2)
                             ((R:segstate.parent.lisp_onset_stop is 0)
                              ((0.510288 0.0913576))
                              ((1.32522 0.867262)))
                             ((R:segstate.parent.n.name is n)
                              ((R:segstate.parent.name is d)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 3.6)
                                ((0.287928 -0.542319))
                                ((1.14488e-06 -0.664525)))
                               ((n.name is ae_2)
                                ((1.39362 1.10264))
                                ((R:segstate.parent.p.ph_vheight is 1)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                   is
                                   0)
                                  ((R:segstate.parent.R:SylStructure.parent.position_type
                                    is
                                    initial)
                                   ((1.04169 0.335789))
                                   ((2.24088 1.17593)))
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                    <
                                    3.6)
                                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                                     <
                                     0.4)
                                    ((0.759278 -0.085626))
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      7.1)
                                     ((1.16282 0.494863))
                                     ((1.66859 0.904664))))
                                   ((n.name is n_1)
                                    ((0.622285 -0.229008))
                                    ((R:segstate.parent.ph_vlng is s)
                                     ((0.632012 -0.201106))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                       <
                                       2.2)
                                      ((0.757784 0.377789))
                                      ((0.902334 0.241453)))))))
                                 ((R:segstate.parent.name is ax)
                                  ((0.122262 -0.473738))
                                  ((n.name is n_1)
                                   ((R:segstate.parent.lisp_coda_nasal
                                     is
                                     0)
                                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                                      <
                                      0.4)
                                     ((0.404397 -0.607842))
                                     ((0.454486 -0.302264)))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                      is
                                      0)
                                     ((1.06354 0.35741))
                                     ((0.482549 -0.447214))))
                                   ((R:segstate.parent.ph_vheight is 3)
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      6.2)
                                     ((0.224643 -0.600057))
                                     ((0.501102 -0.162904)))
                                    ((R:segstate.parent.pp.ph_cplace
                                      is
                                      a)
                                     ((0.932244 0.0926928))
                                     ((R:segstate.parent.p.ph_vheight
                                       is
                                       2)
                                      ((1.17719 0.250087))
                                      ((R:segstate.parent.pp.ph_cplace
                                        is
                                        l)
                                       ((0.994497 0.925767))
                                       ((0.964316 0.446149)))))))))))
                              ((R:segstate.parent.name is ah)
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 6.2)
                                ((R:segstate.parent.p.ph_vlng is l)
                                 ((0.846467 -0.210682))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                   <
                                   0.3)
                                  ((0.126285 -0.739374))
                                  ((0.442723 -0.492073))))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                  is
                                  0)
                                 ((0.126304 -0.741143))
                                 ((0.307066 -0.605775))))
                               ((R:segstate.parent.n.name is pau)
                                ((R:segstate.parent.ph_vlng is a)
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   1.2)
                                  ((2.10249 0.931725))
                                  ((1.67666 0.470248)))
                                 ((R:segstate.parent.nn.ph_cplace is p)
                                  ((R:segstate.parent.nn.ph_ctype is a)
                                   ((1.0882 0.122186))
                                   ((1.59319 0.788535)))
                                  ((n.name is ng_3)
                                   ((R:segstate.parent.pp.ph_cplace
                                     is
                                     a)
                                    ((0.968889 0.094864))
                                    ((0.911365 0.76348)))
                                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                     <
                                     2.1)
                                    ((R:segstate.parent.lisp_onset_glide
                                      is
                                      0)
                                     ((R:segstate.parent.p.name is aa)
                                      ((1.56968 0.710082))
                                      ((R:segstate.parent.p.name is ao)
                                       ((1.25396 0.487963))
                                       ((R:segstate.parent.nn.ph_vfront
                                         is
                                         3)
                                        ((0.346464 -0.527866))
                                        ((R:segstate.parent.nn.ph_vheight
                                          is
                                          1)
                                         ((0.492429 -0.391821))
                                         ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                                           <
                                           0.2)
                                          ((n.name is n_3)
                                           ((0.897061 0.501974))
                                           ((n.name is z_3)
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                              <
                                              1.2)
                                             ((1.25551 0.695497))
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                               is
                                               0)
                                              ((0.679329 0.114842))
                                              ((0.828217 0.143715))))
                                            ((R:segstate.parent.name
                                              is
                                              z)
                                             ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                               <
                                               6)
                                              ((R:segstate.parent.nn.ph_cplace
                                                is
                                                l)
                                               ((0.358165 -0.342354))
                                               ((0.480814 -0.101306)))
                                              ((0.341078 -0.506184)))
                                             ((R:segstate.parent.R:SylStructure.parent.next_accent
                                               <
                                               2.1)
                                              ((R:segstate.parent.ph_cplace
                                                is
                                                p)
                                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                 <
                                                 5.5)
                                                ((0.748675 -0.378))
                                                ((0.830742 -0.0296115)))
                                               ((n.name is pau_5)
                                                ((R:segstate.parent.ph_cvox
                                                  is
                                                  -)
                                                 ((0.856863 -0.0827352))
                                                 ((R:segstate.parent.p.ph_vheight
                                                   is
                                                   3)
                                                  ((0.425509 0.952248))
                                                  ((0.985253 0.434731))))
                                                ((R:segstate.parent.name
                                                  is
                                                  v)
                                                 ((0.550791 -0.345565))
                                                 ((R:segstate.parent.p.name
                                                   is
                                                   eh)
                                                  ((0.828904 -0.205077))
                                                  ((R:segstate.parent.nn.ph_ctype
                                                    is
                                                    r)
                                                   ((0.847189
                                                     -0.162318))
                                                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                     <
                                                     3.6)
                                                    ((n.name is pau_1)
                                                     ((R:segstate.parent.p.name
                                                       is
                                                       ih)
                                                      ((1.616 0.729224))
                                                      ((R:segstate.parent.R:SylStructure.parent.position_type
                                                        is
                                                        single)
                                                       ((0.861187
                                                         -0.113998))
                                                       ((0.842488
                                                         -0.233596))))
                                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                       <
                                                       2.4)
                                                      ((2.23357 1.2408))
                                                      ((R:segstate.parent.nn.ph_ctype
                                                        is
                                                        0)
                                                       ((1.18567
                                                         0.501865))
                                                       ((0.575238
                                                         -0.0999922)))))
                                                    ((0.668822
                                                      -0.250225))))))))
                                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                <
                                                4.6)
                                               ((0.785218 -0.0926281))
                                               ((0.624952 -0.3706)))))))
                                          ((R:segstate.parent.R:SylStructure.parent.last_accent
                                            <
                                            0.4)
                                           ((0.361734 -0.390057))
                                           ((0.520295 -0.100279))))))))
                                     ((n.name is pau_1)
                                      ((0.501566 -0.442297))
                                      ((n.name is pau_5)
                                       ((R:segstate.parent.pp.ph_cplace
                                         is
                                         l)
                                        ((R:segstate.parent.R:SylStructure.parent.syl_in
                                          <
                                          2.1)
                                         ((0.374933 -0.461779))
                                         ((0.533027 -0.177525)))
                                        ((0.704313 0.112419)))
                                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                         <
                                         3.5)
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                          <
                                          2.7)
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                           <
                                           0.3)
                                          ((0.62763 -0.0296263))
                                          ((R:segstate.parent.ph_cplace
                                            is
                                            a)
                                           ((0.828404 0.526903))
                                           ((0.598531 0.00490626))))
                                         ((0.508056 -0.164338)))
                                        ((0.646911 -0.238992))))))
                                    ((R:segstate.parent.nn.ph_vheight
                                      is
                                      2)
                                     ((0.373877 -0.493038))
                                     ((R:segstate.parent.p.ph_vlng
                                       is
                                       s)
                                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                                        <
                                        4.3)
                                       ((R:segstate.parent.lisp_coda_nasal
                                         is
                                         0)
                                        ((R:segstate.parent.R:SylStructure.parent.syl_out
                                          <
                                          2.2)
                                         ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                           <
                                           7.6)
                                          ((0.529349 -0.344785))
                                          ((0.597915 -0.132316)))
                                         ((R:segstate.parent.p.name
                                           is
                                           ah)
                                          ((0.851058 -0.0378379))
                                          ((0.99666 0.292687))))
                                        ((0.824065 0.376329)))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                         <
                                         1.2)
                                        ((0.424204 -0.418347))
                                        ((0.850667 -0.0564215))))
                                      ((R:segstate.parent.p.ph_vheight
                                        is
                                        2)
                                       ((0.446019 -0.543698))
                                       ((R:segstate.parent.R:SylStructure.parent.last_accent
                                         <
                                         0.5)
                                        ((0.915045 -0.0208191))
                                        ((0.481392 -0.393124))))))))))
                                ((R:segstate.parent.n.name is d)
                                 ((R:segstate.parent.p.ph_vheight is 3)
                                  ((R:segstate.parent.nn.ph_cplace
                                    is
                                    g)
                                   ((n.name is n_3)
                                    ((0.446346 -0.709482))
                                    ((0.45101 -0.289448)))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     2.1)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                      <
                                      0.3)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                       is
                                       0)
                                      ((0.735955 -0.174011))
                                      ((0.8141 0.135325)))
                                     ((R:segstate.parent.R:SylStructure.parent.position_type
                                       is
                                       single)
                                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                        <
                                        1.5)
                                       ((0.58993 0.171722))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                         <
                                         0.3)
                                        ((0.872203 0.222114))
                                        ((0.805023 0.642995))))
                                      ((0.499486 0.0692314))))
                                    ((R:segstate.parent.nn.ph_vc is +)
                                     ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                       <
                                       0.2)
                                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                                        <
                                        0.6)
                                       ((n.name is n_3)
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                          is
                                          0)
                                         ((0.594194 -0.507782))
                                         ((0.423123 -0.785119)))
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                          <
                                          0.1)
                                         ((0.400628 -0.215171))
                                         ((0.200183 -0.446955))))
                                       ((0.562921 -0.133225)))
                                      ((n.name is n_3)
                                       ((0.391093 0.437682))
                                       ((0.443584 -0.265151))))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                       <
                                       2.4)
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                        <
                                        0.1)
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                         <
                                         0.3)
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                          is
                                          0)
                                         ((0.490486 -0.0420563))
                                         ((0.740233 -0.0420563)))
                                        ((n.name is n_3)
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                           <
                                           0.1)
                                          ((0.547822 -0.371635))
                                          ((0.624866 -0.763059)))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           6)
                                          ((0.263921 -0.264678))
                                          ((0.51733 0.07211)))))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                         <
                                         1.2)
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                          <
                                          2.7)
                                         ((n.name is n_3)
                                          ((0.574188 0.242012))
                                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                            <
                                            5.4)
                                           ((0.507355 -0.0710573))
                                           ((0.554038 -0.127559))))
                                         ((0.531094 -0.347929)))
                                        ((R:segstate.parent.R:SylStructure.parent.syl_out
                                          <
                                          4.2)
                                         ((0.888914 0.315721))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           3.3)
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            a)
                                           ((0.693811 0.223442))
                                           ((0.400325 -0.0106338)))
                                          ((0.555784 -0.208926))))))
                                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                        <
                                        4.5)
                                       ((1.12504 0.52976))
                                       ((0.539897 0.0477937)))))))
                                  ((n.name is n_3)
                                   ((R:segstate.parent.nn.ph_ctype
                                     is
                                     0)
                                    ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                                      <
                                      0.1)
                                     ((0.353837 1.01253))
                                     ((0.582459 1.31738)))
                                    ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                      <
                                      0.3)
                                     ((0.430535 1.03522))
                                     ((0.651882 0.492692))))
                                   ((R:segstate.parent.p.ph_vlng is s)
                                    ((R:segstate.parent.ph_cplace is a)
                                     ((R:segstate.parent.p.name is ah)
                                      ((0.759582 0.221951))
                                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                                        <
                                        4)
                                       ((0.154511 -0.47644))
                                       ((0.530162 -0.103421))))
                                     ((R:segstate.parent.R:SylStructure.parent.position_type
                                       is
                                       final)
                                      ((0.695933 0.160833))
                                      ((1.26241 0.871508))))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                      <
                                      0.3)
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       8)
                                      ((0.511454 -0.306113))
                                      ((0.518186 -0.603735)))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                       <
                                       2.4)
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                        <
                                        1.1)
                                       ((1.03409 0.446687))
                                       ((0.34579 -0.277035)))
                                      ((0.603447 -0.373335)))))))
                                 ((R:segstate.parent.n.name is ax)
                                  ((R:segstate.parent.lisp_coda_nasal
                                    is
                                    0)
                                   ((R:segstate.parent.p.ph_vheight
                                     is
                                     3)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                      is
                                      0)
                                     ((0.728345 1.19617))
                                     ((0.892911 0.861536)))
                                    ((R:segstate.parent.pos_in_syl
                                      is
                                      1)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                       is
                                       0)
                                      ((1.21005 0.860561))
                                      ((0.881925 0.568916)))
                                     ((0.893484 -0.130425))))
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                     <
                                     0.9)
                                    ((0.392831 -0.676257))
                                    ((R:segstate.parent.lisp_onset_stop
                                      is
                                      0)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                       <
                                       0.1)
                                      ((0.485734 -0.35207))
                                      ((1.15973 0.205151)))
                                     ((0.540551 -0.440272)))))
                                  ((n.name is f_1)
                                   ((R:segstate.parent.name is z)
                                    ((0.347175 0.823443))
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      3.4)
                                     ((R:segstate.parent.pp.ph_cplace
                                       is
                                       a)
                                      ((2.91801 1.12432))
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                        <
                                        0.3)
                                       ((0.508686 -0.324797))
                                       ((0.742 0.107126))))
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       4.6)
                                      ((0.552301 0.105448))
                                      ((0.469087 -0.416705)))))
                                   ((R:segstate.parent.n.ph_cvox is +)
                                    ((n.name is n_3)
                                     ((R:segstate.parent.n.name is l)
                                      ((0.374273 -1.25526))
                                      ((R:segstate.parent.nn.ph_vc
                                        is
                                        +)
                                       ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                         <
                                         2.4)
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                          <
                                          1.2)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                           <
                                           0.3)
                                          ((0.448759 0.522775))
                                          ((0.74988 0.110355)))
                                         ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                           <
                                           1.2)
                                          ((0.841338 1.16704))
                                          ((R:segstate.parent.n.name
                                            is
                                            dh)
                                           ((0.678686 0.802117))
                                           ((1.04878 0.357317)))))
                                        ((R:segstate.parent.pp.ph_cplace
                                          is
                                          0)
                                         ((0.53251 0.266385))
                                         ((0.798963 -0.00618829))))
                                       ((R:segstate.parent.pp.ph_cplace
                                         is
                                         a)
                                        ((0.567846 0.726048))
                                        ((0.830633 1.25078)))))
                                     ((n.name is ng_3)
                                      ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                                        <
                                        0.1)
                                       ((0.64649 0.112558))
                                       ((R:segstate.parent.nn.ph_vheight
                                         is
                                         2)
                                        ((0.903954 1.13026))
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                          <
                                          2.1)
                                         ((0.905502 0.89975))
                                         ((0.550614 0.378684)))))
                                      ((n.name is z_3)
                                       ((R:segstate.parent.n.ph_ctype
                                         is
                                         r)
                                        ((0.745202 -0.0554038))
                                        ((R:segstate.parent.R:SylStructure.parent.next_accent
                                          <
                                          1)
                                         ((R:segstate.parent.p.name
                                           is
                                           ih)
                                          ((2.31276 1.20622))
                                          ((0.499988 0.698065)))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                           <
                                           6.8)
                                          ((0.53895 0.605672))
                                          ((0.750019 -0.111577)))))
                                       ((R:segstate.parent.ph_cplace
                                         is
                                         b)
                                        ((n.name is v_2)
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                           <
                                           0.3)
                                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                                            <
                                            4.8)
                                           ((0.585339 -0.24524))
                                           ((0.509961 0.210232)))
                                          ((R:segstate.parent.p.name
                                            is
                                            ah)
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                             is
                                             0)
                                            ((0.773855 -0.152592))
                                            ((0.476004 -0.335251)))
                                           ((0.465142 -0.504429))))
                                         ((R:segstate.parent.pp.ph_cplace
                                           is
                                           0)
                                          ((R:segstate.parent.R:SylStructure.parent.syl_out
                                            <
                                            5.4)
                                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                                             <
                                             2.4)
                                            ((0.687844 0.052214))
                                            ((0.724682 0.618849)))
                                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                                             <
                                             1)
                                            ((0.808697 0.42583))
                                            ((1.31704 0.861263))))
                                          ((R:segstate.parent.n.name
                                            is
                                            m)
                                           ((0.429877 -0.385055))
                                           ((R:segstate.parent.n.ph_ctype
                                             is
                                             r)
                                            ((0.898989 0.439361))
                                            ((R:segstate.parent.nn.ph_vheight
                                              is
                                              2)
                                             ((R:segstate.parent.nn.ph_vfront
                                               is
                                               1)
                                              ((1.18017 0.419928))
                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                <
                                                0.3)
                                               ((0.687513 0.135548))
                                               ((0.570595 -0.151262))))
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                               is
                                               0)
                                              ((0.415146 -0.447831))
                                              ((0.813354 0.0310359))))))))
                                        ((R:segstate.parent.n.name
                                          is
                                          l)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           2)
                                          ((R:segstate.parent.pos_in_syl
                                            is
                                            0)
                                           ((0.410269 -0.272559))
                                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                                             <
                                             1.8)
                                            ((0.380712 -0.452837))
                                            ((0.0938496 -0.561184))))
                                          ((R:segstate.parent.ph_vheight
                                            is
                                            2)
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                             <
                                             0.3)
                                            ((0.645137 -0.126526))
                                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                                              <
                                              1.2)
                                             ((0.472886 -0.354908))
                                             ((0.312715 -0.566289))))
                                           ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                             <
                                             2.1)
                                            ((R:segstate.parent.p.ph_vrnd
                                              is
                                              +)
                                             ((0.681456 -0.170006))
                                             ((R:segstate.parent.lisp_coda_glide
                                               is
                                               0)
                                              ((R:segstate.parent.ph_cplace
                                                is
                                                a)
                                               ((1.04875 0.874855))
                                               ((0.933479 0.1493)))
                                              ((1.09841 -0.0164584))))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                              is
                                              0)
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                               is
                                               0)
                                              ((1.42505 1.10744))
                                              ((1.22016 0.540514)))
                                             ((1.29038 0.201367))))))
                                         ((R:segstate.parent.ph_vheight
                                           is
                                           2)
                                          ((R:segstate.parent.R:SylStructure.parent.last_accent
                                            <
                                            0.9)
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                             is
                                             0)
                                            ((0.8068 0.00863933))
                                            ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                              <
                                              2.5)
                                             ((0.474173 -0.187824))
                                             ((0.227444 -0.525087))))
                                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                                             <
                                             3.4)
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                              <
                                              1.2)
                                             ((1.59911 0.829594))
                                             ((1.02055 0.229419)))
                                            ((R:segstate.parent.nn.ph_cplace
                                              is
                                              0)
                                             ((R:segstate.parent.p.ph_vlng
                                               is
                                               l)
                                              ((0.42105 -0.0969908))
                                              ((0.622164 -0.244516)))
                                             ((1.30982 0.451986)))))
                                          ((R:segstate.parent.ph_vlng
                                            is
                                            l)
                                           ((R:segstate.parent.R:SylStructure.parent.position_type
                                             is
                                             initial)
                                            ((0.409403 -0.372958))
                                            ((n.name is aa_2)
                                             ((0.474469 -0.494133))
                                             ((n.name is aa_3)
                                              ((6.00697e-07 -0.83898))
                                              ((0.300837 -0.719229)))))
                                           ((n.name is dh_1)
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                              is
                                              0)
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                               <
                                               0.3)
                                              ((0.978125 0.124317))
                                              ((0.329466 -0.489549)))
                                             ((R:segstate.parent.name
                                               is
                                               m)
                                              ((0.440484 -0.240214))
                                              ((R:segstate.parent.ph_cplace
                                                is
                                                a)
                                               ((R:segstate.parent.name
                                                 is
                                                 z)
                                                ((0.220858 -0.62723))
                                                ((R:segstate.parent.p.ph_vlng
                                                  is
                                                  l)
                                                 ((R:segstate.parent.name
                                                   is
                                                   n)
                                                  ((0.175519 -0.472709))
                                                  ((0.205535 -0.575253)))
                                                 ((R:segstate.parent.pp.ph_cplace
                                                   is
                                                   0)
                                                  ((0.101805 -0.523467))
                                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                    is
                                                    0)
                                                   ((0.664407
                                                     -0.148216))
                                                   ((0.173468
                                                     -0.381678))))))
                                               ((0.347142 -0.769211)))))
                                            ((n.name is m_1)
                                             ((R:segstate.parent.name
                                               is
                                               z)
                                              ((0.0692963 -0.780624))
                                              ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                <
                                                1.6)
                                               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                 <
                                                 2.1)
                                                ((0.247089 -0.578788))
                                                ((0.342749 -0.372008)))
                                               ((0.527358 -0.206654))))
                                             ((R:segstate.parent.n.name
                                               is
                                               m)
                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                                is
                                                0)
                                               ((1.82928 0.844855))
                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                                 is
                                                 0)
                                                ((1.03887 0.0688807))
                                                ((0.457641 -0.402911))))
                                              ((R:segstate.parent.n.ph_cplace
                                                is
                                                a)
                                               ((n.name is r_2)
                                                ((0.511368 -0.514983))
                                                ((R:segstate.parent.p.name
                                                  is
                                                  ih)
                                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                                   is
                                                   0)
                                                  ((0.889812 0.103095))
                                                  ((1.44454 0.585509)))
                                                 ((R:segstate.parent.nn.ph_cplace
                                                   is
                                                   a)
                                                  ((0.776311 0.280685))
                                                  ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                    <
                                                    1.8)
                                                   ((R:segstate.parent.lisp_coda_nasal
                                                     is
                                                     0)
                                                    ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                      <
                                                      1.4)
                                                     ((0.746203
                                                       -0.206935))
                                                     ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                       <
                                                       1.5)
                                                      ((0.901265
                                                        0.599037))
                                                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                        <
                                                        2.3)
                                                       ((0.876386
                                                         0.000873783))
                                                       ((0.660993
                                                         0.0859357)))))
                                                    ((0.468256
                                                      -0.376187)))
                                                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                     <
                                                     7.5)
                                                    ((n.name is z_1)
                                                     ((0.299323
                                                       -0.415199))
                                                     ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                                       <
                                                       0.2)
                                                      ((0.812276
                                                        -0.190767))
                                                      ((0.452666
                                                        0.027039))))
                                                    ((0.315329
                                                      -0.612041)))))))
                                               ((R:segstate.parent.ph_cplace
                                                 is
                                                 l)
                                                ((R:segstate.parent.n.ph_ctype
                                                  is
                                                  f)
                                                 ((R:segstate.parent.lisp_onset_glide
                                                   is
                                                   0)
                                                  ((0.583623 -0.284839))
                                                  ((0.171179 -0.449406)))
                                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                   <
                                                   3.5)
                                                  ((n.name is m_2)
                                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                                     <
                                                     1.2)
                                                    ((0.897524
                                                      0.0449665))
                                                    ((0.54724
                                                      -0.212895)))
                                                   ((R:segstate.parent.pos_in_syl
                                                     is
                                                     2)
                                                    ((1.0568 0.586171))
                                                    ((1.03774
                                                      0.0133466))))
                                                  ((0.259814 -0.386524))))
                                                ((R:segstate.parent.pp.ph_vlng
                                                  is
                                                  d)
                                                 ((0.907747 0.143177))
                                                 ((n.name is b_1)
                                                  ((R:segstate.parent.p.ph_vheight
                                                    is
                                                    3)
                                                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                     <
                                                     0.4)
                                                    ((0.136368
                                                      -0.571541))
                                                    ((0.0285532
                                                      -0.669237)))
                                                   ((0.412332
                                                     -0.400844)))
                                                  ((R:segstate.parent.p.name
                                                    is
                                                    eh)
                                                   ((n.name is n_2)
                                                    ((0.248536
                                                      -0.153776))
                                                    ((R:segstate.parent.name
                                                      is
                                                      r)
                                                     ((n.name is r_2)
                                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                                        is
                                                        0)
                                                       ((0.448214
                                                         -0.368353))
                                                       ((0.199199
                                                         -0.596656)))
                                                      ((0.0836833
                                                        -0.635287)))
                                                     ((0.227611
                                                       -0.410425))))
                                                   ((R:segstate.parent.n.ph_cplace
                                                     is
                                                     v)
                                                    ((n.name is ng_1)
                                                     ((R:segstate.parent.p.ph_vrnd
                                                       is
                                                       +)
                                                      ((0.718085
                                                        0.0888254))
                                                      ((0.278276
                                                        -0.400702)))
                                                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                       <
                                                       1.9)
                                                      ((0.654811
                                                        -0.0904049))
                                                      ((R:segstate.parent.name
                                                        is
                                                        r)
                                                       ((0.707389
                                                         -0.226024))
                                                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                         <
                                                         2.4)
                                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                          <
                                                          5.4)
                                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                           <
                                                           2.3)
                                                          ((0.322835
                                                            -0.490488))
                                                          ((R:segstate.parent.p.ph_vrnd
                                                            is
                                                            +)
                                                           ((0.630799
                                                             -0.226807))
                                                           ((0.372871
                                                             -0.384377))))
                                                         ((0.184446
                                                           -0.670217)))
                                                        ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                          <
                                                          9.6)
                                                         ((0.261109
                                                           -0.591754))
                                                         ((0.226567
                                                           -0.74762)))))))
                                                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                      <
                                                      3)
                                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                       <
                                                       4.8)
                                                      ((n.name is w_1)
                                                       ((R:segstate.parent.ph_cplace
                                                         is
                                                         a)
                                                        ((0.344904
                                                          -0.373926))
                                                        ((0.220302
                                                          -0.632751)))
                                                       ((n.name is v_1)
                                                        ((0.0792737
                                                          -0.696309))
                                                        ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                                                          <
                                                          0.2)
                                                         ((n.name
                                                           is
                                                           r_2)
                                                          ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                            <
                                                            4.2)
                                                           ((0.461712
                                                             -0.339285))
                                                           ((0.35312
                                                             -0.493708)))
                                                          ((R:segstate.parent.n.name
                                                            is
                                                            jh)
                                                           ((0.219926
                                                             -0.408516))
                                                           ((R:segstate.parent.n.name
                                                             is
                                                             dh)
                                                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                              <
                                                              1)
                                                             ((0.402026
                                                               -0.125089))
                                                             ((0.445421
                                                               -0.400187)))
                                                            ((R:segstate.parent.ph_cplace
                                                              is
                                                              a)
                                                             ((0.522691
                                                               -0.270836))
                                                             ((R:segstate.parent.p.ph_vheight
                                                               is
                                                               1)
                                                              ((0.90639
                                                                0.196432))
                                                              ((1.07716
                                                                -0.103214)))))))
                                                         ((0.578113
                                                           0.0399362)))))
                                                      ((R:segstate.parent.ph_cvox
                                                        is
                                                        +)
                                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                                         <
                                                         0.1)
                                                        ((0.652922
                                                          0.0676091))
                                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                          <
                                                          0.5)
                                                         ((0.266747
                                                           -0.416836))
                                                         ((R:segstate.parent.p.ph_vlng
                                                           is
                                                           s)
                                                          ((0.760943
                                                            0.0405824))
                                                          ((0.461131
                                                            -0.320241)))))
                                                       ((1.31417
                                                         0.471721))))
                                                     ((R:segstate.parent.p.ph_vlng
                                                       is
                                                       s)
                                                      ((1.64958
                                                        0.474012))
                                                      ((R:segstate.parent.n.ph_cplace
                                                        is
                                                        p)
                                                       ((0.489242
                                                         -0.334652))
                                                       ((0.959538
                                                         0.039165)))))))))))))))))))))
                                    ((n.name is z_3)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                       <
                                       0.3)
                                      ((R:segstate.parent.nn.ph_ctype
                                        is
                                        n)
                                       ((0.682045 1.07979))
                                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                                         <
                                         3)
                                        ((0.761413 0.778907))
                                        ((0.679819 0.273641))))
                                      ((R:segstate.parent.n.name is hh)
                                       ((R:segstate.parent.p.ph_vlng
                                         is
                                         s)
                                        ((0.568913 -0.487628))
                                        ((0.270122 -0.698679)))
                                       ((R:segstate.parent.p.ph_vheight
                                         is
                                         1)
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                          <
                                          2.8)
                                         ((0.196445 -0.817158))
                                         ((0.222259 -0.968406)))
                                        ((R:segstate.parent.n.ph_ctype
                                          is
                                          f)
                                         ((R:segstate.parent.n.name
                                           is
                                           s)
                                          ((0.115489 -1.21329))
                                          ((0.170182 -1.10357)))
                                         ((0.271236 -0.941886))))))
                                     ((n.name is n_3)
                                      ((R:segstate.parent.n.name is k)
                                       ((0.823676 0.702412))
                                       ((R:segstate.parent.p.ph_vheight
                                         is
                                         1)
                                        ((R:segstate.parent.n.ph_cplace
                                          is
                                          a)
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                           <
                                           1.2)
                                          ((0.616681 -0.0659029))
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                            is
                                            0)
                                           ((0.720855 0.683502))
                                           ((0.498614 0.0821876))))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                           <
                                           0.2)
                                          ((R:segstate.parent.n.name
                                            is
                                            hh)
                                           ((0.363194 -0.918386))
                                           ((0.612681 -0.453756)))
                                          ((0.767575 0.0644294))))
                                        ((R:segstate.parent.lisp_coda_stop
                                          is
                                          0)
                                         ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                                           <
                                           0.1)
                                          ((R:segstate.parent.lisp_onset_stop
                                            is
                                            0)
                                           ((R:segstate.parent.p.ph_vheight
                                             is
                                             2)
                                            ((0.422725 -0.120699))
                                            ((0.481005 -0.796579)))
                                           ((0.43775 -0.891186)))
                                          ((R:segstate.parent.n.ph_cplace
                                            is
                                            a)
                                           ((0.664329 0.172952))
                                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                                             <
                                             5.8)
                                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                              <
                                              1.2)
                                             ((0.813965 -0.360957))
                                             ((0.490851 -0.819787)))
                                            ((0.7647 -0.280871)))))
                                         ((R:segstate.parent.nn.ph_ctype
                                           is
                                           s)
                                          ((0.61653 -0.641261))
                                          ((0.470156 -1.03203))))))
                                      ((n.name is v_3)
                                       ((R:segstate.parent.n.ph_vrnd
                                         is
                                         -)
                                        ((0.924929 0.312702))
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                          is
                                          0)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           6)
                                          ((0.171743 -0.865164))
                                          ((0.490568 -0.633509)))
                                         ((0.124034 -0.956935))))
                                       ((n.name is ng_3)
                                        ((R:segstate.parent.p.ph_vheight
                                          is
                                          3)
                                         ((0.56174 -0.960541))
                                         ((R:segstate.parent.R:SylStructure.parent.next_accent
                                           <
                                           0.7)
                                          ((R:segstate.parent.nn.ph_ctype
                                            is
                                            0)
                                           ((0.893429 0.247688))
                                           ((0.616871 -0.265808)))
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                            <
                                            0.3)
                                           ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                             <
                                             0.1)
                                            ((0.446357 -0.718176))
                                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                              <
                                              1.5)
                                             ((0.568115 -0.0225737))
                                             ((0.786664 -0.412522))))
                                           ((0.681301 -0.785098)))))
                                        ((R:segstate.parent.ph_cplace
                                          is
                                          l)
                                         ((n.name is t_1)
                                          ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                            <
                                            2.1)
                                           ((0.32708 -0.619749))
                                           ((0.202132 -1.01691)))
                                          ((R:segstate.parent.lisp_onset_glide
                                            is
                                            0)
                                           ((R:segstate.parent.n.ph_ctype
                                             is
                                             s)
                                            ((n.name is p_1)
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                               is
                                               0)
                                              ((0.179206 -0.50305))
                                              ((0.112511 -0.417968)))
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                               is
                                               0)
                                              ((0.861278 0.222004))
                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                <
                                                0.3)
                                               ((0.673287 0.251999))
                                               ((0.396952 -0.397874)))))
                                            ((R:segstate.parent.name
                                              is
                                              p)
                                             ((0.544118 -0.642342))
                                             ((R:segstate.parent.nn.ph_vfront
                                               is
                                               1)
                                              ((n.name is m_2)
                                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                 <
                                                 7)
                                                ((0.34305 -0.229937))
                                                ((0.168579 -0.424382)))
                                               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                 <
                                                 2.2)
                                                ((0.0598592 -0.50259))
                                                ((0.196647 -0.443228))))
                                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                <
                                                2.2)
                                               ((R:segstate.parent.p.ph_vheight
                                                 is
                                                 3)
                                                ((0.347848 -0.262558))
                                                ((0.711287 0.171653)))
                                               ((R:segstate.parent.nn.ph_cplace
                                                 is
                                                 a)
                                                ((0.153945 -0.442824))
                                                ((0.354434 -0.2831)))))))
                                           ((R:segstate.parent.lisp_onset_stop
                                             is
                                             0)
                                            ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                              <
                                              3.5)
                                             ((0.569203 -0.207184))
                                             ((0.780294 0.0792683)))
                                            ((1.03315 0.494077)))))
                                         ((n.name is t_1)
                                          ((R:segstate.parent.name
                                            is
                                            s)
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                             is
                                             0)
                                            ((1.59674 0.897175))
                                            ((R:segstate.parent.lisp_onset_glide
                                              is
                                              0)
                                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                               <
                                               2.1)
                                              ((0.85957 0.515096))
                                              ((0.756038 0.0279298)))
                                             ((0.684365 -0.123133))))
                                           ((R:segstate.parent.ph_cvox
                                             is
                                             -)
                                            ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                                              <
                                              0.1)
                                             ((0.899519 -0.058534))
                                             ((0.520087 -0.841515)))
                                            ((R:segstate.parent.ph_vlng
                                              is
                                              s)
                                             ((0.187333 -0.361044))
                                             ((R:segstate.parent.R:SylStructure.parent.position_type
                                               is
                                               single)
                                              ((R:segstate.parent.p.ph_vrnd
                                                is
                                                +)
                                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                 <
                                                 2.2)
                                                ((0.499608 -0.202745))
                                                ((0.654654 0.187459)))
                                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                 <
                                                 3.6)
                                                ((0.764487 0.621278))
                                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                                  <
                                                  0.3)
                                                 ((0.576907 0.376749))
                                                 ((0.652348 -0.0237435)))))
                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                                <
                                                1.2)
                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                                 is
                                                 0)
                                                ((R:segstate.parent.lisp_onset_glide
                                                  is
                                                  0)
                                                 ((R:segstate.parent.p.ph_vlng
                                                   is
                                                   l)
                                                  ((0.528462 0.0408753))
                                                  ((R:segstate.parent.nn.ph_vheight
                                                    is
                                                    2)
                                                   ((0.370273
                                                     -0.127036))
                                                   ((0.317958
                                                     -0.339765))))
                                                 ((0.204209 -0.492393)))
                                                ((0.391729 0.134696)))
                                               ((0.648274 0.506993)))))))
                                          ((R:segstate.parent.ph_vrnd
                                            is
                                            +)
                                           ((0.797574 0.615061))
                                           ((R:segstate.parent.name
                                             is
                                             s)
                                            ((R:segstate.parent.pp.ph_cplace
                                              is
                                              d)
                                             ((0.426428 -0.406371))
                                             ((0.166301 -0.722654)))
                                            ((n.name is s_1)
                                             ((R:segstate.parent.name
                                               is
                                               r)
                                              ((1.77343 1.17975))
                                              ((R:segstate.parent.name
                                                is
                                                z)
                                               ((0.032957 -0.814016))
                                               ((R:segstate.parent.name
                                                 is
                                                 n)
                                                ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                  <
                                                  8.5)
                                                 ((R:segstate.parent.p.ph_vheight
                                                   is
                                                   3)
                                                  ((0.563482 0.40946))
                                                  ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                    <
                                                    2.1)
                                                   ((0.342539
                                                     -0.122819))
                                                   ((0.506319 0.190725))))
                                                 ((0.61956 0.685836)))
                                                ((R:segstate.parent.nn.ph_ctype
                                                  is
                                                  0)
                                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                   <
                                                   2.1)
                                                  ((0.492049 0.14046))
                                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                                    is
                                                    0)
                                                   ((0.518518
                                                     -0.287896))
                                                   ((0.426394
                                                     -0.0158391))))
                                                 ((0.367444 -0.334566))))))
                                             ((R:segstate.parent.nn.ph_cplace
                                               is
                                               g)
                                              ((1.65516 0.759426))
                                              ((R:segstate.parent.p.name
                                                is
                                                ao)
                                               ((n.name is r_2)
                                                ((R:segstate.parent.lisp_onset_glide
                                                  is
                                                  0)
                                                 ((R:segstate.parent.pp.ph_cplace
                                                   is
                                                   b)
                                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                    <
                                                    2.1)
                                                   ((0.55538 -0.339285))
                                                   ((0.826971 0.401944)))
                                                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                    <
                                                    0.9)
                                                   ((0.788103 0.393708))
                                                   ((0.705456 0.854917))))
                                                 ((0.647262 -0.378042)))
                                                ((R:segstate.parent.n.ph_ctype
                                                  is
                                                  0)
                                                 ((R:segstate.parent.nn.ph_ctype
                                                   is
                                                   n)
                                                  ((0.347362 -0.52936))
                                                  ((0.680138 -0.223602)))
                                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                   <
                                                   3.5)
                                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                    <
                                                    0.3)
                                                   ((0.462713
                                                     -0.218625))
                                                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                     <
                                                     1.4)
                                                    ((1.2004 0.427854))
                                                    ((0.647513
                                                      -0.0661674))))
                                                  ((0.409032 -0.30604)))))
                                               ((n.name is r_2)
                                                ((R:segstate.parent.p.name
                                                  is
                                                  aa)
                                                 ((R:segstate.parent.n.seg_onsetcoda
                                                   is
                                                   coda)
                                                  ((0.801269 -0.11051))
                                                  ((0.302909 -0.486963)))
                                                 ((R:segstate.parent.n.ph_ctype
                                                   is
                                                   f)
                                                  ((0.0495455
                                                    -0.653032))
                                                  ((0.166482 -0.554685))))
                                                ((R:segstate.parent.p.name
                                                  is
                                                  er)
                                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                   <
                                                   0.9)
                                                  ((R:segstate.parent.nn.ph_vrnd
                                                    is
                                                    -)
                                                   ((1.11664 0.0524329))
                                                   ((R:segstate.parent.n.name
                                                     is
                                                     ae)
                                                    ((0.666209
                                                      -0.103326))
                                                    ((R:segstate.parent.nn.ph_cplace
                                                      is
                                                      a)
                                                     ((0.510849
                                                       -0.396089))
                                                     ((0.48008
                                                       -0.379163)))))
                                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                                    is
                                                    0)
                                                   ((0.512704
                                                     -0.752541))
                                                   ((0.70165 -0.428258))))
                                                 ((n.name is hh_1)
                                                  ((R:segstate.parent.ph_cplace
                                                    is
                                                    b)
                                                   ((0.591822 0.626205))
                                                   ((R:segstate.parent.lisp_coda_nasal
                                                     is
                                                     0)
                                                    ((R:segstate.parent.name
                                                      is
                                                      d)
                                                     ((0.575926
                                                       -0.156652))
                                                     ((R:segstate.parent.R:SylStructure.parent.position_type
                                                       is
                                                       single)
                                                      ((R:segstate.parent.name
                                                        is
                                                        z)
                                                       ((0.32125
                                                         0.236006))
                                                       ((0.996299
                                                         0.73127)))
                                                      ((0.514868
                                                        -0.211972))))
                                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                      is
                                                      0)
                                                     ((0.390566
                                                       -0.422729))
                                                     ((0.556064
                                                       0.0958864)))))
                                                  ((n.name is ax_2)
                                                   ((0.974644 0.356096))
                                                   ((n.name is g_3)
                                                    ((0.981568 0.33523))
                                                    ((R:segstate.parent.name
                                                      is
                                                      z)
                                                     ((R:segstate.parent.n.ph_ctype
                                                       is
                                                       s)
                                                      ((0.330549
                                                        0.198178))
                                                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                        <
                                                        1.6)
                                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                         <
                                                         0.3)
                                                        ((0.0534634
                                                          -0.787929))
                                                        ((0.459171
                                                          -0.639556)))
                                                       ((0.406268
                                                         -0.499527))))
                                                     ((n.name is sh_3)
                                                      ((0.849941
                                                        -0.63534))
                                                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                        <
                                                        2.1)
                                                       ((n.name
                                                         is
                                                         ao_1)
                                                        ((0.454615
                                                          -0.516157))
                                                        ((R:segstate.parent.n.ph_vrnd
                                                          is
                                                          +)
                                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                                           is
                                                           0)
                                                          ((2.08272
                                                            0.479765))
                                                          ((0.770537
                                                            -0.0478651)))
                                                         ((R:segstate.parent.pp.ph_cplace
                                                           is
                                                           v)
                                                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                            <
                                                            1.1)
                                                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                             <
                                                             0.1)
                                                            ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                                                              <
                                                              0.1)
                                                             ((0.330229
                                                               -0.38637))
                                                             ((0.852503
                                                               0.117374)))
                                                            ((0.540944
                                                              -0.415493)))
                                                           ((R:segstate.parent.name
                                                             is
                                                             n)
                                                            ((0.366575
                                                              -0.259217))
                                                            ((0.227671
                                                              -0.701188))))
                                                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                            <
                                                            1.5)
                                                           ((R:segstate.parent.n.ph_cplace
                                                             is
                                                             p)
                                                            ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                                              <
                                                              0.2)
                                                             ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                               <
                                                               0.5)
                                                              ((0.990492
                                                                0.0467243))
                                                              ((0.533078
                                                                -0.201082)))
                                                             ((0.633344
                                                               0.258174)))
                                                            ((R:segstate.parent.n.name
                                                              is
                                                              t)
                                                             ((R:segstate.parent.nn.ph_vheight
                                                               is
                                                               1)
                                                              ((0.90956
                                                                0.329707))
                                                              ((R:segstate.parent.p.ph_vheight
                                                                is
                                                                1)
                                                               ((0.243684
                                                                 -0.45239))
                                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                                                 <
                                                                 0.3)
                                                                ((0.451388
                                                                  -0.1463))
                                                                ((0.408745
                                                                  0.100162)))))
                                                             ((n.name
                                                               is
                                                               v_2)
                                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                                                <
                                                                0.3)
                                                               ((0.267489
                                                                 -0.56082))
                                                               ((0.523006
                                                                 -0.418076)))
                                                              ((R:segstate.parent.name
                                                                is
                                                                v)
                                                               ((0.948948
                                                                 0.548835))
                                                               ((R:segstate.parent.pp.ph_cplace
                                                                 is
                                                                 d)
                                                                ((0.620522
                                                                  -0.00631319))
                                                                ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                                  <
                                                                  6)
                                                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                                   <
                                                                   2.1)
                                                                  ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                                                    <
                                                                    2.4)
                                                                   ((R:segstate.parent.ph_vc
                                                                     is
                                                                     +)
                                                                    ((0.638805
                                                                      -0.567177))
                                                                    ((n.name
                                                                      is
                                                                      p_1)
                                                                     ((0.304046
                                                                       -0.519395))
                                                                     ((R:segstate.parent.p.name
                                                                       is
                                                                       iy)
                                                                      ((R:segstate.parent.nn.ph_cplace
                                                                        is
                                                                        a)
                                                                       ((R:segstate.parent.n.name
                                                                         is
                                                                         ae)
                                                                        ((0.50504
                                                                          -0.256062))
                                                                        ((0.474308
                                                                          -0.346694)))
                                                                       ((0.383648
                                                                         -0.447958)))
                                                                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                                        <
                                                                        10.8)
                                                                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                                         <
                                                                         3.6)
                                                                        ((R:segstate.parent.nn.ph_vc
                                                                          is
                                                                          +)
                                                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                                                           <
                                                                           1.2)
                                                                          ((0.308036
                                                                            -0.459845))
                                                                          ((0.416433
                                                                            -0.251384)))
                                                                         ((R:segstate.parent.ph_cvox
                                                                           is
                                                                           +)
                                                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                                                            <
                                                                            0.3)
                                                                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                                             <
                                                                             1.2)
                                                                            ((0.537994
                                                                              -0.197285))
                                                                            ((0.302328
                                                                              -0.571524)))
                                                                           ((0.454923
                                                                             -0.0969676)))
                                                                          ((R:segstate.parent.pp.ph_cplace
                                                                            is
                                                                            a)
                                                                           ((1.11254
                                                                             0.267648))
                                                                           ((R:segstate.parent.p.ph_vlng
                                                                             is
                                                                             d)
                                                                            ((0.507354
                                                                              -0.306863))
                                                                            ((0.58212
                                                                              -0.0251287))))))
                                                                        ((R:segstate.parent.lisp_onset_stop
                                                                          is
                                                                          0)
                                                                         ((0.663205
                                                                           0.055714))
                                                                         ((0.529561
                                                                           -0.0739872))))
                                                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                                                         <
                                                                         0.2)
                                                                        ((0.253781
                                                                          -0.45261))
                                                                        ((0.476807
                                                                          -0.230312)))))))
                                                                   ((0.661369
                                                                     0.0603372)))
                                                                  ((0.817676
                                                                    0.163554)))
                                                                 ((R:segstate.parent.pp.ph_cplace
                                                                   is
                                                                   l)
                                                                  ((0.345643
                                                                    -0.537153))
                                                                  ((R:segstate.parent.pos_in_syl
                                                                    is
                                                                    2)
                                                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                                     is
                                                                     0)
                                                                    ((0.483632
                                                                      -0.327233))
                                                                    ((0.557639
                                                                      0.050812)))
                                                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                                                     is
                                                                     0)
                                                                    ((R:segstate.parent.pos_in_syl
                                                                      is
                                                                      0)
                                                                     ((0.477044
                                                                       -0.469999))
                                                                     ((0.384744
                                                                       -0.186044)))
                                                                    ((0.280613
                                                                      -0.53613)))))))))))
                                                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                             <
                                                             3.3)
                                                            ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                              <
                                                              1.3)
                                                             ((R:segstate.parent.name
                                                               is
                                                               n)
                                                              ((0.568877
                                                                -0.162247))
                                                              ((0.58256
                                                                -0.43594)))
                                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                               is
                                                               0)
                                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                                                <
                                                                1.3)
                                                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                                 <
                                                                 2.1)
                                                                ((0.735226
                                                                  0.187089))
                                                                ((0.720923
                                                                  -0.0895162)))
                                                               ((0.427181
                                                                 -0.110525)))
                                                              ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                                <
                                                                5.9)
                                                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                                 <
                                                                 4.1)
                                                                ((R:segstate.parent.pp.ph_cplace
                                                                  is
                                                                  a)
                                                                 ((0.506156
                                                                   -0.0439599))
                                                                 ((0.320163
                                                                   -0.373432)))
                                                                ((0.393498
                                                                  -0.538638)))
                                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                                                 <
                                                                 0.3)
                                                                ((0.859494
                                                                  0.0734747))
                                                                ((R:segstate.parent.n.ph_ctype
                                                                  is
                                                                  f)
                                                                 ((0.51772
                                                                   -0.115954))
                                                                 ((0.327315
                                                                   -0.361614)))))))
                                                            ((R:segstate.parent.name
                                                              is
                                                              n)
                                                             ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                               <
                                                               0.2)
                                                              ((0.429634
                                                                -0.166383))
                                                              ((0.189392
                                                                -0.428946)))
                                                             ((R:segstate.parent.nn.ph_ctype
                                                               is
                                                               0)
                                                              ((0.848357
                                                                -0.0842429))
                                                              ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                                <
                                                                0.2)
                                                               ((1.73195
                                                                 0.720871))
                                                               ((0.845769
                                                                 0.133423))))))))))
                                                       ((R:segstate.parent.p.name
                                                         is
                                                         aw)
                                                        ((0.419624
                                                          -0.533043))
                                                        ((n.name
                                                          is
                                                          n_2)
                                                         ((R:segstate.parent.p.ph_vheight
                                                           is
                                                           3)
                                                          ((0.529172
                                                            0.0735497))
                                                          ((R:segstate.parent.p.name
                                                            is
                                                            ih)
                                                           ((0.191759
                                                             -0.392334))
                                                           ((0.320239
                                                             -0.219218))))
                                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                           <
                                                           0.3)
                                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                                            is
                                                            0)
                                                           ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                                             <
                                                             2.2)
                                                            ((0.378535
                                                              -0.360805))
                                                            ((0.75238
                                                              -0.110126)))
                                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                                             is
                                                             0)
                                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                              is
                                                              0)
                                                             ((0.78072
                                                               -0.0226144))
                                                             ((1.09527
                                                               0.4527)))
                                                            ((0.478002
                                                              -0.143996))))
                                                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                            <
                                                            11)
                                                           ((R:segstate.parent.pp.ph_cplace
                                                             is
                                                             a)
                                                            ((0.909373
                                                              0.159167))
                                                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                              <
                                                              1.2)
                                                             ((0.687548
                                                               0.101041))
                                                             ((0.457128
                                                               -0.355892))))
                                                           ((R:segstate.parent.pos_in_syl
                                                             is
                                                             2)
                                                            ((0.668114
                                                              0.11918))
                                                            ((1.32326
                                                              0.86115)))))))))))))))))))))))))))))))))))))))))))))
                 ((n.name is w_1)
                  ((1.84599 2.16699))
                  ((n.name is r_1)
                   ((R:segstate.parent.ph_cplace is a)
                    ((1.74686 2.50202))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                      <
                      0.1)
                     ((0.836786 -0.250215))
                     ((0.719623 0.927635))))
                   ((R:segstate.parent.name is n)
                    ((R:segstate.parent.n.name is ax)
                     ((n.name is n_2)
                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                        <
                        6.6)
                       ((0.314998 -0.360114))
                       ((0.120609 -0.470997)))
                      ((R:segstate.parent.R:SylStructure.parent.position_type
                        is
                        final)
                       ((0.448602 -0.608632))
                       ((0.504413 -1.17658))))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                       <
                       0.1)
                      ((n.name is n_2)
                       ((R:segstate.parent.nn.ph_ctype is f)
                        ((1.18883 0.415625))
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          final)
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           5.2)
                          ((0.152127 -0.47539))
                          ((0.287052 -0.369994)))
                         ((0.489413 -0.178963))))
                       ((R:segstate.parent.pp.ph_cplace is a)
                        ((n.name is n_3)
                         ((0.501845 -0.110688))
                         ((0.163139 -0.467854)))
                        ((R:segstate.parent.R:SylStructure.parent.last_accent
                          <
                          1.2)
                         ((n.name is n_3)
                          ((0.535362 -0.750343))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                            <
                            0.3)
                           ((0.0785844 -0.554598))
                           ((0.276174 -0.375001))))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                           is
                           0)
                          ((0.543168 -0.239143))
                          ((0.249292 -0.560413))))))
                      ((n.name is n_3)
                       ((0.605191 0.501583))
                       ((R:segstate.parent.lisp_coda_stop is 0)
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                          <
                          0.9)
                         ((0.119932 -0.438765))
                         ((0.101724 -0.514741)))
                        ((0.315 -0.289729))))))
                    ((n.name is z_3)
                     ((R:segstate.parent.R:SylStructure.parent.asyl_in
                       <
                       1.2)
                      ((0.715095 1.26214))
                      ((0.465342 0.823064)))
                     ((n.name is k_3)
                      ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                        <
                        2.2)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                         is
                         0)
                        ((R:segstate.parent.R:SylStructure.parent.next_accent
                          <
                          0.4)
                         ((0.268576 -0.896676))
                         ((0.55723 -0.573319)))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                          is
                          0)
                         ((0.402373 -0.40484))
                         ((0.324551 -0.599823))))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                         <
                         1.5)
                        ((0.559102 -0.0926994))
                        ((0.504584 -0.436185))))
                      ((R:segstate.parent.name is k)
                       ((R:segstate.parent.n.ph_ctype is r)
                        ((0.912466 -0.0306743))
                        ((n.name is k_2)
                         ((R:segstate.parent.p.ph_vlng is l)
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            4.3)
                           ((1.04499 -0.0337062))
                           ((0.46466 0.328876)))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            5)
                           ((R:segstate.parent.p.ph_vheight is 2)
                            ((0.7486 0.41754))
                            ((0.768582 0.987536)))
                           ((0.678516 0.244814))))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                           <
                           0.1)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                            is
                            0)
                           ((0.499159 0.852175))
                           ((0.695386 0.188501)))
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            1.4)
                           ((0.679666 0.702683))
                           ((0.651147 1.30845))))))
                       ((n.name is ih_1)
                        ((R:segstate.parent.nn.ph_cplace is v)
                         ((R:segstate.parent.lisp_onset_glide is 0)
                          ((R:segstate.parent.ph_cvox is +)
                           ((0.206043 -0.690909))
                           ((R:segstate.parent.ph_cplace is a)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                              <
                              0.3)
                             ((0.148371 -0.591021))
                             ((0.226061 -0.447412)))
                            ((0.645678 -0.17023))))
                          ((0.283266 -0.220627)))
                         ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                           <
                           2.2)
                          ((R:segstate.parent.ph_cplace is l)
                           ((0.451846 -0.361955))
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             5.2)
                            ((R:segstate.parent.p.ph_vlng is l)
                             ((1.10782 0.661285))
                             ((0.860695 -0.0153393)))
                            ((0.643167 -0.0891075))))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((0.514979 -0.232164))
                           ((0.359843 -0.483019)))))
                        ((n.name is v_2)
                         ((R:segstate.parent.p.name is eh)
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            2.4)
                           ((1.17286 0.721356))
                           ((1.11289 1.69486)))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                            <
                            0.1)
                           ((1.09714 0.850744))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             2.6)
                            ((0.612131 -0.34935))
                            ((R:segstate.parent.pp.ph_cplace is a)
                             ((1.28695 0.471858))
                             ((1.22563 -0.0577444))))))
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           final)
                          ((R:segstate.parent.pp.ph_cplace is b)
                           ((R:segstate.parent.nn.ph_cplace is l)
                            ((0.547762 1.26329))
                            ((n.name is r_2)
                             ((0.244316 -0.545181))
                             ((R:segstate.parent.p.name is eh)
                              ((2.019 1.13962))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                4.5)
                               ((R:segstate.parent.n.ph_vrnd is -)
                                ((0.572136 -0.26719))
                                ((0.952747 0.128823)))
                               ((1.20155 0.24218))))))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                             is
                             0)
                            ((R:segstate.parent.name is g)
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                               is
                               0)
                              ((0.472811 0.875181))
                              ((0.645537 0.582348)))
                             ((n.name is th_3)
                              ((0.437282 -0.474236))
                              ((R:segstate.parent.n.name is ih)
                               ((0.57431 -0.29176))
                               ((R:segstate.parent.p.ph_vheight is 3)
                                ((0.465935 -0.196087))
                                ((R:segstate.parent.lisp_coda_stop
                                  is
                                  0)
                                 ((n.name is iy_1)
                                  ((0.455499 -0.265891))
                                  ((R:segstate.parent.nn.ph_ctype is n)
                                   ((1.35336 1.06225))
                                   ((R:segstate.parent.nn.ph_ctype
                                     is
                                     r)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                      is
                                      0)
                                     ((0.651209 0.090556))
                                     ((0.838666 -0.043201)))
                                    ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                      <
                                      2.3)
                                     ((R:segstate.parent.p.ph_vlng
                                       is
                                       s)
                                      ((0.871991 0.680417))
                                      ((0.628525 0.218038)))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                       <
                                       0.3)
                                      ((0.744713 -0.15011))
                                      ((0.503984 0.3161)))))))
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   1.5)
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                    <
                                    4)
                                   ((0.583671 -0.0225504))
                                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                     <
                                     1.2)
                                    ((R:segstate.parent.n.name is ay)
                                     ((0.522819 0.139331))
                                     ((0.929703 -0.00164996)))
                                    ((0.648561 0.339817))))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                                    <
                                    6.5)
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     8.6)
                                    ((0.454008 -0.297516))
                                    ((0.983162 -0.0305746)))
                                   ((0.750408 0.0483609)))))))))
                            ((n.name is t_2)
                             ((R:segstate.parent.p.name is ih)
                              ((0.0964946 -0.741435))
                              ((R:segstate.parent.n.name is er)
                               ((0.120922 -0.711706))
                               ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                 <
                                 2.4)
                                ((0.539183 -0.32969))
                                ((R:segstate.parent.p.ph_vheight is 2)
                                 ((0.293245 -0.55315))
                                 ((0.151431 -0.710952))))))
                             ((n.name is v_3)
                              ((R:segstate.parent.p.ph_vlng is s)
                               ((0.923114 -0.258029))
                               ((1.2696 0.789306)))
                              ((n.name is s_2)
                               ((R:segstate.parent.p.ph_vheight is 1)
                                ((0.332527 -0.0313698))
                                ((0.661265 0.412623)))
                               ((R:segstate.parent.lisp_onset_glide
                                 is
                                 0)
                                ((R:segstate.parent.name is v)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                   <
                                   5.5)
                                  ((0.273166 -0.780871))
                                  ((0.664982 -0.418856)))
                                 ((n.name is s_3)
                                  ((R:segstate.parent.n.name is t)
                                   ((0.385305 -0.967082))
                                   ((0.416588 -0.248618)))
                                  ((R:segstate.parent.n.ph_ctype is 0)
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                     <
                                     2.4)
                                    ((R:segstate.parent.nn.ph_cplace
                                      is
                                      v)
                                     ((0.395861 -0.436195))
                                     ((R:segstate.parent.name is sh)
                                      ((0.61783 -0.420027))
                                      ((R:segstate.parent.lisp_onset_stop
                                        is
                                        0)
                                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                                         <
                                         7)
                                        ((0.677536 0.16827))
                                        ((0.672256 0.23499)))
                                       ((0.681719 -0.255401)))))
                                    ((R:segstate.parent.n.name is ax)
                                     ((R:segstate.parent.p.ph_vlng
                                       is
                                       d)
                                      ((R:segstate.parent.pp.ph_cplace
                                        is
                                        l)
                                       ((0.508241 -0.84061))
                                       ((0.611802 -0.560262)))
                                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                                        <
                                        7)
                                       ((0.512726 -0.527483))
                                       ((0.629871 -0.174612))))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                       is
                                       0)
                                      ((0.588749 -0.449315))
                                      ((R:segstate.parent.lisp_onset_stop
                                        is
                                        0)
                                       ((R:segstate.parent.pp.ph_cplace
                                         is
                                         a)
                                        ((0.729517 0.17504))
                                        ((0.5653 -0.177417)))
                                       ((0.452645 -0.43551))))))
                                   ((R:segstate.parent.ph_cplace is l)
                                    ((0.615663 -0.266994))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                      is
                                      0)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                       <
                                       0.2)
                                      ((0.506845 0.13087))
                                      ((0.316693 -0.279602)))
                                     ((R:segstate.parent.ph_cplace
                                       is
                                       a)
                                      ((0.611075 0.358636))
                                      ((0.508377 0.00596368))))))))
                                ((n.name is r_2)
                                 ((0.680671 -0.367685))
                                 ((R:segstate.parent.n.name is iy)
                                  ((0.811167 0.610424))
                                  ((R:segstate.parent.n.ph_ctype is r)
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     4.3)
                                    ((1.34368 0.454939))
                                    ((0.707956 0.0322866)))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                                     <
                                     2.4)
                                    ((0.597127 -0.0446444))
                                    ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                      <
                                      2.5)
                                     ((0.47595 -0.409023))
                                     ((0.582786 -0.0906518)))))))))))))
                          ((R:segstate.parent.name is t)
                           ((R:segstate.parent.R:SylStructure.parent.position_type
                             is
                             mid)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                              is
                              0)
                             ((R:segstate.parent.R:SylStructure.parent.syl_out
                               <
                               2.3)
                              ((0.687541 1.08976))
                              ((0.686618 0.341422)))
                             ((0.431144 -0.430737)))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                              <
                              1.8)
                             ((0.681245 1.4673))
                             ((0.740239 0.705984))))
                           ((n.name is v_3)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                              is
                              0)
                             ((1.00528 1.10337))
                             ((0.9856 0.434882)))
                            ((R:segstate.parent.ph_cplace is a)
                             ((n.name is l_3)
                              ((R:segstate.parent.p.ph_vheight is 3)
                               ((0.707473 -0.289792))
                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                 <
                                 7)
                                ((0.692197 0.807047))
                                ((0.668663 1.33606))))
                              ((R:segstate.parent.name is r)
                               ((n.name is iy_1)
                                ((2.12723 1.39731))
                                ((R:segstate.parent.p.ph_vlng is s)
                                 ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                   <
                                   3.2)
                                  ((0.657824 -0.376635))
                                  ((0.0850738 -0.646812)))
                                 ((R:segstate.parent.p.ph_vrnd is +)
                                  ((1.02969 0.352497))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                    is
                                    0)
                                   ((1.19422 1.16084))
                                   ((1.20735 0.716943))))))
                               ((R:segstate.parent.p.ph_vheight is 1)
                                ((n.name is s_3)
                                 ((R:segstate.parent.n.ph_ctype is s)
                                  ((0.519742 -0.27677))
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                    <
                                    4)
                                   ((0.702403 0.680433))
                                   ((0.747612 -0.0385468))))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                   <
                                   1.2)
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                    <
                                    6.3)
                                   ((n.name is s_2)
                                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                      <
                                      3.6)
                                     ((0.308551 0.0254189))
                                     ((0.296151 -0.222391)))
                                    ((R:segstate.parent.lisp_onset_stop
                                      is
                                      0)
                                     ((R:segstate.parent.p.name is iy)
                                      ((0.44163 -0.339309))
                                      ((0.283971 -0.582884)))
                                     ((0.541042 -0.143965))))
                                   ((0.967927 0.184496)))
                                  ((1.19552 0.369312))))
                                ((R:segstate.parent.lisp_onset_glide
                                  is
                                  0)
                                 ((R:segstate.parent.R:SylStructure.parent.position_type
                                   is
                                   single)
                                  ((R:segstate.parent.lisp_onset_stop
                                    is
                                    0)
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                     <
                                     0.7)
                                    ((0.753421 0.251963))
                                    ((1.40297 0.537759)))
                                   ((1.15401 0.951697)))
                                  ((n.name is s_2)
                                   ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                     <
                                     3.2)
                                    ((0.443114 0.456354))
                                    ((0.5752 0.725826)))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                     <
                                     1.1)
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      3.4)
                                     ((0.380103 -0.04333))
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       4.6)
                                      ((1.02818 0.363891))
                                      ((1.06008 0.194423))))
                                    ((0.723097 -0.414075)))))
                                 ((R:segstate.parent.n.ph_ctype is 0)
                                  ((0.400629 -0.324083))
                                  ((0.809656 0.163924)))))))
                             ((n.name is w_3)
                              ((R:segstate.parent.p.name is iy)
                               ((R:segstate.parent.n.name is aa)
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                  <
                                  3.7)
                                 ((0.594223 -0.442137))
                                 ((0.527352 -0.736945)))
                                ((0.673971 0.263761)))
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 5)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                  is
                                  0)
                                 ((0.504044 1.00561))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                   is
                                   0)
                                  ((1.05084 0.575427))
                                  ((0.900904 0.19503))))
                                ((0.941158 1.22707))))
                              ((n.name is jh_2)
                               ((0.919004 0.971113))
                               ((n.name is hh_2)
                                ((R:segstate.parent.p.ph_vlng is l)
                                 ((R:segstate.parent.n.name is ae)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    9.5)
                                   ((1.18349e-06 -0.589325))
                                   ((0.0446018 -0.577808)))
                                  ((0.217863 -0.489316)))
                                 ((R:segstate.parent.n.name is ae)
                                  ((0.152887 -0.481361))
                                  ((1.74459 0.238395))))
                                ((n.name is y_2)
                                 ((0.100732 -0.614492))
                                 ((n.name is p_3)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                    is
                                    0)
                                   ((0.520253 -0.665061))
                                   ((0.665747 -0.246818)))
                                  ((R:segstate.parent.name is p)
                                   ((0.877738 0.589137))
                                   ((n.name is er_1)
                                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                                      <
                                      0.8)
                                     ((0.220817 -0.762431))
                                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                                       <
                                       3.6)
                                      ((0.852962 -0.175846))
                                      ((1.02037 0.0482905))))
                                    ((n.name is aa_1)
                                     ((R:segstate.parent.nn.ph_ctype
                                       is
                                       f)
                                      ((R:segstate.parent.p.name is iy)
                                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                                         <
                                         1.8)
                                        ((0.541869 -0.205799))
                                        ((0.332827 -0.468536)))
                                       ((0.464284 -0.607659)))
                                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                                        <
                                        0.7)
                                       ((0.798154 0.481789))
                                       ((0.614159 -0.0738729))))
                                     ((n.name is hh_3)
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                        is
                                        0)
                                       ((R:segstate.parent.p.ph_vlng
                                         is
                                         d)
                                        ((0.787093 -0.224118))
                                        ((R:segstate.parent.R:SylStructure.parent.next_accent
                                          <
                                          1.1)
                                         ((2.62374e-06 -0.537715))
                                         ((0.208296 -0.463928))))
                                       ((1.25874 0.214919)))
                                      ((n.name is m_3)
                                       ((R:segstate.parent.p.name
                                         is
                                         er)
                                        ((0.569371 -0.932502))
                                        ((R:segstate.parent.p.name
                                          is
                                          iy)
                                         ((0.412966 0.410384))
                                         ((R:segstate.parent.R:SylStructure.parent.next_accent
                                           <
                                           0.7)
                                          ((0.875961 -0.403341))
                                          ((R:segstate.parent.R:SylStructure.parent.last_accent
                                            <
                                            1.2)
                                           ((0.703425 -0.156381))
                                           ((0.831943 0.27098))))))
                                       ((R:segstate.parent.p.name
                                         is
                                         uw)
                                        ((R:segstate.parent.nn.ph_ctype
                                          is
                                          n)
                                         ((0.424314 -0.34554))
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                           is
                                           0)
                                          ((0.565384 0.141586))
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            a)
                                           ((0.673207 -0.0720251))
                                           ((0.5625 -0.366607)))))
                                        ((R:segstate.parent.name is jh)
                                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                           <
                                           3.3)
                                          ((0.773367 -0.150788))
                                          ((0.5351 -0.374004)))
                                         ((n.name is ah_1)
                                          ((0.787268 0.658355))
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            p)
                                           ((0.44266 -0.2954))
                                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                                             <
                                             6.4)
                                            ((n.name is m_2)
                                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                               <
                                               1.2)
                                              ((1.5229 1.04956))
                                              ((0.97756 0.281242)))
                                             ((R:segstate.parent.nn.ph_ctype
                                               is
                                               f)
                                              ((R:segstate.parent.pp.ph_cplace
                                                is
                                                g)
                                               ((1.21255 0.907816))
                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                                 <
                                                 0.3)
                                                ((n.name is w_2)
                                                 ((0.435404 -0.324929))
                                                 ((0.776429 0.0694394)))
                                                ((1.01646 0.53114))))
                                              ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                                <
                                                2.4)
                                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                                 <
                                                 2.4)
                                                ((R:segstate.parent.R:SylStructure.parent.position_type
                                                  is
                                                  single)
                                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                   <
                                                   6.6)
                                                  ((R:segstate.parent.lisp_coda_glide
                                                    is
                                                    0)
                                                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                     <
                                                     0.6)
                                                    ((0.756193
                                                      0.0996022))
                                                    ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                                      <
                                                      1.3)
                                                     ((1.46541 0.69244))
                                                     ((0.972493
                                                       0.946159))))
                                                   ((0.840505
                                                     -0.00433662)))
                                                  ((R:segstate.parent.p.name
                                                    is
                                                    iy)
                                                   ((0.798182 0.163373))
                                                   ((1.11177 -0.205119))))
                                                 ((R:segstate.parent.name
                                                   is
                                                   sh)
                                                  ((0.872498 -0.358427))
                                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                                    <
                                                    1.2)
                                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                                     is
                                                     0)
                                                    ((R:segstate.parent.name
                                                      is
                                                      f)
                                                     ((0.613926
                                                       0.118418))
                                                     ((0.683822
                                                       -0.258679)))
                                                    ((R:segstate.parent.p.ph_vheight
                                                      is
                                                      3)
                                                     ((0.964999
                                                       -0.114958))
                                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                       <
                                                       5.2)
                                                      ((1.24883
                                                        0.283731))
                                                      ((0.754469
                                                        0.17607)))))
                                                   ((0.719344 0.228892)))))
                                                ((R:segstate.parent.p.ph_vheight
                                                  is
                                                  2)
                                                 ((0.375771 -0.366596))
                                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                   <
                                                   4.2)
                                                  ((0.812693 -0.195916))
                                                  ((0.592248 0.139461)))))
                                               ((1.26539 0.606105)))))
                                            ((R:segstate.parent.n.name
                                              is
                                              ey)
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                               <
                                               0.3)
                                              ((0.40127 -0.549694))
                                              ((0.544814 -0.147292)))
                                             ((R:segstate.parent.p.ph_vlng
                                               is
                                               d)
                                              ((R:segstate.parent.name
                                                is
                                                w)
                                               ((0.535137 -0.387014))
                                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                 <
                                                 5.5)
                                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                  <
                                                  0.7)
                                                 ((1.15719 0.38347))
                                                 ((0.666767 0.127406)))
                                                ((R:segstate.parent.p.ph_vrnd
                                                  is
                                                  +)
                                                 ((1.0507 0.0976256))
                                                 ((0.539937 -0.35094)))))
                                              ((n.name is w_2)
                                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                                 <
                                                 10.6)
                                                ((0.897647 0.292134))
                                                ((1.62269 1.26831)))
                                               ((R:segstate.parent.p.ph_vheight
                                                 is
                                                 1)
                                                ((R:segstate.parent.name
                                                  is
                                                  m)
                                                 ((0.296875 -0.335249))
                                                 ((n.name is ae_1)
                                                  ((0.567522 -0.236581))
                                                  ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                                    <
                                                    0.3)
                                                   ((1.01834 0.347011))
                                                   ((0.712654
                                                     -0.105318)))))
                                                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                  <
                                                  1.2)
                                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                                   <
                                                   0.1)
                                                  ((1.01095 0.301033))
                                                  ((0.507122 -0.127669)))
                                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                   <
                                                   0.1)
                                                  ((1.3049 0.349036))
                                                  ((1.55132 0.502465))))))))))))))))))))))))))))))))))))))))))))))))
    ((n.name is t_3)
     ((R:segstate.parent.lisp_coda_stop is 0)
      ((R:segstate.parent.n.name is uw)
       ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 2.4)
        ((1.07743 2.91109))
        ((1.43199 1.26365)))
       ((R:segstate.parent.R:SylStructure.parent.position_type
         is
         single)
        ((R:segstate.parent.nn.ph_ctype is 0)
         ((R:segstate.parent.p.name is s)
          ((0.607251 0.777143))
          ((1.56664 1.90563)))
         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
           <
           0.3)
          ((R:segstate.parent.R:SylStructure.parent.syl_in < 8)
           ((R:segstate.parent.pp.ph_vc is -)
            ((0.563542 0.803278))
            ((R:segstate.parent.pp.ph_vlng is s)
             ((0.428729 0.990804))
             ((0.755566 1.69498))))
           ((0.894523 0.692995)))
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
            is
            0)
           ((R:segstate.parent.pp.ph_vc is +)
            ((0.644456 0.806583))
            ((0.501726 0.345611)))
           ((0.687287 0.900787)))))
        ((R:segstate.parent.R:SylStructure.parent.asyl_out < 4.2)
         ((R:segstate.parent.R:SylStructure.parent.asyl_in < 0.7)
          ((R:segstate.parent.p.ph_ctype is f)
           ((0.697049 0.37089))
           ((0.948535 1.0229)))
          ((R:segstate.parent.R:SylStructure.parent.position_type
            is
            final)
           ((R:segstate.parent.pp.ph_vheight is 3)
            ((0.553884 -0.328471))
            ((R:segstate.parent.nn.ph_cplace is a)
             ((0.85229 0.461167))
             ((0.606793 -0.0203181))))
           ((R:segstate.parent.R:SylStructure.parent.syl_out < 2.3)
            ((R:segstate.parent.p.name is s)
             ((0.373697 0.0117779))
             ((0.896318 0.953124)))
            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
              <
              0.1)
             ((0.566101 0.418685))
             ((R:segstate.parent.pp.ph_vfront is 1)
              ((0.696181 -0.0955523))
              ((0.604514 0.196352)))))))
         ((1.0132 0.996153)))))
      ((R:segstate.parent.seg_onsetcoda is coda)
       ((R:segstate.parent.p.ph_cvox is +)
        ((R:segstate.parent.R:SylStructure.parent.last_accent < 0.5)
         ((R:segstate.parent.pp.ph_vlng is a)
          ((0.136719 -0.665949))
          ((R:segstate.parent.R:SylStructure.parent.asyl_out < 2.5)
           ((0.674819 -0.501805))
           ((0.804835 -0.261062))))
         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
           is
           0)
          ((2.63916e-06 -0.742549))
          ((0.199986 -0.653854))))
        ((R:segstate.parent.p.ph_ctype is s)
         ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 1.8)
          ((0.564349 -0.277111))
          ((0.254299 -0.549954)))
         ((R:segstate.parent.p.name is s)
          ((R:segstate.parent.R:SylStructure.parent.syl_in < 8)
           ((R:segstate.parent.R:SylStructure.parent.asyl_out < 3.6)
            ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 2.6)
             ((0.412074 -0.147025))
             ((R:segstate.parent.n.ph_ctype is 0)
              ((0.613581 0.295656))
              ((0.530343 -0.0604419))))
            ((0.593327 -0.261062)))
           ((0.416317 -0.444483)))
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
            <
            0.3)
           ((0.585942 -0.00426897))
           ((1.38436 0.477217))))))
       ((R:segstate.parent.nn.ph_cplace is v)
        ((0.436672 0.614369))
        ((R:segstate.parent.R:SylStructure.parent.next_accent < 2.1)
         ((R:segstate.parent.R:SylStructure.parent.syl_in < 5.6)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
            <
            0.3)
           ((0.69218 0.493269))
           ((0.412895 -0.00764871)))
          ((0.550082 -0.149289)))
         ((0.38633 0.434419))))))
     ((R:segstate.parent.ph_cvox is +)
      ((R:segstate.parent.p.ph_ctype is n)
       ((n.name is w_2)
        ((R:segstate.parent.pp.ph_vheight is 2)
         ((1.0594 1.34281))
         ((R:segstate.parent.pp.ph_vlng is s)
          ((1.37104 0.218754))
          ((0.976786 0.165385))))
        ((n.name is y_3)
         ((0.575918 0.315498))
         ((R:segstate.parent.R:SylStructure.parent.position_type
           is
           initial)
          ((R:segstate.parent.R:SylStructure.parent.next_accent < 3.5)
           ((R:segstate.parent.R:SylStructure.parent.syl_out < 7.5)
            ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.6)
             ((R:segstate.parent.ph_cplace is a)
              ((0.444173 -0.274964))
              ((0.531955 -0.497847)))
             ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 6.6)
              ((0.905764 0.0560689))
              ((0.644657 -0.319585))))
            ((1.01646 0.29922)))
           ((0.524819 -0.499908)))
          ((n.name is z_2)
           ((R:segstate.parent.p.name is n)
            ((0.193905 -1.05362))
            ((0.453403 -0.724144)))
           ((n.name is d_3)
            ((R:segstate.parent.p.name is n)
             ((R:segstate.parent.n.ph_ctype is s)
              ((R:segstate.parent.n.ph_cplace is l)
               ((0.153812 -0.661915))
               ((R:segstate.parent.R:SylStructure.parent.syl_out < 4.4)
                ((0.460973 -0.338461))
                ((0.330544 -0.498758))))
              ((R:segstate.parent.nn.ph_vrnd is -)
               ((R:segstate.parent.R:SylStructure.parent.next_accent
                 <
                 0.9)
                ((R:segstate.parent.n.ph_cplace is a)
                 ((0.102658 -0.681593))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((0.102659 -0.681595))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    3.2)
                   ((0.155205 -0.578906))
                   ((0.190086 -0.598944)))))
                ((R:segstate.parent.R:SylStructure.parent.syl_out
                  <
                  7.1)
                 ((R:segstate.parent.pp.ph_vfront is 1)
                  ((0.11648 -0.666893))
                  ((4.96019e-06 -0.719163)))
                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                   <
                   10.2)
                  ((1.78814e-06 -0.719163))
                  ((2.3754e-06 -0.719162)))))
               ((R:segstate.parent.R:SylStructure.parent.last_accent
                 <
                 2.4)
                ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 6)
                 ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                   <
                   2.2)
                  ((0.319963 -0.493749))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    5)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                     <
                     0.3)
                    ((0.204545 -0.538831))
                    ((R:segstate.parent.R:SylStructure.parent.asyl_in
                      <
                      1.5)
                     ((0.090404 -0.690539))
                     ((0.147078 -0.613969))))
                   ((0.0655856 -0.70485))))
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                   <
                   1.2)
                  ((1.73475e-06 -0.719163))
                  ((0.105754 -0.679089))))
                ((0.310068 -0.481883)))))
             ((0.868216 -0.163141)))
            ((n.name is jh_2)
             ((0.305602 -1.00752))
             ((n.name is m_3)
              ((0.461342 -1.02603))
              ((R:segstate.parent.n.name is ax)
               ((R:segstate.parent.nn.ph_ctype is l)
                ((0.226041 -0.756551))
                ((R:segstate.parent.name is dh)
                 ((n.name is ax_1)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                    <
                    0.3)
                   ((0.522099 -0.292832))
                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                     <
                     5.5)
                    ((0.0873217 -0.573671))
                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                      <
                      9)
                     ((0.209391 -0.450888))
                     ((0.104231 -0.550902)))))
                  ((R:segstate.parent.pp.ph_vfront is 2)
                   ((0.424462 -0.487381))
                   ((R:segstate.parent.R:SylStructure.parent.asyl_out
                     <
                     3.5)
                    ((n.name is dh_2)
                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                       <
                       1.2)
                      ((0.0856983 -0.590652))
                      ((2.64793e-06 -0.639091)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                       is
                       0)
                      ((7.4871e-06 -0.684526))
                      ((8.67234e-06 -0.684532))))
                    ((0.191837 -0.567153)))))
                 ((n.name is ax_1)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    3)
                   ((0.115539 -0.676561))
                   ((0.440008 -0.431057)))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                    is
                    0)
                   ((0.841669 -0.0518326))
                   ((0.345696 -0.44823))))))
               ((R:segstate.parent.name is g)
                ((R:segstate.parent.p.name is ng)
                 ((n.name is g_3)
                  ((0.325177 -1.00451))
                  ((0.2251 -0.730724)))
                 ((0.590512 -0.370293)))
                ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                  <
                  0.3)
                 ((n.name is d_2)
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    10.6)
                   ((R:segstate.parent.n.ph_cplace is a)
                    ((R:segstate.parent.R:SylStructure.parent.asyl_in
                      <
                      2.4)
                     ((0.239786 -0.259031))
                     ((0.227755 -0.409033)))
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      2.8)
                     ((0.308482 -0.344633))
                     ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                       <
                       0.2)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        5)
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                         <
                         4.5)
                        ((0.275695 -0.386402))
                        ((0.170314 -0.457229)))
                       ((0.117065 -0.527183)))
                      ((0.128999 -0.51569)))))
                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                     <
                     0.8)
                    ((0.745244 -0.0847342))
                    ((1.65703 0.0359349))))
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    7.2)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                     <
                     2.4)
                    ((R:segstate.parent.n.ph_ctype is f)
                     ((R:segstate.parent.nn.ph_vfront is 1)
                      ((0.126372 -0.591017))
                      ((0.0518702 -0.654893)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                       is
                       0)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                        <
                        5.5)
                       ((R:segstate.parent.n.ph_cvox is +)
                        ((0.136299 -0.571416))
                        ((0.246849 -0.459683)))
                       ((0.162339 -0.601068)))
                      ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                        <
                        0.1)
                       ((0.12786 -0.590038))
                       ((0.0954096 -0.62961)))))
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      6.5)
                     ((1.01192 -0.23622))
                     ((0.0954095 -0.62961))))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((0.446389 -0.363708))
                    ((0.203287 -0.543079)))))
                 ((n.name is l_2)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                    is
                    0)
                   ((0.317382 -0.69653))
                   ((0.207506 -0.727743)))
                  ((R:segstate.parent.n.name is ae)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((0.236055 -0.541295))
                    ((0.157483 -0.702639)))
                   ((n.name is dh_2)
                    ((0.12625 -0.59438))
                    ((R:segstate.parent.lisp_coda_glide is 0)
                     ((n.name is w_3)
                      ((R:segstate.parent.n.name is ih)
                       ((0.303584 -0.751805))
                       ((0.708236 -0.45759)))
                      ((n.name is d_2)
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                         <
                         3.9)
                        ((R:segstate.parent.R:SylStructure.parent.syl_in
                          <
                          4.5)
                         ((R:segstate.parent.nn.ph_vfront is 2)
                          ((0.0830495 -0.548353))
                          ((R:segstate.parent.R:SylStructure.parent.asyl_out
                            <
                            2.4)
                           ((0.422084 -0.285846))
                           ((0.190303 -0.473552))))
                         ((0.5597 -0.170924)))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                          is
                          0)
                         ((0.202946 -0.495005))
                         ((7.02996e-06 -0.567407))))
                       ((R:segstate.parent.lisp_onset_glide is 0)
                        ((R:segstate.parent.n.ph_ctype is f)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                           is
                           0)
                          ((0.58695 -0.32092))
                          ((0.175249 -0.595395)))
                         ((n.name is z_3)
                          ((0.745838 0.0236057))
                          ((R:segstate.parent.n.name is iy)
                           ((0.114369 -0.632151))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             3.3)
                            ((R:segstate.parent.name is b)
                             ((0.444574 -0.443172))
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               7)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                                <
                                0.3)
                               ((0.584623 -0.287652))
                               ((R:segstate.parent.pp.ph_vfront is 2)
                                ((0.669729 -0.248337))
                                ((R:segstate.parent.pp.ph_vheight is 1)
                                 ((0.751895 0.0265872))
                                 ((1.40652 0.409165)))))
                              ((0.388502 -0.375598))))
                            ((R:segstate.parent.R:SylStructure.parent.next_accent
                              <
                              3)
                             ((R:segstate.parent.lisp_coda_nasal is 0)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                is
                                0)
                               ((0.114627 -0.488659))
                               ((0.487309 -0.531798)))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                is
                                0)
                               ((0.501049 -0.340761))
                               ((0.964378 0.0916853))))
                             ((0.400487 -0.580781)))))))
                        ((R:segstate.parent.ph_cplace is p)
                         ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                           <
                           1.2)
                          ((0.244441 -0.527161))
                          ((0.687633 -0.297269)))
                         ((R:segstate.parent.syl_final is 0)
                          ((R:segstate.parent.n.ph_vrnd is -)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                             is
                             0)
                            ((R:segstate.parent.p.name is n)
                             ((0.575906 -0.353527))
                             ((0.598329 -0.0456305)))
                            ((R:segstate.parent.pp.ph_vlng is d)
                             ((0.629497 -0.271938))
                             ((R:segstate.parent.nn.ph_ctype is f)
                              ((0.493397 0.0699046))
                              ((0.864335 0.362967)))))
                           ((1.17153 0.296634)))
                          ((0.486639 -0.352083)))))))
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       4.9)
                      ((0.507326 -0.396419))
                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                        <
                        1.2)
                       ((1.71497 0.660581))
                       ((0.73894 0.0220912))))))))))))))))))
       ((n.name is r_1)
        ((R:segstate.parent.name is b)
         ((0.634936 0.502466))
         ((R:segstate.parent.nn.ph_vheight is 2)
          ((1.58085 1.85724))
          ((1.04753 1.1379))))
        ((n.name is dh_3)
         ((R:segstate.parent.R:SylStructure.parent.syl_in < 1.6)
          ((R:segstate.parent.p.name is d)
           ((0.401321 -0.537657))
           ((R:segstate.parent.R:SylStructure.parent.syl_codasize
             <
             1.2)
            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
              is
              0)
             ((R:segstate.parent.R:SylStructure.parent.asyl_out < 3.4)
              ((0.817932 0.219281))
              ((1.01728 0.51302)))
             ((R:segstate.parent.nn.ph_cplace is l)
              ((0.81481 0.284808))
              ((R:segstate.parent.R:SylStructure.parent.asyl_out < 5.2)
               ((R:segstate.parent.R:SylStructure.parent.asyl_out < 3)
                ((0.915198 0.9233))
                ((1.46335 1.50721)))
               ((1.06489 0.490424)))))
            ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 7.6)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
               <
               0.3)
              ((0.644247 0.229324))
              ((R:segstate.parent.R:SylStructure.parent.next_accent
                <
                0.7)
               ((0.7649 0.412094))
               ((0.734344 0.67401))))
             ((0.629276 -0.0383033)))))
          ((R:segstate.parent.nn.ph_ctype is f)
           ((R:segstate.parent.nn.ph_cplace is a)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              1.6)
             ((0.881212 0.123253))
             ((0.999526 0.892381)))
            ((R:segstate.parent.nn.ph_cplace is b)
             ((1.05022 0.233402))
             ((0.403005 -0.511742))))
           ((R:segstate.parent.p.name is z)
            ((1.22385 0.425151))
            ((R:segstate.parent.p.ph_cvox is +)
             ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 1.8)
              ((0.999868 -0.170487))
              ((R:segstate.parent.pp.ph_vheight is 2)
               ((0.113059 -0.662766))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                 is
                 0)
                ((0.189977 -0.619249))
                ((0.460475 -0.358151)))))
             ((R:segstate.parent.nn.ph_ctype is r)
              ((0.973749 0.314181))
              ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 4.5)
               ((R:segstate.parent.pp.ph_cplace is a)
                ((0.489845 -0.371206))
                ((0.834947 0.14229)))
               ((0.394089 -0.478913))))))))
         ((n.name is l_2)
          ((R:segstate.parent.R:SylStructure.parent.asyl_in < 2.5)
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
             <
             0.3)
            ((R:segstate.parent.nn.ph_ctype is 0)
             ((0.469513 -0.559192))
             ((0.678322 -0.289892)))
            ((R:segstate.parent.pp.ph_vc is +)
             ((R:segstate.parent.R:SylStructure.parent.syl_out < 7.8)
              ((R:segstate.parent.R:SylStructure.parent.asyl_in < 0.2)
               ((0.513728 -0.32115))
               ((0.329262 -0.552169)))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                is
                0)
               ((0.337506 -0.587284))
               ((0.132396 -0.748399))))
             ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 2)
              ((0.0869941 -0.790169))
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 2.4)
               ((0.134286 -0.732832))
               ((R:segstate.parent.nn.ph_cplace is a)
                ((0.103439 -0.71994))
                ((0.251043 -0.581431)))))))
           ((R:segstate.parent.nn.ph_cplace is a)
            ((0.0796595 -0.797973))
            ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
              <
              1.1)
             ((0.218338 -0.659977))
             ((0.0799597 -0.757005)))))
          ((R:segstate.parent.p.name is r)
           ((n.name is m_2)
            ((0.701939 1.72958))
            ((R:segstate.parent.name is m)
             ((n.name is m_3)
              ((0.362784 -1.21966))
              ((0.125572 -0.471425)))
             ((n.name is w_2)
              ((0.0699821 -0.722802))
              ((R:segstate.parent.name is n)
               ((0.771609 -0.309942))
               ((n.name is dh_2)
                ((0.61717 -0.374891))
                ((R:segstate.parent.name is dh)
                 ((1.356 1.54618))
                 ((n.name is d_2)
                  ((0.773371 0.61847))
                  ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                    <
                    1.2)
                   ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                     <
                     3.5)
                    ((R:segstate.parent.n.ph_ctype is r)
                     ((0.573708 -0.195839))
                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                       <
                       1.6)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                        <
                        0.3)
                       ((R:segstate.parent.nn.ph_cplace is a)
                        ((0.645385 0.04909))
                        ((0.506887 -0.221281)))
                       ((n.name is w_3)
                        ((0.441203 0.418234))
                        ((R:segstate.parent.nn.ph_cplace is a)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                           <
                           0.1)
                          ((0.486487 0.320989))
                          ((0.759821 -0.110204)))
                         ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                           <
                           1.3)
                          ((1.05846 0.0176891))
                          ((R:segstate.parent.nn.ph_ctype is 0)
                           ((0.698436 0.307733))
                           ((0.861759 0.527535)))))))
                      ((R:segstate.parent.name is d)
                       ((0.811938 0.746615))
                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                         <
                         3.2)
                        ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                          <
                          1.3)
                         ((R:segstate.parent.lisp_onset_glide is 0)
                          ((0.943342 0.148862))
                          ((0.477316 0.315524)))
                         ((0.943138 0.463593)))
                        ((0.624488 0.0666459))))))
                    ((0.990494 -0.346676)))
                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                     <
                     4.9)
                    ((0.534754 -0.305888))
                    ((0.648563 -0.0186144)))))))))))
           ((n.name is n_3)
            ((R:segstate.parent.p.ph_ctype is s)
             ((R:segstate.parent.n.name is aa)
              ((0.671352 0.475501))
              ((R:segstate.parent.pp.ph_vrnd is -)
               ((0.858724 0.877574))
               ((1.74769 1.55831))))
             ((R:segstate.parent.n.name is ow)
              ((0.867071 0.267498))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                is
                0)
               ((R:segstate.parent.nn.ph_cplace is a)
                ((0.740684 -0.75858))
                ((0.808105 -0.343903)))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                 is
                 0)
                ((0.574866 0.223377))
                ((0.911439 -0.192629))))))
            ((n.name is z_2)
             ((R:segstate.parent.R:SylStructure.parent.asyl_in < 2.5)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                <
                0.2)
               ((0.122515 -1.16447))
               ((0.358047 -0.99653)))
              ((1.26641 -0.582942)))
             ((R:segstate.parent.p.name is p)
              ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.6)
               ((n.name is l_3)
                ((0.337245 -0.744677))
                ((R:segstate.parent.name is l)
                 ((0.482789 -0.0966315))
                 ((R:segstate.parent.nn.ph_cplace is l)
                  ((0.315818 -0.61725))
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    3.5)
                   ((1.23465 -0.0454967))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                     <
                     0.3)
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      3.8)
                     ((0.143602 -0.699272))
                     ((0.391637 -0.474896)))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                      is
                      0)
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                       <
                       5.2)
                      ((0.655497 -0.202045))
                      ((0.418681 -0.491299)))
                     ((0.333071 -0.513798))))))))
               ((n.name is r_3)
                ((0.154105 -0.764081))
                ((R:segstate.parent.R:SylStructure.parent.syl_out < 5)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((0.407084 -0.398752))
                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                    <
                    2.4)
                   ((0.336036 -0.529751))
                   ((0.0675969 -0.637711))))
                 ((0.198827 -0.656085)))))
              ((n.name is y_2)
               ((R:segstate.parent.p.ph_cvox is +)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                  <
                  1.2)
                 ((0.0782813 -0.623733))
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                   <
                   1.8)
                  ((0.263064 -0.527156))
                  ((0.577423 -0.219994))))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                  is
                  0)
                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                   <
                   2.5)
                  ((1.16241 1.39728))
                  ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                    <
                    1.2)
                   ((0.959521 0.970307))
                   ((0.963832 0.836114))))
                 ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                   <
                   1.1)
                  ((0.697665 -0.074765))
                  ((1.17102 0.494535)))))
               ((R:segstate.parent.pp.ph_vc is +)
                ((R:segstate.parent.p.name is pau)
                 ((R:segstate.parent.lisp_onset_glide is 0)
                  ((R:segstate.parent.lisp_coda_stop is 0)
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     1.2)
                    ((R:segstate.parent.pp.ph_vfront is 2)
                     ((1.00402 0.0848419))
                     ((R:segstate.parent.ph_cplace is l)
                      ((1.31805 0.211125))
                      ((0.969881 0.544368))))
                    ((0.661974 -0.18395)))
                   ((0.635767 -0.295786)))
                  ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                    <
                    1.3)
                   ((1.71016 0.730563))
                   ((R:segstate.parent.ph_cplace is a)
                    ((1.0801 0.411999))
                    ((R:segstate.parent.nn.ph_ctype is f)
                     ((0.862508 -0.10591))
                     ((1.11087 0.267583))))))
                 ((n.name is r_2)
                  ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                    <
                    1.1)
                   ((R:segstate.parent.pp.ph_vfront is 1)
                    ((0.641206 0.438547))
                    ((1.41084 1.11717)))
                   ((R:segstate.parent.nn.ph_ctype is 0)
                    ((1.15725 0.668399))
                    ((R:segstate.parent.p.name is t)
                     ((0.29987 -0.536946))
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                       <
                       2.4)
                      ((0.424233 -0.394191))
                      ((0.583956 -0.075737))))))
                  ((R:segstate.parent.p.name is k)
                   ((R:segstate.parent.ph_cplace is p)
                    ((0.397103 -0.768267))
                    ((R:segstate.parent.nn.ph_cplace is l)
                     ((0.725921 0.0172547))
                     ((R:segstate.parent.R:SylStructure.parent.position_type
                       is
                       mid)
                      ((0.334112 -0.609042))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                        is
                        0)
                       ((R:segstate.parent.ph_cplace is a)
                        ((0.587026 -0.336559))
                        ((0.363767 -0.668728)))
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         4.2)
                        ((1.03553 -0.0248856))
                        ((0.629533 -0.248209)))))))
                   ((n.name is v_2)
                    ((0.110923 -0.784372))
                    ((R:segstate.parent.name is n)
                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                       <
                       6.6)
                      ((R:segstate.parent.R:SylStructure.parent.next_accent
                        <
                        0.8)
                       ((0.33119 -0.317447))
                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                         <
                         2.4)
                        ((n.name is aa_1)
                         ((0.0785834 -0.554598))
                         ((R:segstate.parent.n.ph_vrnd is +)
                          ((0.0528093 -0.548414))
                          ((0.14429 -0.461571))))
                        ((R:segstate.parent.R:SylStructure.parent.asyl_out
                          <
                          2.2)
                         ((0.0889527 -0.528927))
                         ((0.598624 -0.300795)))))
                      ((R:segstate.parent.R:SylStructure.parent.position_type
                        is
                        single)
                       ((0.323715 -0.362179))
                       ((0.789612 -0.0874943))))
                     ((R:segstate.parent.n.ph_vc is +)
                      ((n.name is r_3)
                       ((R:segstate.parent.p.name is l)
                        ((0.481503 0.046066))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                          <
                          3.6)
                         ((R:segstate.parent.pp.ph_vfront is 2)
                          ((0.516534 -0.198101))
                          ((0.498062 -0.469398)))
                         ((0.317703 -0.629269))))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                         is
                         0)
                        ((n.name is ay_1)
                         ((1.17339 0.450847))
                         ((n.name is w_2)
                          ((R:segstate.parent.pp.ph_vfront is 1)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                             <
                             0.3)
                            ((0.902098 0.242103))
                            ((1.21323 0.67659)))
                           ((0.671837 -0.29245)))
                          ((R:segstate.parent.R:SylStructure.parent.asyl_in
                            <
                            2.4)
                           ((n.name is iy_1)
                            ((0.743701 0.13806))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              5)
                             ((R:segstate.parent.R:SylStructure.parent.next_accent
                               <
                               4.2)
                              ((R:segstate.parent.n.name is ay)
                               ((0.508932 -0.469525))
                               ((R:segstate.parent.name is jh)
                                ((0.565214 -0.546144))
                                ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                  <
                                  1.2)
                                 ((R:segstate.parent.nn.ph_cplace is v)
                                  ((0.904678 0.244375))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                    is
                                    0)
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                     <
                                     0.3)
                                    ((0.739328 0.196461))
                                    ((0.701137 -0.110503)))
                                   ((R:segstate.parent.p.ph_ctype is s)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                      <
                                      0.3)
                                     ((0.386455 -0.504624))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                       <
                                       1.2)
                                      ((0.544612 -0.30025))
                                      ((0.660046 -0.0381784))))
                                    ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                      <
                                      0.1)
                                     ((R:segstate.parent.pp.ph_vheight
                                       is
                                       1)
                                      ((0.749928 0.203494))
                                      ((0.788536 -0.102797)))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                       <
                                       3.4)
                                      ((R:segstate.parent.R:SylStructure.parent.next_accent
                                        <
                                        0.3)
                                       ((0.774957 -0.298502))
                                       ((1.10057 -0.0847136)))
                                      ((0.525437 -0.405128)))))))
                                 ((R:segstate.parent.n.name is ax)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    2.6)
                                   ((0.659951 -0.314789))
                                   ((0.690588 -0.0128186)))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                    <
                                    0.1)
                                   ((0.785429 -0.16285))
                                   ((R:segstate.parent.nn.ph_cplace
                                     is
                                     a)
                                    ((0.492822 -0.342472))
                                    ((0.33366 -0.705064))))))))
                              ((0.866044 0.139971)))
                             ((R:segstate.parent.pp.ph_vlng is a)
                              ((0.731852 -0.0192387))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                5.5)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                 is
                                 0)
                                ((0.527 -0.446672))
                                ((0.375303 -0.590436)))
                               ((R:segstate.parent.name is w)
                                ((0.344528 -0.621129))
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  1.2)
                                 ((R:segstate.parent.pp.ph_vheight
                                   is
                                   2)
                                  ((0.162456 -0.619448))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                                    <
                                    2.2)
                                   ((0.464702 -0.413966))
                                   ((0.626458 -0.13317))))
                                 ((0.82463 -0.0649379))))))))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                             <
                             0.1)
                            ((R:segstate.parent.nn.ph_ctype is f)
                             ((0.60663 -0.452578))
                             ((0.45366 -0.122036)))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                              <
                              1.2)
                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                               <
                               1.2)
                              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                <
                                1.2)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                 <
                                 0.3)
                                ((0.84444 -0.0793057))
                                ((0.744865 0.114738)))
                               ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                 <
                                 3.3)
                                ((0.419646 -0.628072))
                                ((0.609849 -0.0903629))))
                              ((0.965751 0.18847)))
                             ((0.85708 0.280742)))))))
                        ((n.name is ah_1)
                         ((1.46596 0.820861))
                         ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                           <
                           2.2)
                          ((n.name is iy_1)
                           ((R:segstate.parent.pp.ph_vfront is 1)
                            ((0.41951 -0.250041))
                            ((1.52191 0.725325)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                             is
                             0)
                            ((R:segstate.parent.R:SylStructure.parent.position_type
                              is
                              final)
                             ((0.380583 -0.595721))
                             ((R:segstate.parent.R:SylStructure.parent.next_accent
                               <
                               3)
                              ((R:segstate.parent.nn.ph_ctype is f)
                               ((0.381633 -0.389552))
                               ((R:segstate.parent.R:SylStructure.parent.position_type
                                 is
                                 single)
                                ((R:segstate.parent.nn.ph_ctype is s)
                                 ((1.03046 0.333781))
                                 ((0.604285 0.0430212)))
                                ((R:segstate.parent.pp.ph_vheight is 2)
                                 ((0.48464 -0.326544))
                                 ((0.624207 -0.101945)))))
                              ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                <
                                2.4)
                               ((0.330007 -0.341985))
                               ((0.595601 -0.463639)))))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                              <
                              5)
                             ((R:segstate.parent.pp.ph_vheight is 1)
                              ((R:segstate.parent.name is dh)
                               ((1.19636 0.618321))
                               ((n.name is w_2)
                                ((1.06289 0.458906))
                                ((R:segstate.parent.p.ph_cvox is +)
                                 ((R:segstate.parent.lisp_onset_glide
                                   is
                                   0)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                    is
                                    0)
                                   ((1.14737 0.308286))
                                   ((0.752769 -0.28807)))
                                  ((1.39745 0.6608)))
                                 ((R:segstate.parent.nn.ph_ctype is f)
                                  ((0.800781 -0.00647334))
                                  ((0.690568 -0.427512))))))
                              ((n.name is ih_1)
                               ((1.18418 0.39261))
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 4.8)
                                ((R:segstate.parent.nn.ph_vrnd is -)
                                 ((0.390851 -0.477521))
                                 ((R:segstate.parent.nn.ph_ctype is r)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                    <
                                    1.1)
                                   ((0.587908 -0.407))
                                   ((0.789691 -0.0917938)))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    5.5)
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                     <
                                     0.3)
                                    ((R:segstate.parent.nn.ph_cplace
                                      is
                                      a)
                                     ((0.431936 -0.353992))
                                     ((1.0538 0.00212463)))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                      is
                                      0)
                                     ((R:segstate.parent.p.ph_cvox
                                       is
                                       +)
                                      ((1.2111 0.0725448))
                                      ((0.725952 -0.246517)))
                                     ((R:segstate.parent.lisp_onset_stop
                                       is
                                       0)
                                      ((R:segstate.parent.pp.ph_vlng
                                        is
                                        s)
                                       ((0.969066 0.350024))
                                       ((0.695111 -0.175893)))
                                      ((1.01697 0.512137)))))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     7)
                                    ((0.457479 -0.405111))
                                    ((R:segstate.parent.pp.ph_vfront
                                      is
                                      3)
                                     ((0.830995 0.122578))
                                     ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                       <
                                       0.2)
                                      ((0.613855 -0.0410219))
                                      ((R:segstate.parent.R:SylStructure.parent.next_accent
                                        <
                                        1.2)
                                       ((0.413102 -0.438892))
                                       ((0.67233 -0.101976)))))))))
                                ((1.44491 0.299112)))))
                             ((R:segstate.parent.nn.ph_ctype is s)
                              ((0.639109 -0.0786935))
                              ((R:segstate.parent.p.ph_ctype is f)
                               ((0.28703 -0.482451))
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 9.6)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                  is
                                  0)
                                 ((0.806273 -0.0567836))
                                 ((0.448626 -0.390458)))
                                ((0.546281 -0.432333))))))))
                          ((R:segstate.parent.name is w)
                           ((2.06735 1.39366))
                           ((R:segstate.parent.nn.ph_ctype is s)
                            ((0.667135 -0.401526))
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              2.7)
                             ((0.776873 -0.226062))
                             ((R:segstate.parent.ph_cplace is l)
                              ((0.777619 -0.0507194))
                              ((R:segstate.parent.pos_in_syl is 3)
                               ((0.989607 0.0293634))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                 is
                                 0)
                                ((0.875377 0.282556))
                                ((1.32312 0.692905))))))))))))
                      ((R:segstate.parent.nn.ph_vheight is 1)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                         is
                         0)
                        ((0.482498 -0.378251))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          3.3)
                         ((R:segstate.parent.lisp_onset_stop is 0)
                          ((1.0707 0.0985047))
                          ((0.775242 -0.401252)))
                         ((1.18072 0.508827))))
                       ((R:segstate.parent.name is g)
                        ((0.176317 -0.884369))
                        ((R:segstate.parent.n.ph_ctype is f)
                         ((R:segstate.parent.n.name is dh)
                          ((0.452211 -0.404838))
                          ((0.339556 -0.811746)))
                         ((R:segstate.parent.lisp_coda_stop is 0)
                          ((R:segstate.parent.R:SylStructure.parent.asyl_in
                            <
                            1.2)
                           ((0.796939 -0.142173))
                           ((1.11916 0.408635)))
                          ((R:segstate.parent.p.name is l)
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             5.4)
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              0.4)
                             ((0.475976 -0.227087))
                             ((0.388857 -0.475581)))
                            ((0.619411 -0.00465475)))
                           ((n.name is d_3)
                            ((0.197132 -0.641243))
                            ((R:segstate.parent.pos_in_syl is 3)
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               4.5)
                              ((0.679061 -0.116487))
                              ((0.419506 -0.415573)))
                             ((0.183488 -0.583881)))))))))))))))
                ((n.name is b_3)
                 ((R:segstate.parent.p.name is pau)
                  ((R:segstate.parent.pp.ph_cplace is a)
                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                     <
                     0.7)
                    ((0.620941 -0.509126))
                    ((R:segstate.parent.nn.ph_cplace is a)
                     ((0.126468 -0.789026))
                     ((0.10339 -0.845051))))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                     <
                     1.2)
                    ((R:segstate.parent.R:SylStructure.parent.asyl_out
                      <
                      2.6)
                     ((0.102202 -0.847651))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                       is
                       0)
                      ((1.53421e-06 -0.897019))
                      ((1.41867e-06 -0.897018))))
                    ((0.146645 -0.814738))))
                  ((0.911986 0.0389189)))
                 ((R:segstate.parent.name is b)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in < 2)
                   ((R:segstate.parent.R:SylStructure.parent.asyl_out
                     <
                     3.4)
                    ((n.name is ah_1)
                     ((0.348543 -0.224745))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                       <
                       1.2)
                      ((0.15507 -0.812558))
                      ((0.557644 -0.644773))))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                      <
                      1.2)
                     ((n.name is b_2)
                      ((0.586433 -0.733468))
                      ((0.671435 -0.0379301)))
                     ((1.41629 0.340543))))
                   ((R:segstate.parent.p.name is d)
                    ((0.746484 -0.360178))
                    ((R:segstate.parent.n.ph_vrnd is -)
                     ((n.name is ih_1)
                      ((0.840898 -0.0351002))
                      ((R:segstate.parent.R:SylStructure.parent.next_accent
                        <
                        2.4)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                         is
                         0)
                        ((0.867989 0.220666))
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          initial)
                         ((1.55185 0.933946))
                         ((0.999809 0.45475))))
                       ((0.820261 0.163519))))
                     ((R:segstate.parent.R:SylStructure.parent.asyl_in
                       <
                       1.5)
                      ((0.943256 0.817167))
                      ((1.22679 1.26988))))))
                  ((n.name is w_3)
                   ((R:segstate.parent.pos_in_syl is 1)
                    ((0.116472 -0.889105))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                      <
                      2.5)
                     ((R:segstate.parent.nn.ph_cplace is a)
                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                        <
                        1.2)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                         is
                         0)
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                          is
                          0)
                         ((0.532381 -0.281826))
                         ((0.403052 -0.386266)))
                        ((0.478207 -0.641845)))
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                         <
                         3.1)
                        ((0.600075 -0.359519))
                        ((0.816763 -0.0296438))))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                        is
                        0)
                       ((0.514957 -0.393907))
                       ((0.422049 -0.641846))))
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                       <
                       2)
                      ((0.637738 -0.388247))
                      ((R:segstate.parent.n.name is ih)
                       ((0.149173 -0.869584))
                       ((0.454649 -0.633576))))))
                   ((n.name is g_3)
                    ((R:segstate.parent.pp.ph_cplace is a)
                     ((0.866744 -0.703283))
                     ((0.134277 -1.1002)))
                    ((n.name is dh_2)
                     ((R:segstate.parent.p.name is d)
                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                        <
                        4)
                       ((1.06659e-06 -0.63909))
                       ((0.612387 -0.426892)))
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                        <
                        2.3)
                       ((0.545158 -0.237246))
                       ((R:segstate.parent.nn.ph_vc is +)
                        ((0.42095 -0.308583))
                        ((R:segstate.parent.nn.ph_cplace is l)
                         ((R:segstate.parent.pp.ph_vc is -)
                          ((1.21098 0.0842274))
                          ((1.34084 0.655948)))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                           <
                           2.4)
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                            <
                            9.6)
                           ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                             <
                             1.2)
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              12)
                             ((R:segstate.parent.R:SylStructure.parent.syl_out
                               <
                               6.4)
                              ((0.67461 -0.277431))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                is
                                0)
                               ((1.35988 0.238465))
                               ((1.16153 -0.0122656))))
                             ((0.479077 -0.329097)))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                              <
                              0.2)
                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                               <
                               3.4)
                              ((1.48473 0.597903))
                              ((1.28835 0.162313)))
                             ((0.787681 -0.030174))))
                           ((0.431401 -0.309722)))
                          ((1.19986 0.523385)))))))
                     ((R:segstate.parent.lisp_onset_glide is 0)
                      ((n.name is v_3)
                       ((0.193406 -0.906148))
                       ((n.name is ao_1)
                        ((0.611055 0.307912))
                        ((R:segstate.parent.nn.ph_ctype is r)
                         ((R:segstate.parent.name is m)
                          ((0.319381 -0.720738))
                          ((R:segstate.parent.R:SylStructure.parent.last_accent
                            <
                            1)
                           ((R:segstate.parent.name is dh)
                            ((R:segstate.parent.pp.ph_vc is -)
                             ((0.277513 -0.357126))
                             ((n.name is eh_1)
                              ((0.105408 -0.546563))
                              ((0.264805 -0.444637))))
                            ((R:segstate.parent.name is n)
                             ((0.19506 -0.408445))
                             ((0.570145 -0.186555))))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                             is
                             0)
                            ((0.322177 -0.607601))
                            ((0.336291 -0.47119)))))
                         ((R:segstate.parent.p.ph_ctype is f)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                            <
                            0.1)
                           ((0.821263 0.190984))
                           ((R:segstate.parent.lisp_onset_nasal is 0)
                            ((0.644698 -0.404624))
                            ((0.476831 -0.0511889))))
                          ((R:segstate.parent.n.name is iy)
                           ((R:segstate.parent.name is m)
                            ((0.404496 -0.276978))
                            ((2.13631 0.343584)))
                           ((R:segstate.parent.n.name is aw)
                            ((R:segstate.parent.name is n)
                             ((0.276888 -0.331079))
                             ((0.996672 0.244031)))
                            ((R:segstate.parent.name is dh)
                             ((R:segstate.parent.pp.ph_cplace is a)
                              ((R:segstate.parent.p.name is d)
                               ((0.126782 -0.538755))
                               ((n.name is ax_1)
                                ((0.422406 -0.102629))
                                ((1.25552 0.483125))))
                              ((n.name is ax_1)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                 <
                                 0.3)
                                ((0.27254 -0.408312))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                  <
                                  4.6)
                                 ((0.333329 -0.383915))
                                 ((R:segstate.parent.pp.ph_vc is -)
                                  ((0.526806 0.132222))
                                  ((0.478905 -0.123678)))))
                               ((R:segstate.parent.nn.ph_ctype is f)
                                ((0.607359 -0.371264))
                                ((R:segstate.parent.R:SylStructure.parent.syl_in
                                  <
                                  1.3)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                   <
                                   9)
                                  ((0.0536712 -0.598972))
                                  ((0.0777768 -0.583159)))
                                 ((0.165644 -0.483537))))))
                             ((n.name is m_3)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                <
                                0.3)
                               ((0.440016 -0.92351))
                               ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                 <
                                 0.4)
                                ((0.519195 -0.670639))
                                ((0.706923 -0.231217))))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                is
                                0)
                               ((n.name is d_3)
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  1.3)
                                 ((0.197276 -0.601556))
                                 ((0.232807 -0.478721)))
                                ((R:segstate.parent.nn.ph_cplace is b)
                                 ((0.253952 -0.594515))
                                 ((R:segstate.parent.name is v)
                                  ((0.399013 -0.575918))
                                  ((n.name is d_2)
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                     is
                                     0)
                                    ((0.183569 -0.41226))
                                    ((0.123648 -0.522156)))
                                   ((R:segstate.parent.lisp_onset_nasal
                                     is
                                     0)
                                    ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                      <
                                      1.2)
                                     ((0.672613 -0.285991))
                                     ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                       <
                                       2.4)
                                      ((0.668782 0.318501))
                                      ((0.81031 0.0288365))))
                                    ((R:segstate.parent.n.name is ay)
                                     ((0.768048 -0.130666))
                                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                                       <
                                       0.8)
                                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                        <
                                        0.5)
                                       ((R:segstate.parent.nn.ph_cplace
                                         is
                                         a)
                                        ((0.0186857 -0.540092))
                                        ((0.200932 -0.485443)))
                                       ((0.19036 -0.400899)))
                                      ((n.name is m_2)
                                       ((0.832708 -0.110658))
                                       ((R:segstate.parent.R:SylStructure.parent.position_type
                                         is
                                         initial)
                                        ((0.270192 -0.371846))
                                        ((0.143955 -0.443787)))))))))))
                               ((R:segstate.parent.name is jh)
                                ((0.411405 -0.73942))
                                ((R:segstate.parent.name is d)
                                 ((n.name is d_2)
                                  ((0.117065 -0.527183))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                    <
                                    1.3)
                                   ((0.312224 -0.387439))
                                   ((0.633888 0.000917464))))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   2)
                                  ((R:segstate.parent.n.ph_vrnd is +)
                                   ((0.157576 -0.420716))
                                   ((R:segstate.parent.name is m)
                                    ((0.0739118 -0.50382))
                                    ((0.199593 -0.601728))))
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                    <
                                    3.6)
                                   ((0.40256 -0.300524))
                                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                     <
                                     3.5)
                                    ((0.230004 -0.511678))
                                    ((0.193599 -0.46327))))))))))))))))
                      ((R:segstate.parent.n.name is oy)
                       ((1.51315 0.869071))
                       ((n.name is r_3)
                        ((R:segstate.parent.p.name is b)
                         ((0.624037 0.0660557))
                         ((R:segstate.parent.lisp_coda_nasal is 0)
                          ((R:segstate.parent.n.name is iy)
                           ((0.103213 -0.808518))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                             is
                             0)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                              is
                              0)
                             ((0.515079 -0.520266))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                               is
                               0)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                <
                                1.2)
                               ((0.324753 -0.432403))
                               ((0.491232 0.0189352)))
                              ((0.76442 0.0135103))))
                            ((R:segstate.parent.R:SylStructure.parent.syl_in
                              <
                              6.4)
                             ((0.537129 -0.350706))
                             ((0.169458 -0.783854)))))
                          ((R:segstate.parent.p.name is f)
                           ((0.152048 -0.816054))
                           ((0.35312 -0.588768)))))
                        ((R:segstate.parent.p.name is f)
                         ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                           <
                           1.2)
                          ((0.907728 -0.0549866))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            4.8)
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             4.5)
                            ((0.333937 -0.389196))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              2.6)
                             ((0.277583 -0.418579))
                             ((0.192892 -0.634518))))
                           ((0.0755087 -0.647626))))
                         ((R:segstate.parent.lisp_onset_stop is 0)
                          ((R:segstate.parent.lisp_coda_nasal is 0)
                           ((n.name is er_1)
                            ((0.337636 -0.634402))
                            ((n.name is aa_1)
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                               is
                               0)
                              ((1.02986 -0.0504035))
                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                <
                                3.6)
                               ((0.628907 -0.491066))
                               ((0.24541 -0.740776))))
                             ((n.name is r_2)
                              ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                <
                                2.4)
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 4.8)
                                ((0.784542 0.253698))
                                ((0.498973 -0.119662)))
                               ((0.67495 0.523521)))
                              ((R:segstate.parent.name is r)
                               ((R:segstate.parent.n.ph_vrnd is +)
                                ((0.192024 -0.584369))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                  is
                                  0)
                                 ((0.263269 -0.419109))
                                 ((0.813421 -0.071234))))
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 4.8)
                                ((n.name is ih_1)
                                 ((R:segstate.parent.nn.ph_cplace is d)
                                  ((R:segstate.parent.R:SylStructure.parent.next_accent
                                    <
                                    0.3)
                                   ((0.574783 -0.264828))
                                   ((0.389338 -0.579649)))
                                  ((0.611776 -0.0815546)))
                                 ((R:segstate.parent.p.name is d)
                                  ((n.name is w_2)
                                   ((1.26841 1.31436))
                                   ((0.574118 -0.185343)))
                                  ((n.name is w_2)
                                   ((R:segstate.parent.pp.ph_cplace
                                     is
                                     0)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                      is
                                      0)
                                     ((0.528415 -0.423709))
                                     ((0.179058 -0.657597)))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                      <
                                      0.1)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                       <
                                       0.3)
                                      ((0.747622 -0.24154))
                                      ((1.63456 0.948109)))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                       <
                                       1.8)
                                      ((0.883815 0.19207))
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                        is
                                        0)
                                       ((0.807762 0.0446083))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                         is
                                         0)
                                        ((0.566973 -0.333756))
                                        ((0.33862 -0.499282)))))))
                                   ((R:segstate.parent.name is w)
                                    ((n.name is iy_1)
                                     ((0.458799 -0.117911))
                                     ((R:segstate.parent.pp.ph_cplace
                                       is
                                       0)
                                      ((0.87925 0.774084))
                                      ((0.605182 0.19691))))
                                    ((n.name is ay_1)
                                     ((0.190947 -0.393293))
                                     ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                       <
                                       2.2)
                                      ((R:segstate.parent.nn.ph_cplace
                                        is
                                        0)
                                       ((0.895043 0.448235))
                                       ((n.name is uw_1)
                                        ((0.474479 -0.333487))
                                        ((R:segstate.parent.n.name
                                          is
                                          eh)
                                         ((0.508568 -0.229881))
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                           <
                                           0.3)
                                          ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                            <
                                            1.2)
                                           ((0.621832 0.426587))
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                             is
                                             0)
                                            ((0.597411 0.010741))
                                            ((0.771658 0.258925))))
                                          ((R:segstate.parent.p.name
                                            is
                                            pau)
                                           ((0.565638 0.269243))
                                           ((0.42999 -0.217832)))))))
                                      ((0.481485 -0.351112))))))))
                                ((0.406025 -0.441814)))))))
                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                             <
                             1.2)
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              6.1)
                             ((1.21811 0.852))
                             ((R:segstate.parent.pp.ph_cplace is a)
                              ((0.876931 0.545081))
                              ((0.702944 -0.0382461))))
                            ((0.571769 -0.206544))))
                          ((n.name is ey_1)
                           ((0.412219 0.413521))
                           ((n.name is ih_1)
                            ((0.906562 0.32267))
                            ((R:segstate.parent.p.ph_cvox is +)
                             ((R:segstate.parent.nn.ph_cplace is a)
                              ((R:segstate.parent.n.ph_vrnd is +)
                               ((0.668294 -0.333927))
                               ((n.name is r_2)
                                ((0.606024 -0.284379))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                  <
                                  3)
                                 ((0.654257 -0.0287816))
                                 ((0.88152 0.719097)))))
                              ((R:segstate.parent.nn.ph_ctype is 0)
                               ((0.819714 -0.0314844))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                 is
                                 0)
                                ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                  <
                                  0.4)
                                 ((0.427588 -0.386198))
                                 ((0.427712 -0.141949)))
                                ((R:segstate.parent.p.name is g)
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   0.5)
                                  ((0.25739 -0.565315))
                                  ((0.435679 -0.210604)))
                                 ((0.139844 -0.611487))))))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                               <
                               0.4)
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                <
                                5)
                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                 <
                                 4)
                                ((0.279724 -0.398393))
                                ((0.171144 -0.592112)))
                               ((0.189747 -0.6935)))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((R:segstate.parent.ph_cplace is a)
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  1.2)
                                 ((0.201845 -0.562379))
                                 ((0.562007 -0.358001)))
                                ((0.22378 -0.693574)))
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 6)
                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                  <
                                  0.8)
                                 ((0.914619 0.0616626))
                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                   <
                                   6.2)
                                  ((0.700246 -0.214852))
                                  ((0.222877 -0.580353))))
                                ((0.390716 -0.511433))))))))))))))))))))))))))))
      ((R:segstate.parent.p.ph_ctype is r)
       ((R:segstate.parent.n.ph_ctype is 0)
        ((n.name is ey_2)
         ((0.838238 1.62117))
         ((n.name is iy_3)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
            is
            0)
           ((0.0818446 -1.12297))
           ((0.434364 -0.839262)))
          ((R:segstate.parent.ph_vheight is 2)
           ((R:segstate.parent.lisp_onset_stop is 0)
            ((R:segstate.parent.nn.ph_vheight is 3)
             ((1.39225 0.285936))
             ((R:segstate.parent.nn.ph_cplace is a)
              ((0.728207 -0.171109))
              ((R:segstate.parent.pp.ph_cplace is 0)
               ((0.399824 -0.640941))
               ((0.912733 -0.394448)))))
            ((0.658903 -0.700187)))
           ((n.name is uw_2)
            ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
              <
              1.1)
             ((0.806039 -0.021539))
             ((0.464498 -0.602272)))
            ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 6.6)
             ((R:segstate.parent.p.name is w)
              ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                <
                0.1)
               ((0.795739 0.18596))
               ((1.30386 0.696371)))
              ((R:segstate.parent.name is p)
               ((1.17188 0.61778))
               ((R:segstate.parent.nn.ph_ctype is l)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                  <
                  0.1)
                 ((0.454391 -0.457404))
                 ((0.795869 0.0599684)))
                ((R:segstate.parent.n.ph_vrnd is +)
                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                   <
                   3.4)
                  ((0.971429 0.353771))
                  ((1.4322 0.862296)))
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 6)
                  ((R:segstate.parent.R:SylStructure.parent.position_type
                    is
                    mid)
                   ((0.77724 0.465769))
                   ((R:segstate.parent.name is t)
                    ((0.901685 -0.251349))
                    ((R:segstate.parent.R:SylStructure.parent.position_type
                      is
                      final)
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                       <
                       2.5)
                      ((0.892836 0.58727))
                      ((1.16616 0.228224)))
                     ((R:segstate.parent.n.name is ae)
                      ((0.649394 -0.244208))
                      ((R:segstate.parent.R:SylStructure.parent.asyl_out
                        <
                        2.4)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                         is
                         0)
                        ((R:segstate.parent.name is pau)
                         ((0.941957 0.213366))
                         ((1.00678 0.3147)))
                        ((R:segstate.parent.R:SylStructure.parent.last_accent
                          <
                          0.5)
                         ((0.537139 -0.304371))
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           4.6)
                          ((0.551889 -0.160641))
                          ((0.748835 0.135281)))))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_in
                         <
                         0.3)
                        ((0.609991 0.0546259))
                        ((R:segstate.parent.nn.ph_ctype is s)
                         ((0.678469 0.135873))
                         ((1.00817 0.552751)))))))))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                    <
                    0.1)
                   ((0.442035 -0.356186))
                   ((R:segstate.parent.pos_in_syl is 0)
                    ((0.758118 0.195559))
                    ((0.612223 -0.204615)))))))))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
               <
               1.2)
              ((1.44817 0.902588))
              ((0.686878 0.191564))))))))
        ((n.name is iy_2)
         ((R:segstate.parent.lisp_onset_stop is 0)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
            <
            1.2)
           ((R:segstate.parent.n.ph_cvox is -)
            ((0.728158 0.506092))
            ((0.759127 0.105818)))
           ((0.520028 -0.179603)))
          ((R:segstate.parent.R:SylStructure.parent.syl_in < 4.2)
           ((0.953923 1.21364))
           ((0.647708 0.625198))))
         ((R:segstate.parent.lisp_coda_stop is 0)
          ((n.name is aa_3)
           ((R:segstate.parent.R:SylStructure.parent.position_type
             is
             initial)
            ((0.564209 0.242954))
            ((R:segstate.parent.nn.ph_vrnd is -)
             ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 0.7)
              ((0.0608786 -0.800477))
              ((0.346807 -0.63363)))
             ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 2.7)
              ((R:segstate.parent.R:SylStructure.parent.asyl_in < 0.3)
               ((R:segstate.parent.nn.ph_cplace is a)
                ((0.0233459 -0.833318))
                ((R:segstate.parent.R:SylStructure.parent.next_accent
                  <
                  2)
                 ((0.0806715 -0.787414))
                 ((0.0378249 -0.821994))))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                 <
                 0.3)
                ((0.0822106 -0.799602))
                ((0.21053 -0.712326))))
              ((R:segstate.parent.R:SylStructure.parent.asyl_in < 2.5)
               ((0.0503567 -0.819729))
               ((1.50676e-06 -0.838981))))))
           ((R:segstate.parent.n.ph_cplace is l)
            ((R:segstate.parent.pp.ph_cplace is a)
             ((R:segstate.parent.pos_in_syl is 2)
              ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 1.6)
               ((0.609322 -0.155669))
               ((0.985178 0.143081)))
              ((R:segstate.parent.p.name is y)
               ((0.393821 -0.281397))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                 is
                 0)
                ((0.310492 -0.470186))
                ((0.195516 -0.646192)))))
             ((n.name is ah_3)
              ((0.180433 -0.800123))
              ((n.name is ih_3)
               ((2.77937e-06 -0.756024))
               ((R:segstate.parent.name is ih)
                ((R:segstate.parent.nn.ph_vc is +)
                 ((0.198101 -0.480878))
                 ((0.564783 -0.114473)))
                ((R:segstate.parent.ph_vrnd is -)
                 ((R:segstate.parent.name is ax)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    2)
                   ((0.131082 -0.43556))
                   ((0.0134964 -0.496339)))
                  ((n.name is ah_2)
                   ((0.203009 -0.448935))
                   ((R:segstate.parent.n.ph_ctype is s)
                    ((R:segstate.parent.R:SylStructure.parent.asyl_out
                      <
                      1.2)
                     ((0.577426 -0.423797))
                     ((0.296145 -0.677051)))
                    ((R:segstate.parent.lisp_onset_stop is 0)
                     ((n.name is m_1)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                        <
                        0.1)
                       ((0.223697 -0.667384))
                       ((0.0808971 -0.769215)))
                      ((0.374323 -0.876703)))
                     ((0.355223 -0.591835))))))
                 ((R:segstate.parent.nn.ph_ctype is 0)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                    <
                    0.1)
                   ((0.282066 -0.546139))
                   ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                     <
                     3)
                    ((0.304277 -0.490695))
                    ((0.401728 -0.295042))))
                  ((0.137772 -0.340826))))))))
            ((R:segstate.parent.name is ah)
             ((n.name is n_1)
              ((0.466114 -0.404391))
              ((n.name is ah_2)
               ((R:segstate.parent.R:SylStructure.parent.syl_in < 3.6)
                ((0.239454 -0.31542))
                ((0.833967 0.027971)))
               ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 3)
                ((0.613635 -0.0330396))
                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                  <
                  5.1)
                 ((0.926637 0.553833))
                 ((0.481769 0.208409))))))
             ((n.name is aa_2)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                is
                0)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                 is
                 0)
                ((0.341403 -0.379787))
                ((0.388221 -0.483696)))
               ((R:segstate.parent.nn.ph_cplace is l)
                ((0.257553 -0.474418))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                  is
                  0)
                 ((R:segstate.parent.nn.ph_vc is +)
                  ((0.146459 -0.540134))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                    <
                    0.3)
                   ((R:segstate.parent.nn.ph_ctype is 0)
                    ((0.144389 -0.548019))
                    ((0.104759 -0.595026)))
                   ((R:segstate.parent.pp.ph_vc is +)
                    ((0.0765046 -0.613581))
                    ((2.16601e-06 -0.632137)))))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                   is
                   0)
                  ((0.141068 -0.553278))
                  ((0.353676 -0.442876))))))
              ((R:segstate.parent.ph_vheight is 3)
               ((R:segstate.parent.n.ph_ctype is r)
                ((R:segstate.parent.p.name is w)
                 ((0.452142 -0.152241))
                 ((n.name is r_1)
                  ((0.145512 -0.546035))
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    3.2)
                   ((0.221849 -0.819124))
                   ((0.230231 -0.626454)))))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                  <
                  1.2)
                 ((R:segstate.parent.R:SylStructure.parent.syl_in < 6)
                  ((R:segstate.parent.nn.ph_cplace is a)
                   ((0.509682 -0.213163))
                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                     <
                     5.8)
                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                      <
                      0.2)
                     ((0.848666 0.391741))
                     ((0.926127 0.727939)))
                    ((R:segstate.parent.pp.ph_vfront is 2)
                     ((0.57397 -0.141206))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                       <
                       0.1)
                      ((0.854739 0.45644))
                      ((0.85668 0.142359))))))
                  ((R:segstate.parent.n.ph_ctype is f)
                   ((R:segstate.parent.nn.ph_ctype is 0)
                    ((0.363352 -0.444504))
                    ((0.564607 -0.193796)))
                   ((1.03663 0.18114))))
                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                   <
                   1.4)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    5.5)
                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                     <
                     1.5)
                    ((R:segstate.parent.nn.ph_vc is +)
                     ((0.694872 -0.0324734))
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       6.7)
                      ((0.550342 -0.433567))
                      ((0.724998 -0.224113))))
                    ((R:segstate.parent.nn.ph_cplace is 0)
                     ((0.338389 -0.6724))
                     ((R:segstate.parent.R:SylStructure.parent.asyl_out
                       <
                       2.5)
                      ((0.627701 -0.163786))
                      ((0.240661 -0.596965)))))
                   ((0.767267 0.0102073)))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                    is
                    0)
                   ((R:segstate.parent.pp.ph_cplace is a)
                    ((0.655172 -0.227415))
                    ((0.498827 -0.427985)))
                   ((1.24186 0.258516))))))
               ((R:segstate.parent.p.name is y)
                ((n.name is uw_2)
                 ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                   <
                   1.2)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    4.7)
                   ((0.736017 0.218977))
                   ((0.731572 -0.154517)))
                  ((0.757837 0.727645)))
                 ((R:segstate.parent.ph_vheight is 1)
                  ((R:segstate.parent.R:SylStructure.parent.position_type
                    is
                    final)
                   ((1.40904 0.0322667))
                   ((n.name is uw_3)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                      <
                      0.3)
                     ((0.563051 -0.138227))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                       is
                       0)
                      ((0.0705319 -0.672399))
                      ((0.712641 -0.323737))))
                    ((R:segstate.parent.n.ph_ctype is r)
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                       <
                       4.8)
                      ((0.203124 -0.402667))
                      ((0.134737 -0.583821)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                       <
                       0.3)
                      ((0.17967 -0.696596))
                      ((0.128606 -0.627739))))))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    2)
                   ((1.28527 0.573116))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                     is
                     0)
                    ((0.535691 0.0056977))
                    ((0.435424 -0.27107))))))
                ((n.name is uw_3)
                 ((0.8343 0.401382))
                 ((R:segstate.parent.seg_onsetcoda is coda)
                  ((n.name is ey_2)
                   ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                     <
                     0.1)
                    ((0.259718 -0.38353))
                    ((0.598716 0.240942)))
                   ((R:segstate.parent.name is ax)
                    ((R:segstate.parent.n.name is s)
                     ((0.429867 0.121924))
                     ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                       <
                       0.1)
                      ((R:segstate.parent.R:SylStructure.parent.next_accent
                        <
                        3)
                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                         <
                         6.8)
                        ((0.0749423 -0.477594))
                        ((0.155591 -0.374007)))
                       ((0.651524 -0.190419)))
                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                        <
                        0.4)
                       ((0.55553 -0.0589827))
                       ((0.38241 -0.308765)))))
                    ((n.name is ih_2)
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                       <
                       1.2)
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                        <
                        1.5)
                       ((R:segstate.parent.pp.ph_vc is +)
                        ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                          <
                          0.1)
                         ((0.151862 -0.40346))
                         ((0.22709 -0.339275)))
                        ((0.0889585 -0.455243)))
                       ((R:segstate.parent.nn.ph_vrnd is -)
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                          <
                          0.1)
                         ((0.187541 -0.313806))
                         ((0.151862 -0.403461)))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                          is
                          0)
                         ((R:segstate.parent.n.name is dh)
                          ((0.274943 -0.30216))
                          ((0.274189 -0.0884583)))
                         ((0.222396 -0.347366)))))
                      ((0.440649 -0.0660204)))
                     ((n.name is t_1)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        2.4)
                       ((0.445903 -0.342979))
                       ((0.740401 -0.0270081)))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                        <
                        0.3)
                       ((R:segstate.parent.ph_vlng is l)
                        ((R:segstate.parent.n.ph_cplace is a)
                         ((0.275489 -0.845182))
                         ((0.248193 -0.682057)))
                        ((R:segstate.parent.lisp_onset_stop is 0)
                         ((R:segstate.parent.lisp_coda_glide is 0)
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            1)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                             <
                             0.2)
                            ((R:segstate.parent.n.name is dh)
                             ((0.230784 -0.333373))
                             ((0.364634 -0.54539)))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              3.1)
                             ((n.name is ih_3)
                              ((0.180456 -0.679564))
                              ((R:segstate.parent.nn.ph_vc is +)
                               ((0.332199 -0.525283))
                               ((0.239527 -0.610242))))
                             ((0.310003 -0.475921))))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             5)
                            ((R:segstate.parent.R:SylStructure.parent.position_type
                              is
                              initial)
                             ((0.523823 -0.30923))
                             ((R:segstate.parent.name is er)
                              ((0.339707 -0.491075))
                              ((0.300877 -0.599039))))
                            ((0.684491 -0.215141))))
                          ((0.519495 -0.645778)))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                           is
                           0)
                          ((0.3227 -0.523277))
                          ((0.235413 -0.771907)))))
                       ((n.name is s_1)
                        ((0.627676 0.0225024))
                        ((n.name is uw_2)
                         ((0.185938 -0.792422))
                         ((R:segstate.parent.pp.ph_cplace is d)
                          ((0.761805 -0.130358))
                          ((n.name is l_1)
                           ((R:segstate.parent.nn.ph_cplace is 0)
                            ((1.07263 0.0470504))
                            ((0.504671 -0.42771)))
                           ((R:segstate.parent.n.name is l)
                            ((R:segstate.parent.ph_vlng is s)
                             ((0.26036 -0.589632))
                             ((0.445163 -0.963529)))
                            ((R:segstate.parent.pp.ph_vrnd is +)
                             ((0.619498 -0.19757))
                             ((R:segstate.parent.n.ph_ctype is r)
                              ((0.377922 -0.824499))
                              ((R:segstate.parent.name is ey)
                               ((R:segstate.parent.nn.ph_vheight is 2)
                                ((0.484672 -0.524538))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                  <
                                  2.4)
                                 ((0.387996 -0.29423))
                                 ((0.686662 -0.0170104))))
                               ((R:segstate.parent.n.name is v)
                                ((0.171037 -0.688693))
                                ((R:segstate.parent.R:SylStructure.parent.next_accent
                                  <
                                  4)
                                 ((R:segstate.parent.pp.ph_vlng is a)
                                  ((0.612276 -0.15921))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                    is
                                    0)
                                   ((R:segstate.parent.n.ph_ctype is s)
                                    ((R:segstate.parent.ph_vlng is s)
                                     ((0.230898 -0.602351))
                                     ((0.271362 -0.745397)))
                                    ((R:segstate.parent.lisp_onset_stop
                                      is
                                      0)
                                     ((R:segstate.parent.name is ih)
                                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                                        <
                                        1.1)
                                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                                         <
                                         7.8)
                                        ((n.name is ih_3)
                                         ((0.250442 -0.391169))
                                         ((0.158159 -0.52647)))
                                        ((0.167084 -0.575707)))
                                       ((0.548693 -0.30274)))
                                      ((R:segstate.parent.n.ph_cvox
                                        is
                                        +)
                                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                                         <
                                         5)
                                        ((0.277768 -0.630098))
                                        ((0.148727 -0.771073)))
                                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                                         <
                                         0.3)
                                        ((0.203356 -0.613201))
                                        ((0.378533 -0.382376)))))
                                     ((R:segstate.parent.pp.ph_cplace
                                       is
                                       a)
                                      ((0.462571 -0.497986))
                                      ((0.370257 -0.265842)))))
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                     <
                                     4)
                                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                                      <
                                      0.3)
                                     ((R:segstate.parent.nn.ph_vc is +)
                                      ((0.434915 -0.316642))
                                      ((0.307168 -0.429544)))
                                     ((R:segstate.parent.name is eh)
                                      ((0.123471 -0.7107))
                                      ((R:segstate.parent.nn.ph_vc
                                        is
                                        +)
                                       ((0.413208 -0.401704))
                                       ((0.278554 -0.5781)))))
                                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                                      <
                                      8.9)
                                     ((0.553993 -0.0713517))
                                     ((0.421734 -0.427968))))))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                   <
                                   5.2)
                                  ((R:segstate.parent.pp.ph_cplace
                                    is
                                    a)
                                   ((0.430141 -0.251409))
                                   ((0.307278 -0.548014)))
                                  ((1.13582 0.00238328))))))))))))))))))
                  ((0.995997 0.258096))))))))))
          ((R:segstate.parent.n.name is s)
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
             is
             0)
            ((1.21267 0.613275))
            ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.5)
             ((0.689626 0.103455))
             ((1.10459 0.305849))))
           ((n.name is eh_3)
            ((R:segstate.parent.nn.ph_cplace is a)
             ((0.608608 -0.10207))
             ((1.08803 0.946333)))
            ((n.name is ae_3)
             ((1.05113 0.518283))
             ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.4)
              ((R:segstate.parent.n.ph_cplace is l)
               ((R:segstate.parent.R:SylStructure.parent.next_accent
                 <
                 0.5)
                ((0.627574 -0.215042))
                ((1.25524 0.558581)))
               ((R:segstate.parent.p.name is r)
                ((R:segstate.parent.p.seg_onsetcoda is coda)
                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                   <
                   4.6)
                  ((1.09339 0.311272))
                  ((0.778659 -0.0781339)))
                 ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                   <
                   2)
                  ((R:segstate.parent.nn.ph_vheight is 3)
                   ((0.232765 -0.587937))
                   ((R:segstate.parent.ph_vrnd is +)
                    ((0.453334 -0.475836))
                    ((R:segstate.parent.R:SylStructure.parent.position_type
                      is
                      initial)
                     ((0.49786 -0.447177))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                       <
                       0.3)
                      ((1.22895 0.0599484))
                      ((R:segstate.parent.ph_vlng is s)
                       ((0.34783 -0.507596))
                       ((R:segstate.parent.n.ph_cvox is +)
                        ((0.536786 -0.0446624))
                        ((0.549515 -0.29519))))))))
                  ((0.416482 -0.642835))))
                ((R:segstate.parent.ph_vlng is s)
                 ((n.name is ah_3)
                  ((0.538097 -0.322429))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    8)
                   ((R:segstate.parent.pp.ph_vc is +)
                    ((0.581909 0.0239231))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                      <
                      0.6)
                     ((R:segstate.parent.name is ah)
                      ((1.14363 0.368791))
                      ((0.597086 -0.0389699)))
                     ((1.28483 0.618198))))
                   ((0.730533 -0.166028))))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                   <
                   0.3)
                  ((0.500512 -0.470785))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                    is
                    0)
                   ((R:segstate.parent.pp.ph_vc is +)
                    ((0.714743 -0.236742))
                    ((0.697404 -0.451799)))
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     1.2)
                    ((0.96236 0.15639))
                    ((0.604921 -0.262043))))))))
              ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 5.6)
               ((R:segstate.parent.ph_vlng is s)
                ((R:segstate.parent.nn.ph_ctype is 0)
                 ((0.223734 -0.560563))
                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                   <
                   9.5)
                  ((R:segstate.parent.ph_vheight is 2)
                   ((0.237992 -0.480451))
                   ((0.213878 -0.387089)))
                  ((0.525024 -0.234774))))
                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                  <
                  4.4)
                 ((R:segstate.parent.ph_vlng is l)
                  ((0.403918 -0.548439))
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    4.9)
                   ((0.58397 -0.241685))
                   ((0.440442 -0.452549))))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                   is
                   0)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    9.6)
                   ((0.644561 0.193655))
                   ((R:segstate.parent.nn.ph_cplace is a)
                    ((0.556781 -0.0797967))
                    ((0.584558 -0.274789))))
                  ((R:segstate.parent.lisp_onset_stop is 0)
                   ((0.460285 -0.403407))
                   ((0.406668 -0.196775))))))
               ((0.474498 -0.718114))))))))))
       ((R:segstate.parent.lisp_onset_nasal is 0)
        ((n.name is p_2)
         ((R:segstate.parent.pp.ph_vc is +)
          ((R:segstate.parent.p.ph_ctype is f)
           ((R:segstate.parent.pp.ph_vheight is 1)
            ((0.207011 -1.08003))
            ((0.325216 -0.890157)))
           ((R:segstate.parent.p.name is m)
            ((R:segstate.parent.R:SylStructure.parent.asyl_out < 1.2)
             ((0.280432 -0.975525))
             ((0.506227 -0.743432)))
            ((R:segstate.parent.pp.ph_vlng is s)
             ((0.606923 -0.53802))
             ((0.673991 0.088002)))))
          ((R:segstate.parent.lisp_coda_stop is 0)
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
             is
             0)
            ((0.103616 -1.15159))
            ((0.348597 -1.00753)))
           ((0.556451 -0.812934))))
         ((n.name is k_2)
          ((R:segstate.parent.p.ph_ctype is f)
           ((R:segstate.parent.R:SylStructure.parent.asyl_out < 2.4)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              1.1)
             ((0.122884 -1.25022))
             ((5.79053e-06 -1.29245)))
            ((1.00133 -0.877903)))
           ((R:segstate.parent.pp.ph_vheight is 3)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              0.7)
             ((0.870073 -0.286571))
             ((1.03964 0.182836)))
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              1.1)
             ((R:segstate.parent.R:SylStructure.parent.last_accent
               <
               0.1)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                is
                0)
               ((4.69388e-06 -1.29245))
               ((R:segstate.parent.p.ph_cvox is +)
                ((0.648968 -0.864949))
                ((0.396407 -1.10245))))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                is
                0)
               ((0.879129 -0.177788))
               ((0.865791 -0.621862))))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
               is
               0)
              ((0.596809 -0.627453))
              ((1.16714 0.249897))))))
          ((R:segstate.parent.name is pau)
           ((n.name is pau_2)
            ((R:segstate.parent.n.ph_cvox is +)
             ((R:segstate.parent.p.name is pau)
              ((R:segstate.parent.n.name is m)
               ((0.730418 1.69401))
               ((R:segstate.parent.n.name is w)
                ((R:segstate.parent.nn.ph_vfront is 1)
                 ((1.37998 1.37878))
                 ((1.47411 0.972538)))
                ((R:segstate.parent.n.name is n)
                 ((R:segstate.parent.nn.ph_vrnd is +)
                  ((1.23287 1.1226))
                  ((1.02815 1.04384)))
                 ((R:segstate.parent.nn.ph_vheight is 2)
                  ((1.08557 0.739675))
                  ((1.04942 0.243468))))))
              ((R:segstate.parent.pp.ph_vfront is 2)
               ((R:segstate.parent.nn.ph_vheight is 3)
                ((1.28929 0.0660248))
                ((0.297523 -0.354977)))
               ((R:segstate.parent.n.name is dh)
                ((1.93876e-06 -0.58585))
                ((R:segstate.parent.p.ph_ctype is s)
                 ((0.153252 -0.520917))
                 ((0.233872 -0.376622))))))
             ((R:segstate.parent.n.ph_ctype is f)
              ((R:segstate.parent.p.name is pau)
               ((R:segstate.parent.n.name is s)
                ((0.56681 0.889725))
                ((R:segstate.parent.n.name is hh)
                 ((R:segstate.parent.nn.ph_vheight is 1)
                  ((0.665379 0.691162))
                  ((0.672445 0.392471)))
                 ((0.749236 0.258277))))
               ((R:segstate.parent.p.ph_cvox is +)
                ((R:segstate.parent.p.name is n)
                 ((0.215077 -0.481959))
                 ((R:segstate.parent.nn.ph_vfront is 1)
                  ((0.30834 -0.389083))
                  ((0.593351 -0.215929))))
                ((R:segstate.parent.n.name is f)
                 ((0.0999703 -0.542561))
                 ((0.262079 -0.422315)))))
              ((R:segstate.parent.p.ph_ctype is n)
               ((R:segstate.parent.nn.ph_cplace is 0)
                ((0.622352 -0.191443))
                ((R:segstate.parent.pp.ph_vlng is a)
                 ((1.56654 0.47472))
                 ((1.22357 0.136621))))
               ((R:segstate.parent.p.name is d)
                ((R:segstate.parent.n.ph_vrnd is -)
                 ((0.473016 -0.315297))
                 ((0.477906 -0.00527407)))
                ((R:segstate.parent.n.name is ae)
                 ((R:segstate.parent.p.ph_ctype is f)
                  ((0.505533 -0.124106))
                  ((R:segstate.parent.p.name is pau)
                   ((R:segstate.parent.pp.ph_cplace is 0)
                    ((0.428534 -0.239542))
                    ((0.585475 -0.221314)))
                   ((0.286299 -0.339437))))
                 ((R:segstate.parent.n.name is ax)
                  ((0.0881316 -0.549776))
                  ((R:segstate.parent.pp.ph_cplace is 0)
                   ((R:segstate.parent.nn.ph_ctype is s)
                    ((0.110352 -0.519252))
                    ((R:segstate.parent.nn.ph_ctype is 0)
                     ((R:segstate.parent.p.name is pau)
                      ((R:segstate.parent.nn.ph_vc is +)
                       ((0.463261 -0.412696))
                       ((0.272915 -0.447327)))
                      ((0.234756 -0.435782)))
                     ((R:segstate.parent.nn.ph_cplace is a)
                      ((R:segstate.parent.pp.ph_vc is 0)
                       ((0.09269 -0.549396))
                       ((0.482638 -0.239541)))
                      ((0.504023 -0.175411)))))
                   ((R:segstate.parent.p.ph_ctype is f)
                    ((0.138033 -0.538082))
                    ((0.310821 -0.412696))))))))))
            ((R:segstate.parent.p.name is pau)
             ((n.name is pau_1)
              ((R:segstate.parent.nn.ph_vheight is 2)
               ((0.583411 -0.673034))
               ((R:segstate.parent.nn.ph_vc is +)
                ((R:segstate.parent.nn.ph_vfront is 2)
                 ((0.638806 -1.01644))
                 ((0.50539 -0.865707)))
                ((R:segstate.parent.nn.ph_ctype is 0)
                 ((0.476457 -1.12272))
                 ((R:segstate.parent.nn.ph_ctype is f)
                  ((0.910901 -1.13277))
                  ((R:segstate.parent.nn.ph_cplace is a)
                   ((0.611891 -0.982686))
                   ((0.605314 -0.978121)))))))
              ((n.name is pau_3)
               ((R:segstate.parent.n.seg_onsetcoda is coda)
                ((R:segstate.parent.nn.ph_cplace is 0)
                 ((R:segstate.parent.pp.ph_vc is 0)
                  ((R:segstate.parent.nn.ph_vfront is 2)
                   ((0.10379 -0.541412))
                   ((0.106709 -0.509712)))
                  ((R:segstate.parent.nn.ph_vheight is 3)
                   ((0.218893 -0.419636))
                   ((0.252568 -0.476399))))
                 ((R:segstate.parent.n.name is ay)
                  ((0.190294 -0.469578))
                  ((R:segstate.parent.pp.ph_vc is 0)
                   ((R:segstate.parent.n.name is pau)
                    ((0.248197 -0.437701))
                    ((0.394666 -0.320232)))
                   ((0.608207 -0.224822)))))
                ((R:segstate.parent.pp.ph_cplace is 0)
                 ((R:segstate.parent.n.ph_ctype is s)
                  ((0.651 0.0830676))
                  ((R:segstate.parent.nn.ph_vrnd is -)
                   ((R:segstate.parent.nn.ph_vfront is 1)
                    ((0.350814 -0.180149))
                    ((0.931938 -0.143345)))
                   ((0.32563 -0.361895))))
                 ((1.14226 0.492428))))
               ((n.name is y_1)
                ((0.712308 -0.944146))
                ((n.name is dh_1)
                 ((R:segstate.parent.nn.ph_vfront is 1)
                  ((0.655237 -0.750744))
                  ((0.670612 -0.640976)))
                 ((R:segstate.parent.pp.ph_vc is -)
                  ((R:segstate.parent.nn.ph_ctype is r)
                   ((0.386295 -0.322786))
                   ((R:segstate.parent.nn.ph_cplace is a)
                    ((R:segstate.parent.pp.ph_cplace is 0)
                     ((0.446242 -0.875204))
                     ((0.520116 -0.518171)))
                    ((R:segstate.parent.nn.ph_vc is +)
                     ((R:segstate.parent.n.ph_ctype is r)
                      ((0.311412 -0.42818))
                      ((R:segstate.parent.n.ph_cvox is -)
                       ((0.351393 -0.465366))
                       ((R:segstate.parent.n.ph_cplace is a)
                        ((0.129249 -0.525433))
                        ((0.429571 -0.837193)))))
                     ((0.501645 -0.40442)))))
                  ((R:segstate.parent.nn.ph_ctype is 0)
                   ((n.name is m_1)
                    ((0.232874 -0.670308))
                    ((n.name is w_1)
                     ((R:segstate.parent.nn.ph_vfront is 1)
                      ((0.265588 -0.524632))
                      ((0.0935088 -0.574428)))
                     ((R:segstate.parent.nn.ph_vfront is 1)
                      ((R:segstate.parent.nn.ph_vheight is 1)
                       ((n.name is hh_1)
                        ((0.301845 -0.650695))
                        ((0.439308 -0.435235)))
                       ((0.31739 -0.259979)))
                      ((R:segstate.parent.n.ph_cplace is a)
                       ((R:segstate.parent.n.ph_cvox is +)
                        ((0.171247 -0.522007))
                        ((0.429119 -0.371834)))
                       ((0.492917 -0.206854))))))
                   ((n.name is ae_1)
                    ((0.52925 -0.400937))
                    ((R:segstate.parent.n.seg_onsetcoda is coda)
                     ((R:segstate.parent.nn.ph_cplace is a)
                      ((0.646441 -0.0245672))
                      ((0.50135 -0.234792)))
                     ((0.405679 -0.307571))))))))))
             ((R:segstate.parent.n.seg_onsetcoda is coda)
              ((R:segstate.parent.p.name is t)
               ((R:segstate.parent.n.ph_vrnd is -)
                ((R:segstate.parent.pp.ph_vrnd is -)
                 ((R:segstate.parent.pp.ph_vheight is 3)
                  ((0.575961 -0.417136))
                  ((0.551559 -0.47506)))
                 ((0.759019 -0.134005)))
                ((0.6653 -0.642015)))
               ((n.name is ae_1)
                ((R:segstate.parent.pp.ph_vc is +)
                 ((R:segstate.parent.pp.ph_vfront is 1)
                  ((0.623439 -0.392122))
                  ((R:segstate.parent.pp.ph_vlng is a)
                   ((0.852986 0.0132319))
                   ((1.45421 0.155389))))
                 ((0.488124 -0.500194)))
                ((R:segstate.parent.nn.ph_vheight is 3)
                 ((n.name is pau_3)
                  ((1.43448 0.395398))
                  ((0.889434 -0.550034)))
                 ((n.name is pau_1)
                  ((R:segstate.parent.pp.ph_vlng is a)
                   ((1.48141 0.110953))
                   ((1.43378 1.06286)))
                  ((R:segstate.parent.n.name is ae)
                   ((1.38024 0.8067))
                   ((R:segstate.parent.pp.ph_vlng is d)
                    ((0.46619 -0.305835))
                    ((R:segstate.parent.nn.ph_cplace is l)
                     ((0.581248 -0.25634))
                     ((n.name is ih_1)
                      ((R:segstate.parent.p.ph_cvox is +)
                       ((0.681284 -0.232125))
                       ((0.720572 0.348555)))
                      ((R:segstate.parent.nn.ph_ctype is n)
                       ((0.822613 0.641049))
                       ((R:segstate.parent.pp.ph_vc is +)
                        ((R:segstate.parent.p.ph_ctype is n)
                         ((1.52507 0.596356))
                         ((R:segstate.parent.p.ph_ctype is f)
                          ((0.838194 0.481352))
                          ((0.658179 -0.108408))))
                        ((0.742736 -0.0558191))))))))))))
              ((R:segstate.parent.n.ph_cvox is +)
               ((n.name is b_1)
                ((R:segstate.parent.p.name is t)
                 ((0.768558 -0.200177))
                 ((R:segstate.parent.nn.ph_vheight is 1)
                  ((0.292943 -0.741652))
                  ((R:segstate.parent.pp.ph_vheight is 2)
                   ((0.326935 -0.62676))
                   ((0.580661 -0.417333)))))
                ((n.name is dh_1)
                 ((R:segstate.parent.pp.ph_vfront is 2)
                  ((0.357502 -0.62846))
                  ((R:segstate.parent.pp.ph_vfront is 1)
                   ((0.502316 -0.538439))
                   ((0.582637 -0.331278))))
                 ((R:segstate.parent.n.ph_ctype is f)
                  ((0.925264 0.312912))
                  ((R:segstate.parent.p.name is t)
                   ((R:segstate.parent.pp.ph_vfront is 1)
                    ((0.625325 -0.461733))
                    ((R:segstate.parent.nn.ph_vfront is 1)
                     ((0.803577 -0.144375))
                     ((0.480653 -0.341421))))
                   ((n.name is w_1)
                    ((R:segstate.parent.p.ph_cvox is +)
                     ((R:segstate.parent.pp.ph_vrnd is -)
                      ((R:segstate.parent.nn.ph_vfront is 2)
                       ((0.609794 -0.288749))
                       ((0.329576 -0.572982)))
                      ((0.541238 -0.127446)))
                     ((R:segstate.parent.nn.ph_vfront is 3)
                      ((0.475272 -0.104271))
                      ((0.711991 0.196516))))
                    ((R:segstate.parent.n.name is g)
                     ((0.903675 -0.379315))
                     ((R:segstate.parent.pp.ph_vheight is 3)
                      ((R:segstate.parent.n.ph_ctype is n)
                       ((0.292747 -0.336066))
                       ((0.784779 -0.0310353)))
                      ((n.name is m_1)
                       ((R:segstate.parent.p.ph_cvox is +)
                        ((0.63607 -0.0633705))
                        ((0.346925 0.0683354)))
                       ((R:segstate.parent.p.ph_ctype is s)
                        ((0.898068 -0.113174))
                        ((R:segstate.parent.n.ph_cplace is a)
                         ((R:segstate.parent.nn.ph_vheight is 1)
                          ((0.867555 0.451691))
                          ((R:segstate.parent.nn.ph_vfront is 2)
                           ((1.80319 0.385902))
                           ((0.899239 -0.120632))))
                         ((0.879223 0.712885))))))))))))
               ((R:segstate.parent.n.ph_cplace is p)
                ((1.39621 0.152895))
                ((R:segstate.parent.n.name is th)
                 ((0.78437 -0.0531552))
                 ((R:segstate.parent.nn.ph_vfront is 3)
                  ((R:segstate.parent.pp.ph_vfront is 1)
                   ((0.33281 -0.622658))
                   ((R:segstate.parent.p.ph_ctype is s)
                    ((1.21892 0.0400131))
                    ((0.370739 -0.421091))))
                  ((R:segstate.parent.n.name is s)
                   ((R:segstate.parent.p.name is d)
                    ((0.63851 -0.474151))
                    ((1.24269 -0.195571)))
                   ((R:segstate.parent.pp.ph_cplace is a)
                    ((R:segstate.parent.n.ph_ctype is f)
                     ((0.164546 -0.596541))
                     ((0.333128 -0.720785)))
                    ((n.name is k_1)
                     ((0.150854 -0.661929))
                     ((R:segstate.parent.pp.ph_vlng is a)
                      ((0.402984 -0.396691))
                      ((R:segstate.parent.nn.ph_vfront is 1)
                       ((0.368634 -0.444693))
                       ((R:segstate.parent.p.ph_cvox is +)
                        ((0.272876 -0.554339))
                        ((0.350852 -0.669623)))))))))))))))
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
             <
             0.3)
            ((n.name is pau_5)
             ((R:segstate.parent.name is ax)
              ((2.13856 3.82223))
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 8.4)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                 <
                 0.3)
                ((0.867109 -0.204287))
                ((R:segstate.parent.nn.ph_vfront is 1)
                 ((1.30236 1.29451))
                 ((1.18495 0.424758))))
               ((0.633416 -0.393201))))
             ((R:segstate.parent.pos_in_syl is 3)
              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                <
                3.1)
               ((R:segstate.parent.p.ph_ctype is s)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                  is
                  0)
                 ((0.386103 -0.39533))
                 ((0.323778 -0.696195)))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                  is
                  0)
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   0.6)
                  ((0.527132 -0.427228))
                  ((0.391679 -0.566387)))
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   0.4)
                  ((0.465683 -0.356641))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                    is
                    0)
                   ((1.0037 -0.169253))
                   ((0.781141 -0.00610319))))))
               ((0.386526 -0.784723)))
              ((R:segstate.parent.name is aw)
               ((R:segstate.parent.nn.ph_ctype is 0)
                ((R:segstate.parent.nn.ph_vheight is 2)
                 ((0.694982 -0.309577))
                 ((R:segstate.parent.nn.ph_vc is +)
                  ((0.890364 0.128393))
                  ((0.80095 0.548485))))
                ((R:segstate.parent.R:SylStructure.parent.position_type
                  is
                  single)
                 ((1.06704 0.241198))
                 ((R:segstate.parent.R:SylStructure.parent.asyl_in
                   <
                   1.2)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    4.5)
                   ((0.156437 -1.16992))
                   ((0.129897 -1.2323)))
                  ((R:segstate.parent.nn.ph_cplace is a)
                   ((0.487558 -1.08053))
                   ((0.95639 -0.603106))))))
               ((R:segstate.parent.lisp_coda_glide is 0)
                ((n.name is r_1)
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   4)
                  ((1.68536 1.83362))
                  ((0.939197 1.27085)))
                 ((n.name is ao_3)
                  ((1.40863 1.29144))
                  ((R:segstate.parent.name is aa)
                   ((R:segstate.parent.p.name is l)
                    ((0.772105 0.114011))
                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                      <
                      1.2)
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                       <
                       1.2)
                      ((1.4644 0.928566))
                      ((1.03159 0.401617)))
                     ((1.64047 1.39603))))
                   ((n.name is aa_1)
                    ((R:segstate.parent.ph_vlng is d)
                     ((1.00068 -0.360656))
                     ((R:segstate.parent.nn.ph_ctype is n)
                      ((1.88401 2.17252))
                      ((1.72943 0.782284))))
                    ((n.name is s_2)
                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                       <
                       1.4)
                      ((R:segstate.parent.pp.ph_cplace is 0)
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         5.8)
                        ((0.512997 -0.6447))
                        ((0.617641 -0.485689)))
                       ((0.36879 -0.894576)))
                      ((0.900897 -0.112499)))
                     ((R:segstate.parent.pp.ph_vlng is a)
                      ((R:segstate.parent.p.ph_ctype is s)
                       ((R:segstate.parent.nn.ph_vheight is 3)
                        ((1.39261 1.01755))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                          <
                          0.3)
                         ((R:segstate.parent.n.ph_vc is +)
                          ((0.822476 -0.231905))
                          ((0.990868 0.203403)))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           1.6)
                          ((1.19002 0.739948))
                          ((R:segstate.parent.n.ph_ctype is f)
                           ((1.11794 0.635882))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             2.6)
                            ((0.854022 0.121432))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              4.5)
                             ((1.40111 0.475739))
                             ((0.940633 0.178511))))))))
                       ((n.name is ay_3)
                        ((0.942937 0.71796))
                        ((R:segstate.parent.n.ph_ctype is r)
                         ((0.743993 -0.399018))
                         ((R:segstate.parent.name is iy)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                            <
                            0.1)
                           ((0.742922 0.15586))
                           ((0.775226 -0.197058)))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                            <
                            0.3)
                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                             <
                             3.2)
                            ((R:segstate.parent.R:SylStructure.parent.position_type
                              is
                              initial)
                             ((1.12682 0.479587))
                             ((R:segstate.parent.ph_vheight is 3)
                              ((0.705673 -0.214285))
                              ((R:segstate.parent.n.ph_cplace is a)
                               ((1.0595 0.19887))
                               ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                 <
                                 1.2)
                                ((0.726181 0.15717))
                                ((0.478983 -0.213927))))))
                            ((1.41595 0.528027)))
                           ((R:segstate.parent.ph_vheight is 3)
                            ((1.45133 0.903207))
                            ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                              <
                              2.2)
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                               <
                               6)
                              ((R:segstate.parent.R:SylStructure.parent.syl_out
                                <
                                5.4)
                               ((R:segstate.parent.nn.ph_ctype is 0)
                                ((0.552102 0.12031))
                                ((1.69248 0.580223)))
                               ((0.92208 -0.0395185)))
                              ((1.52807 0.639434)))
                             ((0.755528 -0.0214409)))))))))
                      ((R:segstate.parent.name is ax)
                       ((n.name is ax_3)
                        ((R:segstate.parent.p.name is dh)
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           6)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((2.98174e-06 -0.510848))
                           ((4.46802e-06 -0.510845)))
                          ((0.146846 -0.434565)))
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           4.4)
                          ((3.17492 1.18913))
                          ((R:segstate.parent.p.ph_ctype is s)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                             is
                             0)
                            ((0.384876 -0.304889))
                            ((0.407536 -0.213352)))
                           ((0.81885 0.093954)))))
                        ((n.name is ax_2)
                         ((R:segstate.parent.n.ph_cplace is a)
                          ((R:segstate.parent.nn.ph_vheight is 2)
                           ((0.374111 -0.291227))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                             is
                             0)
                            ((1.35692 0.94457))
                            ((0.820322 0.106537))))
                          ((R:segstate.parent.pp.ph_cplace is 0)
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             1.8)
                            ((1.91849e-06 -0.500363))
                            ((0.1602 -0.422906)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                             is
                             0)
                            ((1.3284 0.00156254))
                            ((0.1379 -0.395796)))))
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           final)
                          ((R:segstate.parent.p.ph_ctype is s)
                           ((1.20664 0.370208))
                           ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                             <
                             2.3)
                            ((0.430285 -0.225932))
                            ((0.18734 -0.376094))))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((1.94605 1.55535))
                           ((1.81376 3.27543))))))
                       ((n.name is ih_3)
                        ((R:segstate.parent.R:SylStructure.parent.next_accent
                          <
                          0.8)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                           <
                           0.3)
                          ((0.426188 -0.427535))
                          ((0.57163 0.0355492)))
                         ((R:segstate.parent.p.name is pau)
                          ((0.494848 -0.163772))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            4)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                             is
                             0)
                            ((1.04939 0.361595))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                              <
                              0.1)
                             ((1.17621 0.707187))
                             ((1.87146 2.02198))))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             2.4)
                            ((1.16288 0.582003))
                            ((0.678033 -0.156348))))))
                        ((R:segstate.parent.n.name is pau)
                         ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                           <
                           1.2)
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            0.5)
                           ((0.762323 -0.0153108))
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             4.2)
                            ((1.28543 0.896484))
                            ((R:segstate.parent.R:SylStructure.parent.next_accent
                              <
                              1.4)
                             ((R:segstate.parent.lisp_onset_stop is 0)
                              ((0.678884 -0.101386))
                              ((1.18768 0.569081)))
                             ((R:segstate.parent.lisp_onset_stop is 0)
                              ((1.35694 0.928133))
                              ((1.10407 0.377453))))))
                          ((0.67201 -0.12664)))
                         ((n.name is t_1)
                          ((R:segstate.parent.lisp_onset_stop is 0)
                           ((R:segstate.parent.pp.ph_vc is +)
                            ((0.513399 -0.344124))
                            ((R:segstate.parent.ph_vheight is 1)
                             ((0.573846 -0.129137))
                             ((1.04341 0.306835))))
                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                             <
                             0.5)
                            ((2.14779 2.10076))
                            ((0.666092 0.224042))))
                          ((n.name is er_3)
                           ((R:segstate.parent.R:SylStructure.parent.asyl_in
                             <
                             0.8)
                            ((0.759873 -0.461235))
                            ((R:segstate.parent.R:SylStructure.parent.position_type
                              is
                              final)
                             ((1.29342 0.927838))
                             ((R:segstate.parent.n.name is iy)
                              ((0.930868 0.851857))
                              ((R:segstate.parent.p.ph_cvox is +)
                               ((1.21045 0.493159))
                               ((1.04726 -0.100549))))))
                           ((n.name is t_2)
                            ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                              <
                              2.2)
                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                               <
                               1.6)
                              ((0.692803 -0.0259509))
                              ((R:segstate.parent.pp.ph_vc is +)
                               ((0.53833 -0.226665))
                               ((0.414916 -0.505332))))
                             ((0.338227 -0.624878)))
                            ((R:segstate.parent.p.name is l)
                             ((R:segstate.parent.n.ph_ctype is s)
                              ((R:segstate.parent.name is ay)
                               ((0.301348 -0.670669))
                               ((R:segstate.parent.nn.ph_vrnd is -)
                                ((0.466455 -0.540421))
                                ((0.554898 -0.305066))))
                              ((R:segstate.parent.ph_vheight is 2)
                               ((R:segstate.parent.pp.ph_vheight is 3)
                                ((0.176836 -0.746721))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                  <
                                  1.2)
                                 ((1.05073 0.0155218))
                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                   <
                                   6)
                                  ((0.776702 -0.47996))
                                  ((0.754606 -0.321398)))))
                               ((n.name is iy_3)
                                ((0.618793 -0.423081))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                  is
                                  0)
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   3)
                                  ((0.546674 -0.293184))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    3.6)
                                   ((0.869714 0.364848))
                                   ((0.714836 -0.146637))))
                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                   <
                                   1.2)
                                  ((1.51566 0.519188))
                                  ((R:segstate.parent.syl_final is 0)
                                   ((0.766473 -0.0633643))
                                   ((0.579917 0.288818))))))))
                             ((n.name is ay_3)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                is
                                0)
                               ((0.839138 0.93732))
                               ((1.15962 0.221256)))
                              ((R:segstate.parent.pp.ph_vc is 0)
                               ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                 <
                                 0.3)
                                ((1.1532 0.803505))
                                ((1.04431 0.022583)))
                               ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                 <
                                 0.3)
                                ((R:segstate.parent.ph_vheight is 1)
                                 ((R:segstate.parent.pp.ph_cplace is 0)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                    <
                                    0.3)
                                   ((R:segstate.parent.n.ph_ctype is f)
                                    ((0.506148 -0.294383))
                                    ((0.974048 0.123639)))
                                   ((1.11235 0.361303)))
                                  ((1.8226 0.758427)))
                                 ((n.name is ae_3)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                    <
                                    0.3)
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                     <
                                     4.3)
                                    ((0.241485 -0.655127))
                                    ((0.0805811 -0.809982)))
                                   ((0.710307 -0.481206)))
                                  ((R:segstate.parent.p.name is n)
                                   ((1.0697 0.44638))
                                   ((n.name is ah_3)
                                    ((0.365166 -0.652764))
                                    ((R:segstate.parent.pp.ph_vheight
                                      is
                                      2)
                                     ((0.295622 -0.485725))
                                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                                       <
                                       1.6)
                                      ((R:segstate.parent.ph_vheight
                                        is
                                        2)
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                         is
                                         0)
                                        ((1.06274 0.121198))
                                        ((0.534667 -0.250561)))
                                       ((R:segstate.parent.nn.ph_cplace
                                         is
                                         a)
                                        ((n.name is n_1)
                                         ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                           <
                                           2.5)
                                          ((0.394415 -0.20344))
                                          ((0.14528 -0.428893)))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           4.8)
                                          ((0.791476 0.125797))
                                          ((0.40634 -0.235683))))
                                        ((R:segstate.parent.n.name
                                          is
                                          n)
                                         ((0.397983 -0.310952))
                                         ((0.465669 -0.650055)))))
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                        is
                                        0)
                                       ((0.566226 -0.215694))
                                       ((1.05382 0.302665)))))))))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                  <
                                  0.3)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                   is
                                   0)
                                  ((n.name is iy_3)
                                   ((0.754729 0.627687))
                                   ((R:segstate.parent.p.name is b)
                                    ((0.411692 -0.554208))
                                    ((R:segstate.parent.name is ih)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                       <
                                       0.3)
                                      ((0.458292 -0.26096))
                                      ((0.17823 -0.48492)))
                                     ((n.name is iy_2)
                                      ((0.950853 -0.443871))
                                      ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                        <
                                        1.2)
                                       ((R:segstate.parent.pp.ph_vc
                                         is
                                         +)
                                        ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                                          <
                                          0.1)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                                           <
                                           4.9)
                                          ((0.77877 0.229416))
                                          ((1.11747 0.764408)))
                                         ((0.830075 0.0760711)))
                                        ((R:segstate.parent.p.name
                                          is
                                          t)
                                         ((0.763533 -0.215589))
                                         ((R:segstate.parent.R:SylStructure.parent.last_accent
                                           <
                                           1.5)
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            a)
                                           ((R:segstate.parent.R:SylStructure.parent.position_type
                                             is
                                             single)
                                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                              <
                                              6.5)
                                             ((0.713321 -0.118207))
                                             ((0.851748 0.0919167)))
                                            ((0.513063 -0.330341)))
                                           ((R:segstate.parent.lisp_coda_nasal
                                             is
                                             0)
                                            ((R:segstate.parent.name
                                              is
                                              hh)
                                             ((0.775907 0.0106104))
                                             ((R:segstate.parent.p.ph_ctype
                                               is
                                               f)
                                              ((1.37566 0.253665))
                                              ((0.864026 0.633877))))
                                            ((0.815644 -0.0795198))))
                                          ((1.098 0.50744)))))
                                       ((0.806922 -0.19706)))))))
                                  ((R:segstate.parent.n.ph_ctype is r)
                                   ((0.727115 0.284418))
                                   ((R:segstate.parent.n.name is iy)
                                    ((n.name is iy_1)
                                     ((0.375598 -0.313533))
                                     ((0.337677 -0.850208)))
                                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                                      <
                                      2.7)
                                     ((n.name is er_2)
                                      ((R:segstate.parent.pp.ph_vfront
                                        is
                                        1)
                                       ((1.14272 0.447871))
                                       ((1.04817 -0.131704)))
                                      ((n.name is iy_2)
                                       ((0.542816 -0.470971))
                                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                         <
                                         3.6)
                                        ((R:segstate.parent.ph_vheight
                                          is
                                          2)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                           <
                                           5)
                                          ((0.307921 -0.433777))
                                          ((0.344461 -0.251133)))
                                         ((R:segstate.parent.p.name
                                           is
                                           d)
                                          ((0.440908 0.265881))
                                          ((R:segstate.parent.lisp_coda_stop
                                            is
                                            0)
                                           ((R:segstate.parent.nn.ph_vc
                                             is
                                             +)
                                            ((R:segstate.parent.p.name
                                              is
                                              t)
                                             ((0.61141 -0.0758434))
                                             ((0.834346 -0.298397)))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                              <
                                              0.3)
                                             ((0.744431 -0.240217))
                                             ((R:segstate.parent.name
                                               is
                                               iy)
                                              ((0.599951 -0.0300956))
                                              ((1.18982 0.219682)))))
                                           ((0.296293 -0.397682)))))
                                        ((R:segstate.parent.lisp_onset_stop
                                          is
                                          0)
                                         ((R:segstate.parent.p.ph_cvox
                                           is
                                           +)
                                          ((0.746571 -0.361844))
                                          ((0.641345 -0.193036)))
                                         ((0.444489 -0.395259))))))
                                     ((R:segstate.parent.name is iy)
                                      ((0.818809 -0.132284))
                                      ((R:segstate.parent.pp.ph_vlng
                                        is
                                        s)
                                       ((R:segstate.parent.p.ph_ctype
                                         is
                                         s)
                                        ((0.307322 -0.572603))
                                        ((0.361967 -0.701883)))
                                       ((0.483679 -0.335671))))))))
                                 ((n.name is ax_1)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    3.1)
                                   ((1.28356 1.04367))
                                   ((0.853405 0.636577)))
                                  ((n.name is er_2)
                                   ((0.985727 0.800249))
                                   ((R:segstate.parent.p.ph_ctype is f)
                                    ((R:segstate.parent.pp.ph_vc is +)
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       4.8)
                                      ((R:segstate.parent.nn.ph_ctype
                                        is
                                        0)
                                       ((R:segstate.parent.pp.ph_vheight
                                         is
                                         3)
                                        ((0.75474 0.262206))
                                        ((1.5713 0.94459)))
                                       ((0.730813 0.0883239)))
                                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                        <
                                        2.8)
                                       ((0.629514 -0.403363))
                                       ((R:segstate.parent.R:SylStructure.parent.last_accent
                                         <
                                         1.2)
                                        ((R:segstate.parent.p.name
                                          is
                                          s)
                                         ((1.05863 0.219385))
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                           <
                                           0.1)
                                          ((0.309909 -0.347653))
                                          ((1.09758 0.0308713))))
                                        ((1.05498 0.328148)))))
                                     ((n.name is iy_2)
                                      ((0.364666 -0.566788))
                                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                                        <
                                        8)
                                       ((R:segstate.parent.n.name is n)
                                        ((1.13324 0.406467))
                                        ((n.name is ae_3)
                                         ((0.895573 0.396144))
                                         ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                           <
                                           2.1)
                                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                            <
                                            1.2)
                                           ((R:segstate.parent.nn.ph_vrnd
                                             is
                                             -)
                                            ((0.757587 0.397404))
                                            ((0.872672 -0.0147174)))
                                           ((R:segstate.parent.n.ph_cplace
                                             is
                                             a)
                                            ((0.652349 -0.326744))
                                            ((R:segstate.parent.pp.ph_cplace
                                              is
                                              a)
                                             ((0.975456 0.154863))
                                             ((0.684452 -0.240326)))))
                                          ((R:segstate.parent.p.ph_cvox
                                            is
                                            +)
                                           ((0.596883 -0.528718))
                                           ((R:segstate.parent.nn.ph_ctype
                                             is
                                             0)
                                            ((R:segstate.parent.pp.ph_cplace
                                              is
                                              a)
                                             ((0.712291 0.062402))
                                             ((0.807051 -0.273457)))
                                            ((0.578435 -0.339312)))))))
                                       ((R:segstate.parent.ph_vheight
                                         is
                                         2)
                                        ((0.85113 0.0478203))
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                          is
                                          0)
                                         ((0.404449 -0.690119))
                                         ((R:segstate.parent.n.ph_cplace
                                           is
                                           a)
                                          ((0.354289 -0.524788))
                                          ((0.948737 -0.0576118))))))))
                                    ((R:segstate.parent.name is ah)
                                     ((1.35795 0.521082))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                       <
                                       0.3)
                                      ((R:segstate.parent.ph_vlng is s)
                                       ((R:segstate.parent.n.ph_cvox
                                         is
                                         -)
                                        ((1.00196 0.133979))
                                        ((1.88055 1.3419)))
                                       ((R:segstate.parent.p.name is d)
                                        ((1.56835 0.79582))
                                        ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                          <
                                          1.2)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                           <
                                           7)
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            a)
                                           ((1.02017 0.234296))
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                             <
                                             0.3)
                                            ((1.51704 0.223812))
                                            ((R:segstate.parent.nn.ph_vfront
                                              is
                                              1)
                                             ((0.686633 -0.161156))
                                             ((0.538149 -0.388202)))))
                                          ((R:segstate.parent.name
                                            is
                                            hh)
                                           ((1.87927 0.863549))
                                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                             <
                                             6.4)
                                            ((1.01899 0.362856))
                                            ((0.631913 -0.225878)))))
                                         ((0.718371 -0.385705)))))
                                      ((R:segstate.parent.n.ph_ctype
                                        is
                                        s)
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                         is
                                         0)
                                        ((0.509373 -0.390894))
                                        ((0.719443 -0.101111)))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                         <
                                         1.2)
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                          <
                                          7)
                                         ((R:segstate.parent.nn.ph_vrnd
                                           is
                                           -)
                                          ((R:segstate.parent.pos_in_syl
                                            is
                                            0)
                                           ((R:segstate.parent.nn.ph_vheight
                                             is
                                             2)
                                            ((1.49263 0.373426))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                              is
                                              0)
                                             ((0.697075 -0.300747))
                                             ((0.919364 0.130816))))
                                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                                             <
                                             8)
                                            ((0.783906 -0.211606))
                                            ((0.472074 -0.505806))))
                                          ((R:segstate.parent.n.ph_ctype
                                            is
                                            0)
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                             <
                                             0.3)
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                              <
                                              0.1)
                                             ((1.78512 0.34202))
                                             ((0.643408 0.0626632)))
                                            ((R:segstate.parent.name
                                              is
                                              hh)
                                             ((1.14227 0.183464))
                                             ((R:segstate.parent.pp.ph_cplace
                                               is
                                               0)
                                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                                <
                                                3.4)
                                               ((0.595738 -0.21197))
                                               ((0.566654 -0.381205)))
                                              ((R:segstate.parent.R:SylStructure.parent.position_type
                                                is
                                                initial)
                                               ((0.58577 0.0266974))
                                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                                 <
                                                 4.5)
                                                ((0.781436 -0.108218))
                                                ((0.693949 -0.268939)))))))
                                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                             <
                                             4.5)
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                              is
                                              0)
                                             ((1.04184 0.545347))
                                             ((1.09098 0.84363)))
                                            ((1.10944 0.17484)))))
                                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                           <
                                           1.2)
                                          ((0.598788 -0.238075))
                                          ((0.497472 -0.440546))))
                                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                          <
                                          3)
                                         ((0.892042 0.111612))
                                         ((0.980264 0.552362)))))))))))))))))))))))))))
                ((R:segstate.parent.n.ph_vc is +)
                 ((R:segstate.parent.ph_cplace is a)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                    <
                    0.1)
                   ((0.446973 -0.275155))
                   ((0.663855 -0.0490051)))
                  ((R:segstate.parent.pp.ph_cplace is 0)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                     is
                     0)
                    ((R:segstate.parent.pp.ph_vrnd is -)
                     ((0.718088 -0.203171))
                     ((1.21268 0.264896)))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                      is
                      0)
                     ((0.997034 0.043452))
                     ((1.42409 0.718353))))
                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                     <
                     4.2)
                    ((0.929607 0.0936774))
                    ((0.909546 -0.33019)))))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                   is
                   0)
                  ((n.name is r_1)
                   ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                     <
                     3)
                    ((0.195888 -0.535311))
                    ((0.637836 -0.277441)))
                   ((R:segstate.parent.ph_vlng is a)
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                      is
                      0)
                     ((0.332499 -0.35999))
                     ((0.0897397 -0.474502)))
                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                      <
                      1.6)
                     ((R:segstate.parent.nn.ph_vfront is 2)
                      ((0.361413 -0.530062))
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        4.5)
                       ((R:segstate.parent.R:SylStructure.parent.last_accent
                         <
                         0.4)
                        ((0.972117 0.177343))
                        ((R:segstate.parent.R:SylStructure.parent.syl_out
                          <
                          5.2)
                         ((0.718503 -0.451463))
                         ((0.668603 -0.2174))))
                       ((1.13684 0.433575))))
                     ((R:segstate.parent.lisp_onset_stop is 0)
                      ((R:segstate.parent.n.name is l)
                       ((1.04152 0.180867))
                       ((1.095 0.607527)))
                      ((0.725148 -0.136385))))))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    2.7)
                   ((R:segstate.parent.nn.ph_cplace is a)
                    ((0.825038 0.0530811))
                    ((R:segstate.parent.ph_vlng is l)
                     ((0.268441 -0.597504))
                     ((R:segstate.parent.p.ph_cvox is +)
                      ((0.758251 -0.0860974))
                      ((0.425271 -0.393737)))))
                   ((R:segstate.parent.nn.ph_ctype is s)
                    ((0.250108 -0.753888))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      2.2)
                     ((0.485389 -0.255338))
                     ((0.450259 -0.520597)))))))))))
            ((R:segstate.parent.pos_in_syl is 0)
             ((n.name is ih_2)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                <
                1.2)
               ((R:segstate.parent.R:SylStructure.parent.syl_out < 6.7)
                ((0.734103 0.73936))
                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                  <
                  7.5)
                 ((1.36354 2.11105))
                 ((0.862008 1.07835))))
               ((R:segstate.parent.pp.ph_vfront is 2)
                ((1.9236 0.927079))
                ((R:segstate.parent.p.name is pau)
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   2.4)
                  ((R:segstate.parent.nn.ph_ctype is f)
                   ((0.412802 -0.158938))
                   ((0.913287 0.531835)))
                  ((1.57376 0.960974)))
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   2.1)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    2)
                   ((1.14072 0.372577))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                     <
                     0.3)
                    ((0.275786 -0.374935))
                    ((R:segstate.parent.nn.ph_ctype is f)
                     ((0.784928 0.195619))
                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                       <
                       1.2)
                      ((0.29571 -0.247579))
                      ((0.489269 -0.0453085))))))
                  ((1.70651 0.623531))))))
              ((n.name is r_1)
               ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                 <
                 0.2)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                  <
                  0.3)
                 ((0.303424 -0.539997))
                 ((1.05758 -0.00450486)))
                ((R:segstate.parent.name is p)
                 ((0.53863 0.365909))
                 ((R:segstate.parent.name is f)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    3.2)
                   ((1.19089 0.202689))
                   ((1.78787 1.28582)))
                  ((R:segstate.parent.R:SylStructure.parent.next_accent
                    <
                    1)
                   ((1.08304 1.35585))
                   ((1.42283 2.06934))))))
               ((n.name is ae_3)
                ((R:segstate.parent.lisp_coda_nasal is 0)
                 ((R:segstate.parent.n.ph_cplace is a)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    0.9)
                   ((0.591568 0.16277))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                     <
                     0.3)
                    ((0.359503 -0.66058))
                    ((0.631241 -0.371589))))
                  ((0.740491 0.425406)))
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   0.7)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                    is
                    0)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                     is
                     0)
                    ((0.100559 -0.751457))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                      <
                      1.2)
                     ((0.0655374 -0.783264))
                     ((0.0500666 -0.813575))))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                     <
                     0.1)
                    ((0.288599 -0.619138))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      3.1)
                     ((0.0939842 -0.771813))
                     ((0.142964 -0.706926)))))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    4)
                   ((0.217833 -0.726011))
                   ((0.259433 -0.582879)))))
                ((n.name is ch_2)
                 ((R:segstate.parent.pp.ph_vc is +)
                  ((0.601341 -0.918556))
                  ((0.340136 -1.16184)))
                 ((n.name is ah_3)
                  ((R:segstate.parent.n.name is v)
                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                     <
                     0.6)
                    ((0.741996 -0.339701))
                    ((R:segstate.parent.p.ph_ctype is f)
                     ((0.177271 -0.798315))
                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                       <
                       0.9)
                      ((0.164049 -0.788189))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                        is
                        0)
                       ((0.186951 -0.71087))
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                         <
                         3)
                        ((0.251188 -0.679194))
                        ((0.713453 -0.281928)))))))
                   ((0.878027 0.256108)))
                  ((n.name is iy_1)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                     <
                     0.3)
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      5)
                     ((0.403909 -0.572876))
                     ((R:segstate.parent.R:SylStructure.parent.asyl_out
                       <
                       3.4)
                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                        <
                        7.7)
                       ((0.200827 -0.711407))
                       ((0.244694 -0.838702)))
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                        <
                        7.1)
                       ((0.268758 -0.604211))
                       ((0.0584998 -0.707561)))))
                    ((R:segstate.parent.p.ph_ctype is f)
                     ((0.305532 -0.511839))
                     ((R:segstate.parent.p.ph_cvox is +)
                      ((0.994217 0.605904))
                      ((R:segstate.parent.R:SylStructure.parent.asyl_in
                        <
                        1.2)
                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                         <
                         0.7)
                        ((0.571919 -0.287957))
                        ((0.516751 -0.623975)))
                       ((R:segstate.parent.pp.ph_cplace is a)
                        ((0.759235 -0.230156))
                        ((1.81543 0.450206)))))))
                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                     <
                     2.4)
                    ((n.name is hh_2)
                     ((R:segstate.parent.p.name is pau)
                      ((R:segstate.parent.n.name is iy)
                       ((R:segstate.parent.nn.ph_cplace is a)
                        ((0.900751 1.00648))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                          <
                          0.3)
                         ((1.24783 1.98609))
                         ((R:segstate.parent.pp.ph_vc is -)
                          ((1.03535 1.12657))
                          ((R:segstate.parent.R:SylStructure.parent.asyl_out
                            <
                            3.2)
                           ((1.14004 1.57954))
                           ((0.234197 1.5066))))))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                         <
                         0.4)
                        ((1.06931 0.648655))
                        ((0.552722 -0.359002))))
                      ((R:segstate.parent.n.name is iy)
                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                         <
                         2.5)
                        ((0.982871 0.183466))
                        ((0.699752 -0.151713)))
                       ((R:segstate.parent.p.ph_ctype is f)
                        ((R:segstate.parent.R:SylStructure.parent.last_accent
                          <
                          1.4)
                         ((0.26007 -0.157471))
                         ((0.198485 -0.320616)))
                        ((R:segstate.parent.pp.ph_vheight is 1)
                         ((R:segstate.parent.n.name is ih)
                          ((0.0607817 -0.566293))
                          ((0.10868 -0.51375)))
                         ((R:segstate.parent.pp.ph_cplace is a)
                          ((0.0961784 -0.534047))
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            5)
                           ((0.284652 -0.326457))
                           ((0.161082 -0.502954))))))))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                       <
                       0.3)
                      ((n.name is f_2)
                       ((R:segstate.parent.n.name is ih)
                        ((1.46008 1.68867))
                        ((1.38597 1.04008)))
                       ((n.name is p_3)
                        ((0.528378 -1.01033))
                        ((n.name is er_1)
                         ((1.0485 0.585394))
                         ((R:segstate.parent.ph_vheight is 2)
                          ((R:segstate.parent.n.ph_ctype is n)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                             is
                             0)
                            ((1.16299 0.549951))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              5.2)
                             ((1.1351 0.175523))
                             ((1.31025 0.366449))))
                           ((n.name is ax_2)
                            ((1.06743 0.398143))
                            ((R:segstate.parent.n.ph_cplace is l)
                             ((0.846294 0.191392))
                             ((R:segstate.parent.n.ph_ctype is f)
                              ((0.58525 -0.0537677))
                              ((0.70914 -0.325921))))))
                          ((n.name is ae_2)
                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                             <
                             0.7)
                            ((0.693259 0.553124))
                            ((0.611721 0.177058)))
                           ((R:segstate.parent.ph_vheight is 3)
                            ((n.name is ay_3)
                             ((R:segstate.parent.n.ph_cvox is +)
                              ((R:segstate.parent.nn.ph_vfront is 3)
                               ((0.550897 0.164334))
                               ((0.520008 -0.209392)))
                              ((0.442129 -0.498778)))
                             ((R:segstate.parent.nn.ph_vheight is 3)
                              ((n.name is ay_2)
                               ((0.212331 -0.317041))
                               ((0.585237 -0.963158)))
                              ((R:segstate.parent.nn.ph_ctype is s)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 5.8)
                                ((0.31586 -0.3912))
                                ((0.444614 -0.60911)))
                               ((n.name is ay_2)
                                ((0.213286 -0.432122))
                                ((R:segstate.parent.n.name is l)
                                 ((0.333929 -0.293574))
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                   <
                                   4.5)
                                  ((0.587673 -0.185028))
                                  ((0.399522 0.0471145))))))))
                            ((R:segstate.parent.name is hh)
                             ((R:segstate.parent.pp.ph_vc is -)
                              ((n.name is hh_3)
                               ((R:segstate.parent.n.name is iy)
                                ((0.144774 -0.449688))
                                ((0.972298 0.413531)))
                               ((0.124903 -0.631248)))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                7)
                               ((R:segstate.parent.nn.ph_cplace is l)
                                ((3.38577e-07 -0.537715))
                                ((0.122035 -0.50766)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                 is
                                 0)
                                ((0.0863662 -0.55429))
                                ((2.71465e-07 -0.537715)))))
                             ((R:segstate.parent.pp.ph_vrnd is -)
                              ((R:segstate.parent.n.ph_ctype is 0)
                               ((R:segstate.parent.n.name is ax)
                                ((R:segstate.parent.R:SylStructure.parent.syl_in
                                  <
                                  3)
                                 ((R:segstate.parent.p.ph_cvox is +)
                                  ((1.40635 -0.0585923))
                                  ((0.761153 -0.188444)))
                                 ((n.name is ax_1)
                                  ((0.417283 -0.342232))
                                  ((0.692621 -0.708051))))
                                ((n.name is t_2)
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   1.2)
                                  ((0.202514 -0.596187))
                                  ((0.486633 -0.251894)))
                                 ((R:segstate.parent.name is t)
                                  ((0.861647 0.7017))
                                  ((n.name is s_2)
                                   ((1.66434 0.353052))
                                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                     <
                                     0.2)
                                    ((0.899099 -0.0248406))
                                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                                      <
                                      2.1)
                                     ((R:segstate.parent.p.name is n)
                                      ((0.658933 -0.625058))
                                      ((0.590836 -0.268014)))
                                     ((0.411373 -0.0902189))))))))
                               ((R:segstate.parent.n.ph_cplace is a)
                                ((0.668162 -0.173151))
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  4.6)
                                 ((0.621481 -0.425331))
                                 ((R:segstate.parent.name is s)
                                  ((0.603355 -0.589336))
                                  ((0.426297 -0.90944))))))
                              ((n.name is s_3)
                               ((0.816194 0.635655))
                               ((R:segstate.parent.n.name is iy)
                                ((1.40068 0.777608))
                                ((n.name is s_2)
                                 ((0.331887 -0.561409))
                                 ((n.name is f_3)
                                  ((0.653986 -0.451234))
                                  ((R:segstate.parent.nn.ph_ctype is l)
                                   ((1.25947 0.537313))
                                   ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                     <
                                     2.2)
                                    ((n.name is t_1)
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       10.5)
                                      ((0.471521 -0.416082))
                                      ((0.628465 -0.143017)))
                                     ((R:segstate.parent.nn.ph_ctype
                                       is
                                       r)
                                      ((R:segstate.parent.n.name is t)
                                       ((0.794152 0.643403))
                                       ((0.79071 -0.0212943)))
                                      ((R:segstate.parent.n.ph_ctype
                                        is
                                        s)
                                       ((0.495132 -0.529334))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                         is
                                         0)
                                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                          <
                                          0.1)
                                         ((0.681781 -0.121701))
                                         ((0.993254 0.365225)))
                                        ((R:segstate.parent.lisp_onset_stop
                                          is
                                          0)
                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                           <
                                           11.2)
                                          ((0.353051 -0.39658))
                                          ((0.520199 -0.06596)))
                                         ((0.774547 0.0625222)))))))
                                    ((0.449863 -0.433929))))))))))))))))
                      ((n.name is s_2)
                       ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                         <
                         0.2)
                        ((R:segstate.parent.p.ph_ctype is f)
                         ((0.452913 -0.930131))
                         ((R:segstate.parent.R:SylStructure.parent.asyl_in
                           <
                           0.2)
                          ((1.17891 0.0955708))
                          ((R:segstate.parent.p.ph_cvox is +)
                           ((0.749479 -0.303961))
                           ((0.481878 -0.826428)))))
                        ((1.27785 0.267525)))
                       ((n.name is sh_3)
                        ((R:segstate.parent.p.ph_cvox is +)
                         ((1.18661 1.0754))
                         ((0.685869 0.574957)))
                        ((R:segstate.parent.name is k)
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           3.2)
                          ((n.name is k_3)
                           ((R:segstate.parent.nn.ph_cplace is a)
                            ((0.583655 0.127584))
                            ((0.799664 -0.411109)))
                           ((R:segstate.parent.R:SylStructure.parent.asyl_out
                             <
                             2.5)
                            ((0.370399 0.444069))
                            ((0.660683 0.547456))))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                            <
                            0.3)
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             7.6)
                            ((0.476312 0.425296))
                            ((0.760423 -0.0285473)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                             is
                             0)
                            ((1.85389 1.2558))
                            ((n.name is k_3)
                             ((0.710773 0.153618))
                             ((0.579997 0.807468))))))
                         ((n.name is w_1)
                          ((1.55925 1.13547))
                          ((n.name is f_2)
                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                             <
                             3)
                            ((R:segstate.parent.p.ph_ctype is s)
                             ((0.460139 -0.473929))
                             ((R:segstate.parent.n.ph_vrnd is -)
                              ((0.918177 0.25939))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                is
                                0)
                               ((0.401252 -0.704884))
                               ((0.908115 0.0667202)))))
                            ((0.2596 -0.763936)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                             <
                             2.2)
                            ((n.name is v_1)
                             ((R:segstate.parent.p.ph_ctype is n)
                              ((0.662077 0.0603917))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((0.558612 -0.23708))
                               ((R:segstate.parent.p.ph_cvox is +)
                                ((0.14068 -0.695932))
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  2.5)
                                 ((0.314685 -0.49144))
                                 ((0.574469 -0.133518))))))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                               is
                               0)
                              ((n.name is eh_2)
                               ((2.30653 1.10867))
                               ((R:segstate.parent.n.ph_ctype is f)
                                ((R:segstate.parent.ph_vrnd is +)
                                 ((0.695955 0.0188445))
                                 ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                   <
                                   1.1)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                    <
                                    0.1)
                                   ((1.29709 0.680146))
                                   ((1.20743 1.20097)))
                                  ((R:segstate.parent.pp.ph_vfront
                                    is
                                    1)
                                   ((0.774813 -0.0442622))
                                   ((R:segstate.parent.R:SylStructure.parent.last_accent
                                     <
                                     2)
                                    ((1.01552 0.425132))
                                    ((1.40917 0.738445))))))
                                ((n.name is s_3)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_in
                                   <
                                   3.9)
                                  ((0.629363 0.780535))
                                  ((0.62438 0.344337)))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   5.4)
                                  ((R:segstate.parent.name is p)
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     7)
                                    ((0.779884 0.162467))
                                    ((0.822956 0.751761)))
                                   ((R:segstate.parent.name is hh)
                                    ((R:segstate.parent.nn.ph_ctype
                                      is
                                      f)
                                     ((1.11893 -0.0500376))
                                     ((R:segstate.parent.pp.ph_vheight
                                       is
                                       2)
                                      ((1.13211 0.147791))
                                      ((R:segstate.parent.p.ph_cvox
                                        is
                                        +)
                                       ((1.35565 0.511096))
                                       ((1.52315 0.939656)))))
                                    ((R:segstate.parent.nn.ph_ctype
                                      is
                                      n)
                                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                                       <
                                       1.5)
                                      ((0.551167 -0.4811))
                                      ((0.74881 -0.0181271)))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                       is
                                       0)
                                      ((R:segstate.parent.name is sh)
                                       ((0.947868 -0.247492))
                                       ((R:segstate.parent.name is ao)
                                        ((0.596663 -0.118051))
                                        ((R:segstate.parent.nn.ph_vheight
                                          is
                                          1)
                                         ((0.889869 -0.178523))
                                         ((R:segstate.parent.pp.ph_vfront
                                           is
                                           3)
                                          ((1.36311 0.519344))
                                          ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                            <
                                            4.5)
                                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                             <
                                             2.5)
                                            ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                              <
                                              1.2)
                                             ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                                               <
                                               0.2)
                                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                                is
                                                0)
                                               ((0.752865 -0.0339056))
                                               ((0.723933 0.168858)))
                                              ((0.652821 -0.178552)))
                                             ((R:segstate.parent.p.ph_cvox
                                               is
                                               +)
                                              ((R:segstate.parent.pp.ph_vheight
                                                is
                                                1)
                                               ((0.546494 0.17616))
                                               ((1.06208 -0.0535205)))
                                              ((1.2581 0.773308))))
                                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                              <
                                              2.4)
                                             ((1.42211 0.574093))
                                             ((0.811107 0.286958))))
                                           ((0.836598 -0.128946)))))))
                                      ((R:segstate.parent.n.seg_onsetcoda
                                        is
                                        coda)
                                       ((R:segstate.parent.nn.ph_ctype
                                         is
                                         s)
                                        ((n.name is n_1)
                                         ((0.221096 -0.427126))
                                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                                           <
                                           3.4)
                                          ((0.369788 -0.29716))
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                            <
                                            0.1)
                                           ((0.632019 -0.0991861))
                                           ((1.24035 0.435239)))))
                                        ((R:segstate.parent.pp.ph_vheight
                                          is
                                          2)
                                         ((0.623907 -0.13757))
                                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                           <
                                           0.3)
                                          ((R:segstate.parent.nn.ph_cplace
                                            is
                                            a)
                                           ((0.612144 0.102984))
                                           ((0.699074 -0.133107)))
                                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                            <
                                            2.5)
                                           ((0.753002 0.553945))
                                           ((0.983404 0.208463))))))
                                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                                         <
                                         6.2)
                                        ((0.614826 -0.394501))
                                        ((0.743283 -0.127861))))))))
                                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                                    <
                                    4)
                                   ((0.877142 0.42814))
                                   ((0.965249 0.65165)))))))
                              ((n.name is n_1)
                               ((R:segstate.parent.name is ih)
                                ((0.920648 0.377938))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                  <
                                  2.1)
                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                   <
                                   8.6)
                                  ((0.140964 -0.455341))
                                  ((0.241376 -0.349567)))
                                 ((0.176327 -0.54148))))
                               ((R:segstate.parent.n.name is iy)
                                ((R:segstate.parent.pp.ph_vrnd is -)
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                   <
                                   4)
                                  ((0.710242 -0.231749))
                                  ((0.420782 -0.643935)))
                                 ((0.956714 0.0351192)))
                                ((R:segstate.parent.R:SylStructure.parent.position_type
                                  is
                                  initial)
                                 ((R:segstate.parent.name is eh)
                                  ((0.410568 -0.503702))
                                  ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                    <
                                    2.1)
                                   ((R:segstate.parent.lisp_coda_stop
                                     is
                                     0)
                                    ((R:segstate.parent.n.ph_cvox is +)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                       <
                                       0.3)
                                      ((R:segstate.parent.ph_vheight
                                        is
                                        2)
                                       ((0.550216 -0.121773))
                                       ((0.863028 -0.349881)))
                                      ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                        <
                                        2.1)
                                       ((0.866527 0.0577229))
                                       ((0.709688 0.263223))))
                                     ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                       <
                                       2.2)
                                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                                        <
                                        4)
                                       ((2.02427 0.889523))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                         <
                                         1.2)
                                        ((0.985596 0.167727))
                                        ((1.02893 0.358056))))
                                      ((R:segstate.parent.pp.ph_vfront
                                        is
                                        1)
                                       ((0.953847 0.2655))
                                       ((0.451163 -0.251533)))))
                                    ((0.48046 -0.372043)))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     13.2)
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      2.4)
                                     ((R:segstate.parent.pp.ph_vheight
                                       is
                                       2)
                                      ((0.395606 -0.28989))
                                      ((R:segstate.parent.pp.ph_vlng
                                        is
                                        s)
                                       ((0.718722 0.151882))
                                       ((1.01134 -0.141158))))
                                     ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                       <
                                       5.5)
                                      ((R:segstate.parent.p.ph_ctype
                                        is
                                        n)
                                       ((0.33577 -0.422176))
                                       ((R:segstate.parent.R:SylStructure.parent.last_accent
                                         <
                                         1.4)
                                        ((0.699046 0.0791165))
                                        ((0.53307 -0.328126))))
                                      ((0.53127 -0.552452))))
                                    ((0.875491 0.183667)))))
                                 ((R:segstate.parent.pp.ph_vfront is 3)
                                  ((n.name is ax_3)
                                   ((0.966265 0.541833))
                                   ((R:segstate.parent.p.name is t)
                                    ((1.07195 0.301163))
                                    ((R:segstate.parent.n.ph_cplace
                                      is
                                      a)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                       <
                                       0.1)
                                      ((0.392961 -0.397343))
                                      ((0.840176 0.121779)))
                                     ((R:segstate.parent.p.ph_ctype
                                       is
                                       f)
                                      ((0.376344 -0.463356))
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                        is
                                        0)
                                       ((0.533817 -0.32211))
                                       ((0.590098 -0.153263)))))))
                                  ((n.name is ay_2)
                                   ((0.366606 -0.3279))
                                   ((n.name is ae_1)
                                    ((1.25598 0.68104))
                                    ((R:segstate.parent.nn.ph_cplace
                                      is
                                      v)
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                       <
                                       0.3)
                                      ((0.68218 0.113466))
                                      ((0.807426 -0.370288)))
                                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                                       <
                                       3.3)
                                      ((R:segstate.parent.n.name is er)
                                       ((1.31262 0.616627))
                                       ((R:segstate.parent.nn.ph_cplace
                                         is
                                         g)
                                        ((0.584602 -0.304212))
                                        ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                          <
                                          1.6)
                                         ((R:segstate.parent.n.ph_cvox
                                           is
                                           -)
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                            is
                                            0)
                                           ((0.978717 0.137612))
                                           ((1.07998 0.543058)))
                                          ((R:segstate.parent.n.name
                                            is
                                            n)
                                           ((0.944876 0.492453))
                                           ((R:segstate.parent.p.ph_ctype
                                             is
                                             n)
                                            ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                              <
                                              1.2)
                                             ((0.708801 -0.0252911))
                                             ((0.779362 0.232451)))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                              <
                                              1.2)
                                             ((R:segstate.parent.n.ph_vc
                                               is
                                               +)
                                              ((0.743368 -0.251721))
                                              ((0.337557 -0.539564)))
                                             ((0.630704 -0.0604616))))))
                                         ((R:segstate.parent.n.ph_vrnd
                                           is
                                           -)
                                          ((R:segstate.parent.R:SylStructure.parent.syl_out
                                            <
                                            6.2)
                                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                                             <
                                             2.7)
                                            ((R:segstate.parent.R:SylStructure.parent.next_accent
                                              <
                                              0.3)
                                             ((0.521599 -0.173922))
                                             ((0.940664 0.315554)))
                                            ((1.13354 0.717532)))
                                           ((R:segstate.parent.nn.ph_ctype
                                             is
                                             n)
                                            ((0.696942 -0.441861))
                                            ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                              <
                                              1.2)
                                             ((0.590497 -0.21795))
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                               is
                                               0)
                                              ((0.937952 0.257769))
                                              ((0.861717 0.0300221))))))
                                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                            <
                                            5.5)
                                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                             <
                                             3)
                                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                              <
                                              2.2)
                                             ((R:segstate.parent.n.seg_onsetcoda
                                               is
                                               coda)
                                              ((R:segstate.parent.R:SylStructure.parent.last_accent
                                                <
                                                0.6)
                                               ((0.747241 0.0592482))
                                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                                 <
                                                 0.3)
                                                ((1.32346 0.367153))
                                                ((1.19333 0.532176))))
                                              ((R:segstate.parent.nn.ph_vfront
                                                is
                                                1)
                                               ((0.697891 0.179212))
                                               ((0.57645 -0.322842))))
                                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                                               <
                                               1.8)
                                              ((1.34984 1.01602))
                                              ((0.949201 0.320467))))
                                            ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                              <
                                              3.5)
                                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                               <
                                               0.1)
                                              ((0.770967 -0.240523))
                                              ((R:segstate.parent.lisp_coda_stop
                                                is
                                                0)
                                               ((R:segstate.parent.pp.ph_vfront
                                                 is
                                                 2)
                                                ((0.762954 -0.0900772))
                                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                                  <
                                                  1.1)
                                                 ((1.05919 0.181088))
                                                 ((1.4642 0.897877))))
                                               ((0.561011 -0.128677))))
                                             ((0.368017 -0.353489))))
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                             is
                                             0)
                                            ((R:segstate.parent.pp.ph_vlng
                                              is
                                              d)
                                             ((0.860241 0.225957))
                                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                               <
                                               4.5)
                                              ((R:segstate.parent.pp.ph_vfront
                                                is
                                                1)
                                               ((0.888976 0.695509))
                                               ((0.722583 0.0773691)))
                                              ((R:segstate.parent.pp.ph_vheight
                                                is
                                                2)
                                               ((1.1529 0.719492))
                                               ((1.54584 1.20499)))))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                              <
                                              0.1)
                                             ((0.910635 0.387288))
                                             ((R:segstate.parent.R:SylStructure.parent.syl_out
                                               <
                                               11)
                                              ((0.659342 0.0367048))
                                              ((0.582906 -0.364821))))))))))
                                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                        is
                                        0)
                                       ((0.915332 0.0517148))
                                       ((R:segstate.parent.R:SylStructure.parent.next_accent
                                         <
                                         5.2)
                                        ((1.21021 0.877422))
                                        ((0.970798 0.53789))))))))))))))
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                              <
                              5.5)
                             ((R:segstate.parent.name is th)
                              ((0.781006 -0.505982))
                              ((R:segstate.parent.name is ae)
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 0.6)
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  7.4)
                                 ((0.455042 -0.306176))
                                 ((0.297031 -0.48814)))
                                ((0.540238 -0.139513)))
                               ((n.name is t_2)
                                ((0.506008 -0.431525))
                                ((R:segstate.parent.name is hh)
                                 ((R:segstate.parent.p.name is d)
                                  ((R:segstate.parent.R:SylStructure.parent.next_accent
                                    <
                                    0.7)
                                   ((0.710222 -0.260812))
                                   ((0.891677 0.327843)))
                                  ((n.name is hh_3)
                                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                                     <
                                     0.7)
                                    ((0.623341 -0.0212016))
                                    ((0.456664 -0.282192)))
                                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                                     <
                                     8)
                                    ((0.174098 -0.617231))
                                    ((0.70091 -0.406989)))))
                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                   <
                                   1.6)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    10)
                                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                                     <
                                     3.3)
                                    ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                                      <
                                      0.2)
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       5.8)
                                      ((0.956167 0.162019))
                                      ((1.30381 0.623809)))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                       is
                                       0)
                                      ((0.553251 -0.30384))
                                      ((R:segstate.parent.syl_final
                                        is
                                        0)
                                       ((0.886095 0.307524))
                                       ((0.546921 -0.0151987)))))
                                    ((R:segstate.parent.n.name is v)
                                     ((0.532529 -0.315571))
                                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                       <
                                       1.2)
                                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                                        <
                                        3.4)
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                         <
                                         0.1)
                                        ((0.602258 0.025805))
                                        ((1.3239 0.402165)))
                                       ((R:segstate.parent.ph_vheight
                                         is
                                         2)
                                        ((R:segstate.parent.n.ph_cvox
                                          is
                                          +)
                                         ((0.629019 -0.155835))
                                         ((0.750916 0.258198)))
                                        ((R:segstate.parent.R:SylStructure.parent.syl_in
                                          <
                                          4.5)
                                         ((0.648422 -0.466254))
                                         ((0.724811 -0.0657711)))))
                                      ((0.449328 -0.280153)))))
                                   ((R:segstate.parent.nn.ph_cplace
                                     is
                                     a)
                                    ((0.768893 0.0993822))
                                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                      <
                                      6.5)
                                     ((0.444644 -0.639778))
                                     ((0.592887 -0.220833)))))
                                  ((R:segstate.parent.nn.ph_ctype is 0)
                                   ((R:segstate.parent.R:SylStructure.parent.syl_in
                                     <
                                     2.1)
                                    ((0.486592 -0.30522))
                                    ((0.914399 0.116581)))
                                   ((R:segstate.parent.p.ph_ctype is s)
                                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                      <
                                      3.4)
                                     ((R:segstate.parent.R:SylStructure.parent.position_type
                                       is
                                       initial)
                                      ((1.16042 0.478952))
                                      ((1.8481 0.880712)))
                                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                                       <
                                       3.5)
                                      ((0.925169 0.434886))
                                      ((0.769384 -0.0596731))))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                      is
                                      0)
                                     ((0.641808 0.204431))
                                     ((0.582617 -0.251731))))))))))
                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                               <
                               0.9)
                              ((0.58983 -0.248286))
                              ((0.561293 -0.515963))))))))))))
                    ((n.name is er_1)
                     ((0.863118 0.679062))
                     ((R:segstate.parent.R:SylStructure.parent.next_accent
                       <
                       1.8)
                      ((R:segstate.parent.n.name is f)
                       ((1.26967 0.461964))
                       ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                         <
                         3)
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          initial)
                         ((R:segstate.parent.nn.ph_cplace is l)
                          ((R:segstate.parent.R:SylStructure.parent.asyl_out
                            <
                            1.4)
                           ((0.654022 -0.509997))
                           ((0.537527 -0.150878)))
                          ((R:segstate.parent.nn.ph_ctype is s)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                             is
                             0)
                            ((0.851683 0.14807))
                            ((1.89818 0.593909)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                             is
                             0)
                            ((R:segstate.parent.name is s)
                             ((0.526146 -0.226084))
                             ((R:segstate.parent.n.name is ih)
                              ((0.705864 -0.217083))
                              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                <
                                1.2)
                               ((R:segstate.parent.pp.ph_vrnd is -)
                                ((1.05599 -0.0475005))
                                ((0.971287 0.29472)))
                               ((1.43189 0.913967)))))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                              <
                              2)
                             ((R:segstate.parent.p.ph_ctype is n)
                              ((0.730546 0.0199635))
                              ((R:segstate.parent.ph_vlng is s)
                               ((0.471492 -0.426294))
                               ((R:segstate.parent.nn.ph_vheight is 2)
                                ((0.37144 -0.334597))
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  2.6)
                                 ((0.618619 -0.0460104))
                                 ((0.436452 -0.186469))))))
                             ((0.738705 0.0703199))))))
                         ((R:segstate.parent.n.ph_ctype is r)
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            5.2)
                           ((0.599656 -0.661127))
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             10.6)
                            ((0.80067 -0.170453))
                            ((0.518052 -0.441469))))
                          ((R:segstate.parent.nn.ph_cplace is a)
                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                             <
                             2.4)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                              <
                              1.2)
                             ((R:segstate.parent.n.ph_ctype is f)
                              ((0.346985 -0.501632))
                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                <
                                13.1)
                               ((R:segstate.parent.pp.ph_cplace is v)
                                ((0.498301 -0.436186))
                                ((R:segstate.parent.n.ph_ctype is 0)
                                 ((R:segstate.parent.pp.ph_vheight
                                   is
                                   3)
                                  ((0.433519 -0.402827))
                                  ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                    <
                                    2.2)
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                     is
                                     0)
                                    ((0.609056 -0.221129))
                                    ((1.07229 0.0612331)))
                                   ((R:segstate.parent.pp.ph_vc is +)
                                    ((0.578109 -0.164187))
                                    ((0.710911 -0.423195)))))
                                 ((R:segstate.parent.name is ae)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                    is
                                    0)
                                   ((0.319791 -0.325812))
                                   ((0.699929 -0.0802613)))
                                  ((R:segstate.parent.n.name is n)
                                   ((0.832034 0.273619))
                                   ((0.790059 0.0873191))))))
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 1.3)
                                ((0.563098 -0.359314))
                                ((0.479312 -0.485084)))))
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               5.2)
                              ((0.227201 -0.644795))
                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                <
                                11.5)
                               ((0.488805 -0.350464))
                               ((0.301027 -0.540126)))))
                            ((0.699899 0.0745693)))
                           ((R:segstate.parent.ph_vlng is s)
                            ((n.name is n_1)
                             ((0.712114 -0.0287545))
                             ((R:segstate.parent.pp.ph_vfront is 2)
                              ((0.697202 -0.0298917))
                              ((n.name is v_1)
                               ((0.323043 -0.577201))
                               ((R:segstate.parent.name is ih)
                                ((0.398372 -0.462987))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                  <
                                  5.4)
                                 ((0.530954 -0.194579))
                                 ((0.460566 -0.379165)))))))
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              1.5)
                             ((n.name is t_2)
                              ((0.46593 -0.427625))
                              ((R:segstate.parent.pp.ph_vheight is 1)
                               ((0.416702 -0.304447))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                 <
                                 0.1)
                                ((R:segstate.parent.n.ph_ctype is 0)
                                 ((0.910277 0.153058))
                                 ((1.40169 0.50666)))
                                ((R:segstate.parent.lisp_onset_stop
                                  is
                                  0)
                                 ((R:segstate.parent.p.ph_ctype is s)
                                  ((0.927526 0.184296))
                                  ((R:segstate.parent.p.ph_cvox is +)
                                   ((0.619611 -0.183268))
                                   ((0.32597 -0.359488))))
                                 ((0.749339 0.316518))))))
                             ((R:segstate.parent.p.ph_ctype is n)
                              ((0.794613 -0.162602))
                              ((R:segstate.parent.R:SylStructure.parent.last_accent
                                <
                                2.3)
                               ((1.4916 0.61954))
                               ((0.748439 0.12064)))))))))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                          <
                          1.2)
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           11.4)
                          ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                            <
                            1.1)
                           ((0.506607 -0.337038))
                           ((0.396805 -0.628226)))
                          ((0.597078 -0.163916)))
                         ((0.460823 -0.800447)))))
                      ((R:segstate.parent.n.ph_cplace is l)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                         is
                         0)
                        ((1.03567 0.708373))
                        ((0.961814 0.118698)))
                       ((R:segstate.parent.pp.ph_vlng is d)
                        ((R:segstate.parent.R:SylStructure.parent.last_accent
                          <
                          0.2)
                         ((0.643894 -0.507935))
                         ((0.521432 -0.00571161)))
                        ((R:segstate.parent.nn.ph_ctype is 0)
                         ((R:segstate.parent.R:SylStructure.parent.last_accent
                           <
                           0.5)
                          ((1.12737 0.585516))
                          ((1.0275 0.101178)))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                           is
                           0)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((R:segstate.parent.R:SylStructure.parent.syl_out
                             <
                             4.2)
                            ((0.484681 -0.524323))
                            ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                              <
                              0.1)
                             ((0.794199 -0.0466681))
                             ((0.583141 -0.230825))))
                           ((R:segstate.parent.p.ph_cvox is +)
                            ((0.650976 0.14567))
                            ((0.423216 -0.14336))))
                          ((R:segstate.parent.lisp_onset_stop is 0)
                           ((R:segstate.parent.ph_vlng is s)
                            ((0.901219 0.206953))
                            ((R:segstate.parent.pp.ph_cplace is a)
                             ((0.71771 -0.390247))
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               5.4)
                              ((0.765393 -0.234967))
                              ((0.698125 0.179962)))))
                           ((R:segstate.parent.nn.ph_cplace is a)
                            ((1.57621 0.634627))
                            ((0.683499 0.144044)))))))))))))))))
             ((n.name is ae_3)
              ((R:segstate.parent.n.ph_cvox is +)
               ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                 <
                 1.3)
                ((R:segstate.parent.pp.ph_vlng is a)
                 ((1.06707 0.901173))
                 ((0.853922 0.256095)))
                ((R:segstate.parent.R:SylStructure.parent.asyl_in
                  <
                  1.4)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                   <
                   0.1)
                  ((1.24821 0.326008))
                  ((R:segstate.parent.n.ph_cplace is a)
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     1)
                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                      <
                      0.6)
                     ((0.311489 -0.582879))
                     ((0.450263 -0.446904)))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                      is
                      0)
                     ((0.385006 -0.451198))
                     ((0.959148 0.127576))))
                   ((0.871973 0.11675))))
                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                   <
                   0.5)
                  ((0.285882 -0.649674))
                  ((0.640298 -0.292881)))))
               ((R:segstate.parent.p.name is dh)
                ((R:segstate.parent.R:SylStructure.parent.syl_in < 3)
                 ((0.849196 1.1347))
                 ((R:segstate.parent.pp.ph_cplace is 0)
                  ((0.557808 -0.521107))
                  ((1.1298 0.41141))))
                ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                  <
                  0.1)
                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                   <
                   1.4)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    3)
                   ((0.459791 0.892334))
                   ((0.336539 1.02211)))
                  ((0.442244 1.46391)))
                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                   <
                   2.4)
                  ((0.571179 1.40245))
                  ((0.685207 1.91435))))))
              ((R:segstate.parent.n.name is n)
               ((n.name is ae_2)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                  is
                  0)
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   4)
                  ((1.11513 1.52112))
                  ((0.724349 1.88915)))
                 ((1.15 0.411746)))
                ((n.name is n_1)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                   <
                   1.2)
                  ((R:segstate.parent.ph_vlng is d)
                   ((R:segstate.parent.pp.ph_cplace is 0)
                    ((R:segstate.parent.pp.ph_vrnd is -)
                     ((0.545406 -0.375725))
                     ((0.292832 -0.861187)))
                    ((0.505657 -0.307611)))
                   ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                     <
                     2.2)
                    ((R:segstate.parent.R:SylStructure.parent.position_type
                      is
                      initial)
                     ((R:segstate.parent.nn.ph_vfront is 2)
                      ((1.48253 0.434546))
                      ((R:segstate.parent.ph_vlng is s)
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                         <
                         3)
                        ((0.186165 -0.498503))
                        ((0.382397 -0.304177)))
                       ((R:segstate.parent.pp.ph_vc is +)
                        ((0.545171 0.115819))
                        ((0.502838 -0.0705311)))))
                     ((R:segstate.parent.pp.ph_vlng is a)
                      ((0.617912 -0.0656898))
                      ((R:segstate.parent.p.ph_cvox is +)
                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                         <
                         9)
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                          <
                          0.3)
                         ((0.192779 -0.545465))
                         ((R:segstate.parent.R:SylStructure.parent.last_accent
                           <
                           1.2)
                          ((R:segstate.parent.ph_vheight is 2)
                           ((0.249945 -0.406742))
                           ((0.30527 -0.251033)))
                          ((0.193196 -0.457125))))
                        ((0.409376 -0.219213)))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                         is
                         0)
                        ((0.53932 -0.406327))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          5.2)
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           6.4)
                          ((0.296422 -0.363949))
                          ((0.488179 -0.13718)))
                         ((0.60147 -0.0869708)))))))
                    ((R:segstate.parent.ph_vlng is a)
                     ((R:segstate.parent.pp.ph_vfront is 1)
                      ((0.501782 -0.0545945))
                      ((0.185042 -0.390882)))
                     ((1.05355 0.408626)))))
                  ((R:segstate.parent.R:SylStructure.parent.asyl_out
                    <
                    1.5)
                   ((2.33854 1.22648))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                     <
                     0.1)
                    ((0.605701 0.14432))
                    ((0.670383 -0.379262)))))
                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                   <
                   7.2)
                  ((R:segstate.parent.syl_final is 0)
                   ((n.name is iy_2)
                    ((0.289092 -0.59916))
                    ((R:segstate.parent.p.name is g)
                     ((1.0233 0.875652))
                     ((n.name is iy_3)
                      ((0.9235 0.967382))
                      ((R:segstate.parent.p.name is l)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                         is
                         0)
                        ((0.579177 -0.22197))
                        ((0.772992 -0.522701)))
                       ((n.name is ih_2)
                        ((0.259358 -0.330106))
                        ((R:segstate.parent.R:SylStructure.parent.asyl_in
                          <
                          1.2)
                         ((R:segstate.parent.p.ph_cvox is +)
                          ((R:segstate.parent.name is eh)
                           ((0.837955 0.255013))
                           ((R:segstate.parent.R:SylStructure.parent.last_accent
                             <
                             0.6)
                            ((R:segstate.parent.pp.ph_vc is +)
                             ((0.481421 0.074014))
                             ((1.41204 0.742921)))
                            ((R:segstate.parent.nn.ph_cplace is a)
                             ((1.40074 1.29507))
                             ((0.906792 0.625953)))))
                          ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                            <
                            0.2)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                             is
                             0)
                            ((R:segstate.parent.R:SylStructure.parent.position_type
                              is
                              initial)
                             ((0.424909 -0.299456))
                             ((0.837453 0.168401)))
                            ((R:segstate.parent.pp.ph_cplace is 0)
                             ((R:segstate.parent.nn.ph_ctype is s)
                              ((0.777002 -0.122547))
                              ((1.1075 0.539862)))
                             ((1.39371 0.88797))))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             0.3)
                            ((0.544477 -0.0665342))
                            ((0.3864 -0.418711)))))
                         ((R:segstate.parent.ph_vlng is s)
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            5.4)
                           ((0.843243 -0.0453564))
                           ((1.47133 0.65498)))
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            0.8)
                           ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
                             <
                             0.1)
                            ((0.127205 -0.457609))
                            ((0.684324 -0.134091)))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             3.5)
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              1.2)
                             ((R:segstate.parent.nn.ph_cplace is a)
                              ((0.570639 -0.0735318))
                              ((0.811296 0.306124)))
                             ((0.435355 -0.326799)))
                            ((R:segstate.parent.R:SylStructure.parent.asyl_in
                              <
                              2.3)
                             ((0.813812 0.494618))
                             ((0.783964 0.00702364))))))))))))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                     <
                     0.3)
                    ((R:segstate.parent.pp.ph_cplace is a)
                     ((R:segstate.parent.R:SylStructure.parent.syl_in
                       <
                       8.5)
                      ((0.309967 -0.365993))
                      ((0.520646 -0.0479089)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                       <
                       1.2)
                      ((0.348054 -0.310414))
                      ((0.282996 -0.535809))))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                      is
                      0)
                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                       <
                       7)
                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                        <
                        1.6)
                       ((0.606292 -0.215434))
                       ((0.712848 0.271669)))
                      ((0.89939 0.472183)))
                     ((R:segstate.parent.p.ph_cvox is +)
                      ((R:segstate.parent.p.ph_ctype is s)
                       ((0.492859 0.0475958))
                       ((0.512359 -0.158877)))
                      ((0.438036 -0.514178))))))
                  ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                    <
                    2.2)
                   ((R:segstate.parent.pp.ph_vfront is 1)
                    ((1.18246 0.288445))
                    ((R:segstate.parent.R:SylStructure.parent.asyl_out
                      <
                      4.2)
                     ((1.1008 0.653416))
                     ((1.07774 1.11011))))
                   ((0.726153 0.136334))))))
               ((R:segstate.parent.p.ph_ctype is s)
                ((n.name is s_2)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                   <
                   0.3)
                  ((0.239791 -1.0763))
                  ((R:segstate.parent.nn.ph_vc is +)
                   ((0.549425 -0.894575))
                   ((0.386486 -0.754693))))
                 ((n.name is s_3)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                    <
                    0.3)
                   ((0.922023 -0.476878))
                   ((0.650052 -1.18305)))
                  ((R:segstate.parent.n.seg_onsetcoda is coda)
                   ((R:segstate.parent.lisp_coda_glide is 0)
                    ((n.name is m_1)
                     ((R:segstate.parent.nn.ph_cplace is l)
                      ((0.526113 -0.306833))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                        is
                        0)
                       ((0.304142 -0.696019))
                       ((0.239647 -0.586275))))
                     ((n.name is aa_2)
                      ((1.54042 1.02391))
                      ((n.name is ey_2)
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         6.1)
                        ((0.0894996 -0.692063))
                        ((1.05129 -0.218787)))
                       ((R:segstate.parent.n.name is m)
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          single)
                         ((R:segstate.parent.R:SylStructure.parent.last_accent
                           <
                           1.5)
                          ((1.18951 1.1588))
                          ((0.989537 0.659904)))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                           <
                           3.6)
                          ((0.463966 -0.186414))
                          ((1.1695 0.441565))))
                        ((n.name is ae_2)
                         ((R:segstate.parent.nn.ph_ctype is s)
                          ((0.238046 -0.584971))
                          ((0.870722 -0.26188)))
                         ((n.name is ah_2)
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                            is
                            0)
                           ((0.302495 -0.444527))
                           ((0.645524 -0.300452)))
                          ((n.name is ih_2)
                           ((R:segstate.parent.pp.ph_vc is +)
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              0.7)
                             ((R:segstate.parent.nn.ph_cplace is a)
                              ((0.489481 -0.134486))
                              ((0.60176 0.129172)))
                             ((0.237413 -0.237611)))
                            ((R:segstate.parent.n.ph_cplace is a)
                             ((0.18792 -0.31635))
                             ((0.151862 -0.403462))))
                           ((R:segstate.parent.p.ph_cvox is +)
                            ((n.name is ah_3)
                             ((0.907447 -0.252275))
                             ((R:segstate.parent.R:SylStructure.parent.syl_in
                               <
                               6)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((R:segstate.parent.n.name is d)
                                ((0.779951 -0.130114))
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  6.2)
                                 ((R:segstate.parent.ph_vheight is 2)
                                  ((0.814618 0.332743))
                                  ((1.07719 0.800494)))
                                 ((R:segstate.parent.ph_vheight is 2)
                                  ((1.00974 0.00567345))
                                  ((1.25349 0.228341)))))
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 2.5)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                  <
                                  0.3)
                                 ((R:segstate.parent.p.name is b)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    9)
                                   ((0.873819 0.541392))
                                   ((1.00656 1.01643)))
                                  ((R:segstate.parent.n.ph_ctype is f)
                                   ((0.888211 0.514189))
                                   ((1.20811 0.120736))))
                                 ((1.36447 0.843841)))
                                ((R:segstate.parent.n.ph_ctype is f)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                   is
                                   0)
                                  ((1.56502 0.80891))
                                  ((1.34924 0.331074)))
                                 ((R:segstate.parent.ph_vlng is s)
                                  ((1.0351 0.338285))
                                  ((0.768639 -0.0320751))))))
                              ((R:segstate.parent.R:SylStructure.parent.last_accent
                                <
                                0.8)
                               ((R:segstate.parent.nn.ph_ctype is f)
                                ((0.517612 -0.0432895))
                                ((0.447472 -0.491842)))
                               ((R:segstate.parent.n.name is d)
                                ((1.16087 0.437033))
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  2.2)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                   is
                                   0)
                                  ((R:segstate.parent.ph_vheight is 3)
                                   ((0.682095 0.0858624))
                                   ((1.20624 0.337111)))
                                  ((0.68485 -0.0279816)))
                                 ((0.516422 -0.156794)))))))
                            ((R:segstate.parent.n.name is d)
                             ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                               <
                               0.2)
                              ((R:segstate.parent.pp.ph_vheight is 2)
                               ((0.45797 -0.361863))
                               ((R:segstate.parent.R:SylStructure.parent.last_accent
                                 <
                                 1.5)
                                ((n.name is d_1)
                                 ((0.664802 -0.101968))
                                 ((R:segstate.parent.ph_vheight is 1)
                                  ((0.593516 -0.0110622))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                    is
                                    0)
                                   ((0.827581 0.543725))
                                   ((1.02985 0.121478)))))
                                ((0.688435 -0.244081))))
                              ((0.14178 -0.474695)))
                             ((R:segstate.parent.n.name is ng)
                              ((n.name is ih_3)
                               ((0.65929 -0.333143))
                               ((R:segstate.parent.R:SylStructure.parent.syl_in
                                 <
                                 4.5)
                                ((0.774346 -0.104572))
                                ((0.939762 0.248185))))
                              ((R:segstate.parent.pp.ph_vheight is 1)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                 <
                                 3)
                                ((0.87525 0.21885))
                                ((1.06497 0.940029)))
                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                 <
                                 4.6)
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  1.2)
                                 ((0.404795 -0.345987))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                   is
                                   0)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                                    <
                                    0.1)
                                   ((0.73926 0.0839735))
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                     is
                                     0)
                                    ((0.603695 -0.0880694))
                                    ((0.454705 -0.351483))))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                    <
                                    1.2)
                                   ((0.846871 0.037203))
                                   ((0.591142 0.392255)))))
                                ((R:segstate.parent.n.name is pau)
                                 ((R:segstate.parent.nn.ph_ctype is s)
                                  ((0.862156 0.0127681))
                                  ((R:segstate.parent.pp.ph_vc is +)
                                   ((1.06831 0.33629))
                                   ((1.87843 1.17891))))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                   is
                                   0)
                                  ((0.973147 0.410114))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                    is
                                    0)
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                     <
                                     1.2)
                                    ((R:segstate.parent.n.ph_ctype
                                      is
                                      f)
                                     ((1.07671 0.381812))
                                     ((0.733932 0.116457)))
                                    ((0.81722 -0.0272516)))
                                   ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                     <
                                     6)
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                      is
                                      0)
                                     ((0.761316 0.0646262))
                                     ((0.337159 -0.0564624)))
                                    ((0.549183 -0.239653))))))))))))))))))
                    ((R:segstate.parent.name is aa)
                     ((n.name is aa_2)
                      ((0.507598 -0.379787))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                        <
                        1.2)
                       ((0.24252 -0.706201))
                       ((0.0636354 -0.796235))))
                     ((R:segstate.parent.name is ow)
                      ((0.154113 -0.89658))
                      ((n.name is l_1)
                       ((R:segstate.parent.p.name is t)
                        ((1.23941 1.00531))
                        ((R:segstate.parent.nn.ph_cplace is a)
                         ((0.388503 -0.305815))
                         ((0.246706 -0.501401))))
                       ((R:segstate.parent.R:SylStructure.parent.position_type
                         is
                         single)
                        ((R:segstate.parent.p.name is b)
                         ((1.20508 0.385966))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                           <
                           1.2)
                          ((1.01783 0.0963829))
                          ((0.495937 -0.320386))))
                        ((R:segstate.parent.n.name is l)
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           9.4)
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            9)
                           ((R:segstate.parent.nn.ph_cplace is a)
                            ((0.066711 -0.491034))
                            ((0.403497 -0.377271)))
                           ((0.133569 -0.498042)))
                          ((0.153543 -0.542951)))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                           <
                           0.3)
                          ((0.354959 -0.357353))
                          ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                            <
                            2.2)
                           ((0.813836 0.0160622))
                           ((0.493729 -0.188544))))))))))
                   ((R:segstate.parent.pp.ph_vc is +)
                    ((R:segstate.parent.nn.ph_vfront is 2)
                     ((R:segstate.parent.n.ph_ctype is r)
                      ((2.45004 1.04893))
                      ((R:segstate.parent.p.name is g)
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                         <
                         1.2)
                        ((0.943184 0.43664))
                        ((0.87658 0.609819)))
                       ((R:segstate.parent.n.ph_ctype is f)
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          4.5)
                         ((R:segstate.parent.R:SylStructure.parent.syl_in
                           <
                           6)
                          ((0.574099 -0.13714))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                            <
                            0.4)
                           ((0.587258 0.434406))
                           ((0.657186 0.0726024))))
                         ((1.03812 0.569513)))
                        ((R:segstate.parent.name is aa)
                         ((R:segstate.parent.R:SylStructure.parent.asyl_in
                           <
                           1.2)
                          ((0.864946 0.254829))
                          ((0.74781 0.133812)))
                         ((R:segstate.parent.ph_vrnd is -)
                          ((R:segstate.parent.p.name is d)
                           ((0.850826 0.0367347))
                           ((R:segstate.parent.R:SylStructure.parent.next_accent
                             <
                             0.7)
                            ((0.70672 -0.161909))
                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                              <
                              1.3)
                             ((0.275022 -0.606841))
                             ((0.420409 -0.340131)))))
                          ((0.963342 0.162088)))))))
                     ((R:segstate.parent.n.name is f)
                      ((0.446357 -0.566199))
                      ((R:segstate.parent.n.name is l)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                         <
                         0.3)
                        ((0.28262 -0.571271))
                        ((0.80068 -0.10937)))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                         <
                         1.4)
                        ((R:segstate.parent.R:SylStructure.parent.last_accent
                          <
                          3.6)
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           2.8)
                          ((R:segstate.parent.ph_vlng is l)
                           ((1.16497 0.314612))
                           ((R:segstate.parent.R:SylStructure.parent.asyl_in
                             <
                             2.1)
                            ((R:segstate.parent.name is ax)
                             ((0.923434 0.298355))
                             ((0.716018 0.0210029)))
                            ((0.427807 -0.191601))))
                          ((R:segstate.parent.R:SylStructure.parent.position_type
                            is
                            final)
                           ((0.664974 0.0882722))
                           ((0.491144 -0.248209))))
                         ((1.23179 0.405327)))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                          is
                          0)
                         ((R:segstate.parent.ph_vheight is 2)
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            5.2)
                           ((0.68929 0.0298781))
                           ((0.601979 -0.208014)))
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            4.2)
                           ((0.678078 0.0242707))
                           ((1.18734 0.425453))))
                         ((R:segstate.parent.ph_vlng is d)
                          ((0.849831 0.273896))
                          ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                            <
                            6)
                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                             <
                             2)
                            ((0.377889 -0.451093))
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              8)
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               3.5)
                              ((R:segstate.parent.nn.ph_vheight is 2)
                               ((0.362608 -0.452451))
                               ((0.473721 -0.257504)))
                              ((0.495367 -0.134206)))
                             ((R:segstate.parent.nn.ph_vheight is 3)
                              ((0.611166 -0.334748))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                is
                                0)
                               ((1.0938 0.363776))
                               ((0.616232 -0.190285))))))
                           ((0.297482 -0.507246)))))))))
                    ((n.name is er_2)
                     ((R:segstate.parent.pp.ph_cplace is a)
                      ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                        <
                        2.2)
                       ((0.280894 -0.726297))
                       ((0.117791 -0.870838)))
                      ((0.717019 -0.40272)))
                     ((R:segstate.parent.name is ah)
                      ((R:segstate.parent.nn.ph_vheight is 1)
                       ((0.704019 -0.0669207))
                       ((1.01369 0.547928)))
                      ((R:segstate.parent.n.ph_cvox is +)
                       ((n.name is uw_2)
                        ((1.01661 0.135827))
                        ((n.name is r_1)
                         ((0.280552 -0.627326))
                         ((R:segstate.parent.name is aa)
                          ((1.09399 0.189979))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                            is
                            0)
                           ((R:segstate.parent.R:SylStructure.parent.asyl_in
                             <
                             3.2)
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                              is
                              0)
                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                               <
                               3.6)
                              ((R:segstate.parent.n.ph_cplace is l)
                               ((0.237343 -0.324912))
                               ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                 <
                                 2.2)
                                ((0.274183 -0.626999))
                                ((0.388459 -0.38391))))
                              ((R:segstate.parent.nn.ph_vheight is 2)
                               ((0.608861 -0.376673))
                               ((0.520099 -0.112345))))
                             ((R:segstate.parent.n.name is d)
                              ((0.792195 0.20525))
                              ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                <
                                1.8)
                               ((0.734848 0.0942319))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                 <
                                 1.2)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                  is
                                  0)
                                 ((0.548348 -0.239318))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                   <
                                   7.5)
                                  ((0.834155 -0.137611))
                                  ((1.19054 0.101006))))
                                ((R:segstate.parent.R:SylStructure.parent.next_accent
                                  <
                                  1.2)
                                 ((R:segstate.parent.name is ax)
                                  ((0.31458 -0.339432))
                                  ((0.249169 -0.62122)))
                                 ((0.51364 -0.152892)))))))
                            ((R:segstate.parent.pp.ph_cplace is 0)
                             ((0.460809 -0.276326))
                             ((1.32728 0.304323))))
                           ((R:segstate.parent.name is er)
                            ((0.329466 -0.678127))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                              <
                              0.3)
                             ((R:segstate.parent.ph_vlng is d)
                              ((0.586068 -0.8831))
                              ((R:segstate.parent.n.ph_cplace is l)
                               ((0.584042 -0.148384))
                               ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                                 <
                                 1.1)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                  is
                                  0)
                                 ((R:segstate.parent.ph_vlng is s)
                                  ((0.275341 -0.554951))
                                  ((0.123276 -0.511759)))
                                 ((0.211287 -0.425787)))
                                ((0.49626 -0.187614)))))
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                               <
                               1.1)
                              ((R:segstate.parent.p.name is t)
                               ((R:segstate.parent.nn.ph_vfront is 2)
                                ((0.520801 -0.218011))
                                ((0.256169 -0.408844)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                 <
                                 0.3)
                                ((0.712574 -0.135321))
                                ((0.727236 0.171119))))
                              ((R:segstate.parent.R:SylStructure.parent.position_type
                                is
                                initial)
                               ((R:segstate.parent.nn.ph_vfront is 1)
                                ((0.202016 -0.523602))
                                ((R:segstate.parent.pp.ph_cplace is 0)
                                 ((0.292936 -0.487189))
                                 ((0.716289 -0.229291))))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                 is
                                 0)
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  1.2)
                                 ((0.385975 -0.351072))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   1.2)
                                  ((0.3624 -0.29485))
                                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                    <
                                    3.2)
                                   ((0.522629 -0.158963))
                                   ((0.919135 0.0857172)))))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                  is
                                  0)
                                 ((0.477405 -0.380942))
                                 ((0.18554 -0.464212))))))))))))
                       ((R:segstate.parent.nn.ph_vrnd is +)
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          initial)
                         ((R:segstate.parent.R:SylStructure.parent.asyl_in
                           <
                           2.5)
                          ((0.475681 -0.645981))
                          ((0.379179 -0.771492)))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                           <
                           3.6)
                          ((0.804025 0.020498))
                          ((0.453853 -0.132289))))
                        ((n.name is f_1)
                         ((0.606671 0.256093))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                           is
                           0)
                          ((R:segstate.parent.p.name is b)
                           ((0.858282 0.297214))
                           ((R:segstate.parent.n.name is p)
                            ((0.547848 -0.274854))
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              3.6)
                             ((R:segstate.parent.p.name is t)
                              ((R:segstate.parent.pp.ph_cplace is a)
                               ((0.377947 -0.338113))
                               ((0.503747 -0.0315911)))
                              ((R:segstate.parent.n.ph_ctype is s)
                               ((R:segstate.parent.R:SylStructure.parent.syl_out
                                 <
                                 6.1)
                                ((0.417356 -0.360101))
                                ((0.945705 0.0958709)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                 is
                                 0)
                                ((0.733245 0.205837))
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  5)
                                 ((0.49205 -0.0968301))
                                 ((0.763256 0.0790377))))))
                             ((0.826255 0.242703)))))
                          ((n.name is ih_3)
                           ((R:segstate.parent.pp.ph_cplace is 0)
                            ((0.174834 -0.654079))
                            ((0.476298 -0.389023)))
                           ((R:segstate.parent.n.name is hh)
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              2.8)
                             ((0.659175 -0.117822))
                             ((R:segstate.parent.R:SylStructure.parent.syl_in
                               <
                               6.6)
                              ((1.05267 0.354057))
                              ((0.41522 0.107236))))
                            ((R:segstate.parent.ph_vheight is 1)
                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                               <
                               2.5)
                              ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                <
                                1.1)
                               ((0.576622 -0.111625))
                               ((0.386562 -0.369859)))
                              ((n.name is ih_2)
                               ((0.305958 -0.178807))
                               ((R:segstate.parent.name is ih)
                                ((0.774099 0.620017))
                                ((0.495184 0.0522306)))))
                             ((n.name is er_3)
                              ((0.587393 -0.621368))
                              ((n.name is ax_2)
                               ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                 <
                                 1.5)
                                ((0.0858268 -0.472477))
                                ((0.194271 -0.368277)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                 is
                                 0)
                                ((n.name is ax_3)
                                 ((0.464254 -0.274853))
                                 ((0.471563 0.106396)))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                  <
                                  2.4)
                                 ((0.403164 -0.398501))
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   0.5)
                                  ((0.332004 -0.274015))
                                  ((0.511982 -0.126589))))))))))))))))))))
                ((R:segstate.parent.p.name is s)
                 ((n.name is r_1)
                  ((R:segstate.parent.lisp_coda_stop is 0)
                   ((0.889049 0.437703))
                   ((0.809519 1.31027)))
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    2.4)
                   ((R:segstate.parent.pp.ph_vheight is 1)
                    ((R:segstate.parent.n.ph_cplace is a)
                     ((2.00048 1.04544))
                     ((0.933695 0.296079)))
                    ((R:segstate.parent.n.name is pau)
                     ((R:segstate.parent.pp.ph_cplace is 0)
                      ((0.625018 -0.346652))
                      ((1.65265 1.37286)))
                     ((R:segstate.parent.lisp_coda_glide is 0)
                      ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                        <
                        0.2)
                       ((R:segstate.parent.n.ph_cplace is p)
                        ((1.33706 0.481903))
                        ((R:segstate.parent.R:SylStructure.parent.asyl_out
                          <
                          4.2)
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           5.8)
                          ((R:segstate.parent.R:SylStructure.parent.position_type
                            is
                            single)
                           ((0.872429 -0.0166404))
                           ((0.418135 -0.332859)))
                          ((R:segstate.parent.nn.ph_cplace is a)
                           ((1.42445 0.520807))
                           ((R:segstate.parent.nn.ph_vheight is 2)
                            ((0.938018 0.278309))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                              is
                              0)
                             ((0.897985 0.175102))
                             ((0.613678 -0.121461))))))
                         ((R:segstate.parent.n.ph_cplace is l)
                          ((0.460556 -0.488346))
                          ((0.830902 -0.0417972)))))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                         <
                         2.2)
                        ((R:segstate.parent.nn.ph_ctype is 0)
                         ((0.481433 -0.428932))
                         ((0.682283 -0.193791)))
                        ((0.701806 0.145728))))
                      ((0.662082 -0.377037)))))
                   ((n.name is p_3)
                    ((0.311898 0.403432))
                    ((R:segstate.parent.name is p)
                     ((0.208692 -0.90867))
                     ((n.name is ax_3)
                      ((0.811367 0.2615))
                      ((n.name is k_3)
                       ((0.295367 0.309818))
                       ((R:segstate.parent.name is k)
                        ((0.215608 -0.876233))
                        ((R:segstate.parent.name is ao)
                         ((0.859201 0.179705))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                           <
                           7)
                          ((R:segstate.parent.nn.ph_ctype is s)
                           ((R:segstate.parent.ph_vheight is 2)
                            ((0.504434 -0.544082))
                            ((R:segstate.parent.R:SylStructure.parent.syl_out
                              <
                              3.4)
                             ((0.386996 -0.389613))
                             ((R:segstate.parent.nn.ph_cplace is a)
                              ((0.641041 -0.351039))
                              ((R:segstate.parent.R:SylStructure.parent.syl_out
                                <
                                7.2)
                               ((0.406578 -0.19782))
                               ((0.613417 0.0192496))))))
                           ((n.name is l_1)
                            ((0.837543 0.216679))
                            ((R:segstate.parent.lisp_coda_glide is 0)
                             ((R:segstate.parent.ph_vheight is 2)
                              ((R:segstate.parent.pp.ph_vlng is l)
                               ((1.14015 0.396986))
                               ((R:segstate.parent.ph_vlng is s)
                                ((R:segstate.parent.n.ph_cplace is a)
                                 ((R:segstate.parent.n.name is d)
                                  ((0.833839 0.310013))
                                  ((0.857906 0.194839)))
                                 ((R:segstate.parent.n.name is m)
                                  ((0.621638 -0.115244))
                                  ((0.902862 0.0123399))))
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                  is
                                  0)
                                 ((0.714332 0.14713))
                                 ((R:segstate.parent.n.ph_cvox is +)
                                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                                    <
                                    1.5)
                                   ((0.467113 -0.499605))
                                   ((0.645911 -0.262819)))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                    <
                                    0.1)
                                   ((0.398793 -0.277315))
                                   ((0.580281 0.000585567)))))))
                              ((R:segstate.parent.pos_in_syl is 3)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                 is
                                 0)
                                ((0.754597 -0.232931))
                                ((0.14482 -0.636821)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                 is
                                 0)
                                ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                  <
                                  2.2)
                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                   <
                                   1.4)
                                  ((R:segstate.parent.nn.ph_vfront
                                    is
                                    1)
                                   ((0.655066 -0.0875617))
                                   ((R:segstate.parent.n.ph_cvox is -)
                                    ((0.314962 -0.430543))
                                    ((0.566716 -0.15787))))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                    <
                                    0.1)
                                   ((0.392457 -0.322931))
                                   ((0.337695 -0.477617))))
                                 ((0.681679 -0.0332985)))
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                  <
                                  3)
                                 ((0.745483 0.172759))
                                 ((R:segstate.parent.pp.ph_vc is +)
                                  ((0.498025 -0.340043))
                                  ((0.775409 0.0275014)))))))
                             ((n.name is eh_3)
                              ((0.122344 -0.740958))
                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                <
                                6.6)
                               ((0.384654 -0.552856))
                               ((0.722802 -0.112486)))))))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                            is
                            0)
                           ((R:segstate.parent.R:SylStructure.parent.asyl_in
                             <
                             3.5)
                            ((0.653027 -0.353834))
                            ((0.491937 -0.0809992)))
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                             <
                             0.1)
                            ((0.209529 -0.526457))
                            ((0.422238 -0.39906)))))))))))))
                 ((R:segstate.parent.n.ph_ctype is f)
                  ((R:segstate.parent.p.name is jh)
                   ((R:segstate.parent.n.ph_cplace is a)
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      2.2)
                     ((0.678231 0.351622))
                     ((1.41977 1.04497)))
                    ((0.694111 0.0882554)))
                   ((R:segstate.parent.p.name is z)
                    ((R:segstate.parent.R:SylStructure.parent.syl_out
                      <
                      4.4)
                     ((0.813652 0.193116))
                     ((1.02325 0.685387)))
                    ((R:segstate.parent.p.name is f)
                     ((R:segstate.parent.pp.ph_vfront is 2)
                      ((R:segstate.parent.R:SylStructure.parent.last_accent
                        <
                        1.4)
                       ((0.647485 -0.187409))
                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                         <
                         5.4)
                        ((0.547894 0.0288342))
                        ((0.830028 -0.00371507))))
                      ((R:segstate.parent.ph_vlng is d)
                       ((0.509629 0.0676539))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                         is
                         0)
                        ((0.884712 0.155624))
                        ((R:segstate.parent.R:SylStructure.parent.position_type
                          is
                          single)
                         ((0.90287 0.313319))
                         ((1.15788 0.482361))))))
                     ((n.name is ax_3)
                      ((R:segstate.parent.n.name is hh)
                       ((0.980279 0.364481))
                       ((R:segstate.parent.R:SylStructure.parent.syl_in
                         <
                         2.3)
                        ((R:segstate.parent.nn.ph_ctype is 0)
                         ((0.2588 -0.304889))
                         ((0.524346 -0.0817662)))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                          <
                          3)
                         ((0.893769 0.458372))
                         ((R:segstate.parent.p.name is dh)
                          ((R:segstate.parent.n.ph_cplace is b)
                           ((0.7821 0.0949139))
                           ((0.519832 -0.263697)))
                          ((0.739267 0.152799))))))
                      ((R:segstate.parent.pp.ph_vlng is a)
                       ((n.name is ih_2)
                        ((1.30968 0.623535))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                          is
                          0)
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                           is
                           0)
                          ((0.611058 -0.111064))
                          ((R:segstate.parent.ph_vlng is d)
                           ((0.47836 0.0367977))
                           ((1.49418 0.580017))))
                         ((R:segstate.parent.p.ph_ctype is f)
                          ((R:segstate.parent.R:SylStructure.parent.next_accent
                            <
                            0.6)
                           ((0.432206 -0.332722))
                           ((0.813349 0.170909)))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                            is
                            0)
                           ((0.367795 -0.488792))
                           ((0.690212 -0.268596))))))
                       ((R:segstate.parent.name is ah)
                        ((R:segstate.parent.R:SylStructure.parent.syl_out
                          <
                          6.2)
                         ((0.962134 -0.159853))
                         ((1.12571 0.629001)))
                        ((R:segstate.parent.n.name is zh)
                         ((0.934651 0.235378))
                         ((R:segstate.parent.pp.ph_vlng is s)
                          ((R:segstate.parent.R:SylStructure.parent.last_accent
                            <
                            2.6)
                           ((R:segstate.parent.nn.ph_vc is +)
                            ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                              <
                              4.4)
                             ((R:segstate.parent.R:SylStructure.parent.syl_out
                               <
                               2.2)
                              ((0.355742 -0.476326))
                              ((0.328941 -0.673683)))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                               is
                               0)
                              ((0.80641 -0.200961))
                              ((0.474465 -0.506929))))
                            ((0.475269 -0.82879)))
                           ((0.604979 -0.228102)))
                          ((n.name is ae_2)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                             <
                             0.3)
                            ((0.486976 -0.428277))
                            ((0.0940666 -0.704871)))
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                             <
                             2.6)
                            ((R:segstate.parent.ph_vheight is 3)
                             ((0.543919 0.0108138))
                             ((R:segstate.parent.pp.ph_cplace is a)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((R:segstate.parent.R:SylStructure.parent.position_type
                                 is
                                 single)
                                ((R:segstate.parent.n.ph_cplace is b)
                                 ((0.153898 -0.527708))
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                   <
                                   1.2)
                                  ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                    <
                                    2.5)
                                   ((0.16698 -0.590789))
                                   ((0.606266 -0.227266)))
                                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                                    <
                                    4.6)
                                   ((0.809593 -0.080544))
                                   ((0.295545 -0.436312)))))
                                ((0.59179 -0.057978)))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                 is
                                 0)
                                ((R:segstate.parent.n.name is z)
                                 ((0.257959 -0.434801))
                                 ((0.452854 -0.19083)))
                                ((0.736967 0.0640206))))
                              ((n.name is ih_3)
                               ((0.365719 -0.552136))
                               ((R:segstate.parent.R:SylStructure.parent.position_type
                                 is
                                 final)
                                ((0.437316 -0.551263))
                                ((n.name is s_1)
                                 ((0.376626 -0.18739))
                                 ((R:segstate.parent.n.name is z)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    2.2)
                                   ((0.388832 -0.35558))
                                   ((0.816511 -0.145598)))
                                  ((R:segstate.parent.n.name is s)
                                   ((0.267137 -0.515533))
                                   ((R:segstate.parent.name is ax)
                                    ((0.0608542 -0.476843))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                      <
                                      0.3)
                                     ((0.484174 -0.273192))
                                     ((0.328778 -0.444263)))))))))))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                              <
                              1.2)
                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                               <
                               0.7)
                              ((n.name is iy_3)
                               ((R:segstate.parent.nn.ph_vfront is 1)
                                ((0.489937 -0.0595637))
                                ((0.769461 0.276606)))
                               ((R:segstate.parent.name is iy)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                  <
                                  0.1)
                                 ((0.204493 -0.625289))
                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                   <
                                   9.4)
                                  ((n.name is iy_2)
                                   ((0.0721397 -0.755262))
                                   ((0.892272 -0.371465)))
                                  ((0.783194 -0.191357))))
                                ((R:segstate.parent.lisp_onset_stop
                                  is
                                  0)
                                 ((R:segstate.parent.nn.ph_vheight
                                   is
                                   2)
                                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                                    <
                                    12)
                                   ((0.147923 -0.469551))
                                   ((0.680519 -0.284317)))
                                  ((R:segstate.parent.ph_vlng is d)
                                   ((R:segstate.parent.p.name is dh)
                                    ((0.338338 -0.416024))
                                    ((0.723383 -0.200814)))
                                   ((n.name is ih_2)
                                    ((0.22709 -0.339275))
                                    ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                      <
                                      4.5)
                                     ((R:segstate.parent.pp.ph_cplace
                                       is
                                       a)
                                      ((0.439648 -0.259125))
                                      ((R:segstate.parent.ph_vlng is a)
                                       ((0.471947 -0.24503))
                                       ((R:segstate.parent.R:SylStructure.parent.syl_out
                                         <
                                         8.2)
                                        ((0.828909 0.146756))
                                        ((0.700458 -0.111249)))))
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       12.7)
                                      ((0.708073 0.138046))
                                      ((0.454052 0.193112)))))))
                                 ((0.493616 -0.454936)))))
                              ((n.name is s_1)
                               ((0.611502 0.158313))
                               ((R:segstate.parent.nn.ph_vheight is 2)
                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                  <
                                  1.5)
                                 ((0.473295 -0.246785))
                                 ((0.470955 -0.43917)))
                                ((R:segstate.parent.nn.ph_cplace is b)
                                 ((0.304402 -0.403566))
                                 ((R:segstate.parent.pp.ph_vc is +)
                                  ((1.64177 0.365013))
                                  ((n.name is z_1)
                                   ((0.779908 0.121973))
                                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                     is
                                     0)
                                    ((R:segstate.parent.p.name is l)
                                     ((0.740434 -0.150797))
                                     ((0.82363 0.271804)))
                                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                                      <
                                      7)
                                     ((R:segstate.parent.p.name is hh)
                                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                        <
                                        2.4)
                                       ((0.33808 -0.315728))
                                       ((0.266369 -0.558949)))
                                      ((R:segstate.parent.R:SylStructure.parent.syl_out
                                        <
                                        6.9)
                                       ((0.808725 0.0308855))
                                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                         is
                                         0)
                                        ((0.622709 -0.166137))
                                        ((0.361891 -0.407083)))))
                                     ((R:segstate.parent.nn.ph_vheight
                                       is
                                       3)
                                      ((0.410728 -0.332554))
                                      ((1.01777 0.22008)))))))))))
                             ((R:segstate.parent.nn.ph_cplace is a)
                              ((R:segstate.parent.syl_final is 0)
                               ((0.348838 -0.172833))
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 6.5)
                                ((0.320502 -0.375499))
                                ((0.438004 -0.231577))))
                              ((R:segstate.parent.nn.ph_ctype is s)
                               ((0.247309 -0.393922))
                               ((0.200117 -0.521619)))))))))))))))
                  ((n.name is ng_1)
                   ((R:segstate.parent.p.name is l)
                    ((1.38032 1.31075))
                    ((R:segstate.parent.pp.ph_vrnd is -)
                     ((0.254562 -0.39799))
                     ((0.333247 -0.278481))))
                   ((n.name is s_3)
                    ((0.505637 -1.10675))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                      <
                      0.4)
                     ((n.name is eh_2)
                      ((0.570612 0.0977796))
                      ((n.name is w_1)
                       ((0.504806 -0.0147187))
                       ((R:segstate.parent.n.ph_ctype is r)
                        ((R:segstate.parent.nn.ph_ctype is f)
                         ((0.882944 -0.232935))
                         ((R:segstate.parent.R:SylStructure.parent.next_accent
                           <
                           0.9)
                          ((R:segstate.parent.nn.ph_vfront is 1)
                           ((0.209064 -0.397002))
                           ((0.34845 -0.58453)))
                          ((R:segstate.parent.name is eh)
                           ((n.name is eh_3)
                            ((0.165198 -0.683099))
                            ((4.56082e-07 -0.799691)))
                           ((R:segstate.parent.nn.ph_vfront is 3)
                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                              <
                              2.5)
                             ((0.338611 -0.638577))
                             ((0.185189 -0.789142)))
                            ((R:segstate.parent.R:SylStructure.parent.asyl_out
                              <
                              3.5)
                             ((0.478661 -0.464031))
                             ((0.305482 -0.625466)))))))
                        ((R:segstate.parent.ph_vheight is 2)
                         ((n.name is ax_2)
                          ((0.118136 -0.438396))
                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                            <
                            0.1)
                           ((0.291347 -0.381612))
                           ((n.name is ax_3)
                            ((0.279398 -0.360669))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                              is
                              0)
                             ((0.501651 -0.0633784))
                             ((0.467326 -0.238982))))))
                         ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                           <
                           9.6)
                          ((n.name is iy_2)
                           ((0.4708 -0.626784))
                           ((R:segstate.parent.nn.ph_vfront is 3)
                            ((0.546379 -0.334963))
                            ((R:segstate.parent.n.name is l)
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                               <
                               5.4)
                              ((0.090684 -0.509368))
                              ((0.274869 -0.600781)))
                             ((R:segstate.parent.R:SylStructure.parent.asyl_out
                               <
                               3.4)
                              ((0.371561 -0.508906))
                              ((0.384826 -0.392414))))))
                          ((1.08899 -0.214199)))))))
                     ((R:segstate.parent.n.name is pau)
                      ((R:segstate.parent.nn.ph_cplace is b)
                       ((3.1785 1.50795))
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                         <
                         0.9)
                        ((1.26886 0.295889))
                        ((R:segstate.parent.R:SylStructure.parent.asyl_in
                          <
                          3.6)
                         ((R:segstate.parent.R:SylStructure.parent.syl_out
                           <
                           6.6)
                          ((R:segstate.parent.ph_vlng is a)
                           ((0.449403 -0.369964))
                           ((R:segstate.parent.pp.ph_vrnd is -)
                            ((1.20737 0.366611))
                            ((0.921262 0.0156305))))
                          ((R:segstate.parent.p.ph_cvox is +)
                           ((R:segstate.parent.p.ph_ctype is f)
                            ((0.820909 0.000534906))
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              1.4)
                             ((0.46628 -0.495162))
                             ((0.551745 -0.256773))))
                           ((0.317207 -0.580597))))
                         ((0.452222 -0.624974)))))
                      ((R:segstate.parent.lisp_onset_glide is 0)
                       ((n.name is ih_3)
                        ((R:segstate.parent.lisp_coda_stop is 0)
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           single)
                          ((1.21985 0.489971))
                          ((R:segstate.parent.p.name is hh)
                           ((0.165879 -0.658935))
                           ((R:segstate.parent.n.name is l)
                            ((0.386855 -0.45019))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                              <
                              1.2)
                             ((0.454402 -0.20989))
                             ((0.644145 0.372165))))))
                         ((0.755472 1.16819)))
                        ((R:segstate.parent.lisp_coda_glide is 0)
                         ((n.name is aa_3)
                          ((0.286164 0.713974))
                          ((n.name is ae_2)
                           ((R:segstate.parent.n.ph_cvox is +)
                            ((R:segstate.parent.pp.ph_cplace is 0)
                             ((R:segstate.parent.nn.ph_ctype is s)
                              ((0.514236 -0.0901469))
                              ((R:segstate.parent.R:SylStructure.parent.syl_out
                                <
                                5.2)
                               ((0.402519 -0.426628))
                               ((0.619392 -0.268283))))
                             ((0.447872 -0.518995)))
                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                              is
                              0)
                             ((0.148447 -0.584971))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                               <
                               0.1)
                              ((0.0583328 -0.707222))
                              ((0.114087 -0.647475)))))
                           ((R:segstate.parent.syl_final is 0)
                            ((R:segstate.parent.nn.ph_ctype is f)
                             ((R:segstate.parent.R:SylStructure.parent.position_type
                               is
                               initial)
                              ((R:segstate.parent.pp.ph_cplace is a)
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 0.7)
                                ((0.226814 -0.493041))
                                ((0.327873 -0.299589)))
                               ((0.158857 -0.652174)))
                              ((n.name is d_1)
                               ((0.220291 -0.444626))
                               ((R:segstate.parent.pp.ph_vfront is 2)
                                ((0.880617 0.137914))
                                ((R:segstate.parent.nn.ph_cplace is a)
                                 ((0.435496 -0.384171))
                                 ((R:segstate.parent.R:SylStructure.parent.next_accent
                                   <
                                   2.4)
                                  ((R:segstate.parent.pp.ph_cplace
                                    is
                                    a)
                                   ((0.709945 0.125301))
                                   ((R:segstate.parent.p.name is dh)
                                    ((0.388695 -0.286855))
                                    ((0.585024 -0.0766002))))
                                  ((0.556722 -0.325279)))))))
                             ((n.name is m_1)
                              ((1.04678 0.460521))
                              ((n.name is d_1)
                               ((R:segstate.parent.pp.ph_vlng is l)
                                ((0.187953 -0.455306))
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  2.5)
                                 ((0.423029 -0.13505))
                                 ((0.23252 -0.297607))))
                               ((R:segstate.parent.name is ih)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                  is
                                  0)
                                 ((0.43942 -0.0853303))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   2.1)
                                  ((0.363297 -0.235818))
                                  ((0.137871 -0.437424))))
                                ((R:segstate.parent.p.name is dh)
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                   <
                                   3.5)
                                  ((0.336216 -0.374859))
                                  ((1.0745 0.108716)))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                   <
                                   0.1)
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                    is
                                    0)
                                   ((0.592816 -0.301772))
                                   ((0.865501 0.141462)))
                                  ((R:segstate.parent.pp.ph_cplace
                                    is
                                    a)
                                   ((R:segstate.parent.ph_vheight is 2)
                                    ((0.64642 -0.12477))
                                    ((0.859912 0.131261)))
                                   ((R:segstate.parent.ph_vlng is s)
                                    ((1.1347 0.709222))
                                    ((0.976624 0.261972))))))))))
                            ((n.name is ax_3)
                             ((R:segstate.parent.nn.ph_vheight is 1)
                              ((R:segstate.parent.n.name is l)
                               ((5.22565e-06 -0.510846))
                               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                 <
                                 2.5)
                                ((0.211869 -0.396425))
                                ((0.72779 -0.190466))))
                              ((R:segstate.parent.R:SylStructure.parent.last_accent
                                <
                                1.1)
                               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                 <
                                 2.5)
                                ((R:segstate.parent.n.ph_cvox is +)
                                 ((R:segstate.parent.nn.ph_vheight
                                   is
                                   3)
                                  ((0.587385 0.137542))
                                  ((1.00888 0.610483)))
                                 ((0.381519 -0.127199)))
                                ((R:segstate.parent.nn.ph_vheight is 3)
                                 ((0.252915 -0.373541))
                                 ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                   <
                                   1.6)
                                  ((0.688978 0.10703))
                                  ((0.395079 -0.198789)))))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                                 <
                                 0.3)
                                ((0.254633 -0.363733))
                                ((R:segstate.parent.R:SylStructure.parent.next_accent
                                  <
                                  0.4)
                                 ((0.516203 -0.0150181))
                                 ((0.29139 -0.268541))))))
                             ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                               <
                               1)
                              ((R:segstate.parent.p.name is dh)
                               ((R:segstate.parent.name is ey)
                                ((0.424817 -0.564335))
                                ((R:segstate.parent.n.ph_cvox is +)
                                 ((n.name is ax_2)
                                  ((0.176358 -0.388823))
                                  ((0.310717 -0.344205)))
                                 ((0.642184 0.00574452))))
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                 <
                                 1.2)
                                ((R:segstate.parent.R:SylStructure.parent.syl_out
                                  <
                                  7.4)
                                 ((0.343246 -0.50256))
                                 ((0.879239 -0.076308)))
                                ((R:segstate.parent.R:SylStructure.parent.next_accent
                                  <
                                  1.4)
                                 ((1.17451 0.287435))
                                 ((0.665437 -0.120805)))))
                              ((R:segstate.parent.ph_vlng is l)
                               ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                 <
                                 3.6)
                                ((n.name is iy_3)
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                   <
                                   2.2)
                                  ((0.366022 -0.528034))
                                  ((0.359009 -0.737007)))
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                   <
                                   2.2)
                                  ((n.name is iy_2)
                                   ((0.273802 -0.689407))
                                   ((R:segstate.parent.R:SylStructure.parent.position_type
                                     is
                                     initial)
                                    ((0.105011 -0.699415))
                                    ((R:segstate.parent.pp.ph_cplace
                                      is
                                      a)
                                     ((0.618678 -0.017299))
                                     ((0.341316 -0.457056)))))
                                  ((R:segstate.parent.pp.ph_cplace
                                    is
                                    0)
                                   ((1.10368 -0.0578008))
                                   ((0.330975 -0.558062)))))
                                ((0.50279 -0.202393)))
                               ((R:segstate.parent.pp.ph_vlng is s)
                                ((R:segstate.parent.p.ph_ctype is n)
                                 ((0.39144 -0.956087))
                                 ((n.name is er_2)
                                  ((0.334631 -0.656345))
                                  ((R:segstate.parent.n.name is l)
                                   ((0.277298 -0.49926))
                                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                                     <
                                     1.6)
                                    ((0.948076 -0.105899))
                                    ((0.33636 -0.372792))))))
                                ((R:segstate.parent.p.name is sh)
                                 ((0.912036 0.182461))
                                 ((R:segstate.parent.R:SylStructure.parent.asyl_in
                                   <
                                   4.2)
                                  ((R:segstate.parent.p.name is f)
                                   ((R:segstate.parent.ph_vlng is a)
                                    ((R:segstate.parent.R:SylStructure.parent.next_accent
                                      <
                                      1.5)
                                     ((0.176831 -0.51371))
                                     ((0.209574 -0.648152)))
                                    ((R:segstate.parent.n.name is l)
                                     ((0.382584 -0.432087))
                                     ((0.642023 -0.054003))))
                                   ((n.name is er_3)
                                    ((0.743082 0.0978591))
                                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                      <
                                      2.2)
                                     ((R:segstate.parent.R:SylStructure.parent.position_type
                                       is
                                       final)
                                      ((0.580903 -0.479934))
                                      ((R:segstate.parent.pp.ph_vheight
                                        is
                                        3)
                                       ((0.80868 0.0969661))
                                       ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                         <
                                         4)
                                        ((R:segstate.parent.n.name
                                          is
                                          d)
                                         ((0.967076 -0.033047))
                                         ((R:segstate.parent.p.name
                                           is
                                           dh)
                                          ((n.name is ax_2)
                                           ((R:segstate.parent.n.name
                                             is
                                             k)
                                            ((0.351684 -0.104108))
                                            ((R:segstate.parent.nn.ph_vfront
                                              is
                                              1)
                                             ((0.267906 -0.314464))
                                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                                               <
                                               1.1)
                                              ((3.81061e-06 -0.500362))
                                              ((0.115454 -0.444595)))))
                                           ((R:segstate.parent.n.ph_cplace
                                             is
                                             a)
                                            ((0.647834 0.0567234))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                              <
                                              1.2)
                                             ((R:segstate.parent.n.ph_ctype
                                               is
                                               s)
                                              ((0.145974 -0.391109))
                                              ((0.323221 -0.257627)))
                                             ((0.371842 -0.135837)))))
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                            <
                                            0.3)
                                           ((R:segstate.parent.p.ph_cvox
                                             is
                                             +)
                                            ((0.319933 -0.439838))
                                            ((0.213443 -0.610917)))
                                           ((R:segstate.parent.R:SylStructure.parent.syl_in
                                             <
                                             4.4)
                                            ((0.314393 -0.507175))
                                            ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                              is
                                              0)
                                             ((0.334629 -0.389948))
                                             ((0.651266 -0.0138087)))))))
                                        ((R:segstate.parent.pp.ph_cplace
                                          is
                                          a)
                                         ((R:segstate.parent.nn.ph_vheight
                                           is
                                           1)
                                          ((0.253465 -0.477081))
                                          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                            is
                                            0)
                                           ((0.305373 -0.317438))
                                           ((0.48249 -0.0612019))))
                                         ((R:segstate.parent.name
                                           is
                                           ax)
                                          ((R:segstate.parent.pp.ph_cplace
                                            is
                                            b)
                                           ((0.269504 -0.293786))
                                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                             is
                                             0)
                                            ((0.580727 0.00713917))
                                            ((0.389607 -0.177461))))
                                          ((1.06359 0.208389)))))))
                                     ((R:segstate.parent.R:SylStructure.parent.syl_out
                                       <
                                       2.6)
                                      ((0.548596 -0.302295))
                                      ((R:segstate.parent.nn.ph_vfront
                                        is
                                        1)
                                       ((0.381267 -0.373454))
                                       ((0.27936 -0.613319)))))))
                                  ((0.274593 -0.513252)))))))))))
                         ((n.name is ao_2)
                          ((R:segstate.parent.pp.ph_vrnd is -)
                           ((1.31118 1.17781))
                           ((R:segstate.parent.nn.ph_ctype is f)
                            ((0.385879 -0.516283))
                            ((0.836897 -0.0236283))))
                          ((n.name is eh_2)
                           ((R:segstate.parent.R:SylStructure.parent.asyl_out
                             <
                             2.5)
                            ((0.563215 -0.30687))
                            ((0.79939 0.105788)))
                           ((R:segstate.parent.ph_vheight is 3)
                            ((R:segstate.parent.n.name is r)
                             ((n.name is ao_3)
                              ((R:segstate.parent.nn.ph_cplace is l)
                               ((0.281969 -0.540835))
                               ((R:segstate.parent.pp.ph_vc is +)
                                ((0.110709 -0.649751))
                                ((0.0510514 -0.731439))))
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                1.1)
                               ((R:segstate.parent.name is aa)
                                ((0.093764 -0.748454))
                                ((0.0271464 -0.516721)))
                               ((R:segstate.parent.nn.ph_cplace is l)
                                ((0.167671 -0.588283))
                                ((R:segstate.parent.name is aa)
                                 ((0.434471 -0.324989))
                                 ((R:segstate.parent.R:SylStructure.parent.syl_out
                                   <
                                   6.2)
                                  ((0.0359117 -0.510588))
                                  ((0.0647328 -0.477002)))))))
                             ((0.676516 -0.268237)))
                            ((R:segstate.parent.name is ow)
                             ((0.441415 -0.724394))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                               <
                               0.3)
                              ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                <
                                4.5)
                               ((R:segstate.parent.name is ih)
                                ((0.080615 -0.527876))
                                ((R:segstate.parent.p.ph_cvox is +)
                                 ((0.41757 -0.36541))
                                 ((0.699483 -0.456688))))
                               ((0.207275 -0.592477)))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                is
                                0)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 3.6)
                                ((0.971029 0.125402))
                                ((0.518522 -0.339425)))
                               ((n.name is r_1)
                                ((0.192197 -0.616935))
                                ((R:segstate.parent.n.name is l)
                                 ((n.name is l_1)
                                  ((0.380325 -0.356997))
                                  ((0.167895 -0.584767)))
                                 ((0.638723 -0.0631891))))))))))))
                       ((n.name is iy_2)
                        ((R:segstate.parent.n.ph_cplace is l)
                         ((0.451302 -0.0228217))
                         ((0.650467 0.194257)))
                        ((n.name is ay_3)
                         ((R:segstate.parent.R:SylStructure.parent.position_type
                           is
                           single)
                          ((0.205493 -0.901766))
                          ((0.492528 -0.53205)))
                         ((n.name is ih_3)
                          ((R:segstate.parent.pp.ph_vc is +)
                           ((0.0801191 -0.720045))
                           ((0.313387 -0.615851)))
                          ((R:segstate.parent.n.name is g)
                           ((0.744677 -0.0735329))
                           ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                             <
                             1.2)
                            ((R:segstate.parent.ph_vrnd is +)
                             ((R:segstate.parent.n.ph_ctype is s)
                              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                                <
                                1.2)
                               ((0.211584 -0.433113))
                               ((0.092023 -0.465209)))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((0.332971 -0.601686))
                               ((0.159248 -0.707053))))
                             ((R:segstate.parent.n.ph_cplace is a)
                              ((R:segstate.parent.ph_vlng is s)
                               ((R:segstate.parent.R:SylStructure.parent.position_type
                                 is
                                 initial)
                                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.lisp_cg_break
                                  <
                                  0.3)
                                 ((0.0713057 -0.565947))
                                 ((0.303129 -0.538898)))
                                ((0.468028 -0.362444)))
                               ((R:segstate.parent.p.name is n)
                                ((0.39113 -0.488571))
                                ((R:segstate.parent.lisp_coda_stop
                                  is
                                  0)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                                   <
                                   2)
                                  ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                                    <
                                    3.1)
                                   ((0.5759 -0.0717097))
                                   ((0.414219 -0.299194)))
                                  ((0.294899 -0.412749)))
                                 ((0.709698 0.0583593)))))
                              ((R:segstate.parent.n.name is jh)
                               ((0.283541 -0.216139))
                               ((n.name is k_1)
                                ((0.279266 -0.341248))
                                ((R:segstate.parent.R:SylStructure.parent.next_accent
                                  <
                                  1.8)
                                 ((R:segstate.parent.R:SylStructure.parent.last_accent
                                   <
                                   0.6)
                                  ((n.name is p_1)
                                   ((0.141024 -0.402371))
                                   ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                     <
                                     2.4)
                                    ((R:segstate.parent.n.ph_ctype
                                      is
                                      s)
                                     ((0.0597874 -0.525835))
                                     ((0.222186 -0.429245)))
                                    ((R:segstate.parent.nn.ph_cplace
                                      is
                                      0)
                                     ((0.328739 -0.572484))
                                     ((0.0820569 -0.50015)))))
                                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                                    is
                                    0)
                                   ((0.452845 -0.261688))
                                   ((0.210804 -0.461202))))
                                 ((R:segstate.parent.pp.ph_vrnd is -)
                                  ((0.242021 -0.499047))
                                  ((0.211748 -0.66831))))))))
                            ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                              <
                              2.2)
                             ((R:segstate.parent.nn.ph_ctype is 0)
                              ((R:segstate.parent.R:SylStructure.parent.position_type
                                is
                                single)
                               ((0.964558 0.22941))
                               ((0.544688 -0.258243)))
                              ((R:segstate.parent.n.ph_ctype is s)
                               ((R:segstate.parent.R:SylStructure.parent.next_accent
                                 <
                                 1.4)
                                ((0.532004 -0.580736))
                                ((0.534497 -0.36818)))
                               ((0.596004 -0.142305))))
                             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                               <
                               0.3)
                              ((0.246822 -0.635306))
                              ((0.567375 -0.287251))))))))))))))))))))))))
        ((n.name is aa_2)
         ((R:segstate.parent.p.name is m)
          ((1.41953 -0.123932))
          ((R:segstate.parent.R:SylStructure.parent.syl_out < 5.2)
           ((1.06489 1.80124))
           ((0.808314 1.28224))))
         ((n.name is ah_3)
          ((R:segstate.parent.R:SylStructure.parent.syl_in < 6)
           ((0.805536 0.867949))
           ((0.669824 1.38128)))
          ((R:segstate.parent.name is ae)
           ((R:segstate.parent.R:SylStructure.parent.lisp_cg_break
             <
             0.1)
            ((R:segstate.parent.n.ph_cvox is +)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
               <
               0.3)
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 5.6)
               ((1.00704 0.139004))
               ((1.46593 0.346232)))
              ((1.59331 0.708319)))
             ((1.17116 -0.115987)))
            ((n.name is ae_2)
             ((1.77223 1.63349))
             ((R:segstate.parent.R:SylStructure.parent.ssyl_out < 3.4)
              ((1.81855 1.06332))
              ((1.4622 0.554818)))))
           ((R:segstate.parent.p.name is n)
            ((R:segstate.parent.lisp_coda_nasal is 0)
             ((n.name is ow_2)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                <
                0.3)
               ((0.446065 0.730872))
               ((1.02143 1.29749)))
              ((R:segstate.parent.n.seg_onsetcoda is coda)
               ((n.name is ax_2)
                ((1.47329 1.35504))
                ((R:segstate.parent.R:SylStructure.parent.position_type
                  is
                  single)
                 ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                   <
                   2.2)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    1.5)
                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                     <
                     5.5)
                    ((1.54175 1.4626))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.syl_onsetsize
                      <
                      0.1)
                     ((1.1604 0.992446))
                     ((1.47522 0.387782))))
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     0.8)
                    ((R:segstate.parent.R:SylStructure.parent.syl_in
                      <
                      3.3)
                     ((0.562948 0.216324))
                     ((0.830009 -0.0390829)))
                    ((R:segstate.parent.name is ow)
                     ((0.952354 0.115038))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                       <
                       1.2)
                      ((1.11595 0.331214))
                      ((R:segstate.parent.nn.ph_ctype is 0)
                       ((1.75027 1.21541))
                       ((R:segstate.parent.n.ph_cvox is -)
                        ((R:segstate.parent.R:SylStructure.parent.asyl_out
                          <
                          2.2)
                         ((0.990157 0.582198))
                         ((0.58274 0.373906)))
                        ((1.29575 0.879164))))))))
                  ((0.685779 -0.0414944)))
                 ((R:segstate.parent.ph_vheight is 1)
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                    <
                    5.4)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                     is
                     0)
                    ((1.22306 0.616966))
                    ((1.22645 0.113252)))
                   ((0.733452 0.0774622)))
                  ((R:segstate.parent.n.ph_ctype is f)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((0.812357 0.356619))
                    ((0.459078 0.0505076)))
                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                     <
                     8)
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      1.2)
                     ((0.711339 -0.239134))
                     ((0.761254 0.146641)))
                    ((0.385974 -0.363712)))))))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                 is
                 0)
                ((R:segstate.parent.ph_vlng is d)
                 ((R:segstate.parent.ph_vheight is 2)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    5.5)
                   ((0.993601 -0.396145))
                   ((0.671649 -0.10181)))
                  ((0.836734 0.114634)))
                 ((n.name is eh_2)
                  ((0.489487 -0.393144))
                  ((R:segstate.parent.R:SylStructure.parent.next_accent
                    <
                    0.9)
                   ((1.57211 0.767597))
                   ((R:segstate.parent.ph_vlng is l)
                    ((0.895172 -0.129976))
                    ((R:segstate.parent.name is ax)
                     ((1.37555 0.66538))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                       <
                       0.1)
                      ((1.00864 0.504038))
                      ((R:segstate.parent.R:SylStructure.parent.asyl_out
                        <
                        2.5)
                       ((R:segstate.parent.n.ph_cplace is b)
                        ((1.02976 0.150322))
                        ((1.34532 0.462147)))
                       ((0.596607 0.0386173)))))))))
                ((R:segstate.parent.name is uw)
                 ((1.16487 0.299288))
                 ((R:segstate.parent.n.ph_cplace is l)
                  ((R:segstate.parent.R:SylStructure.parent.syl_out
                    <
                    5.4)
                   ((0.444397 -0.575429))
                   ((0.475314 -0.281501)))
                  ((n.name is ow_3)
                   ((0.53456 -0.511643))
                   ((R:segstate.parent.R:SylStructure.parent.syl_out
                     <
                     3.1)
                    ((0.743454 0.165207))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                      <
                      4.4)
                     ((0.400586 -0.40638))
                     ((R:segstate.parent.n.ph_ctype is f)
                      ((0.695564 0.15712))
                      ((1.19252 -0.273766)))))))))))
             ((R:segstate.parent.n.name is n)
              ((0.94065 0.0436408))
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                is
                0)
               ((0.740889 -0.248344))
               ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                 <
                 2.2)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                  is
                  0)
                 ((0.117998 -0.617423))
                 ((0.369304 -0.474556)))
                ((0.0930332 -0.641192))))))
            ((n.name is ao_2)
             ((0.748647 1.00141))
             ((n.name is ow_2)
              ((0.3018 -0.837056))
              ((n.name is ow_3)
               ((0.522899 0.674442))
               ((n.name is ax_2)
                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                  <
                  3.6)
                 ((R:segstate.parent.R:SylStructure.parent.position_type
                   is
                   final)
                  ((1.78164 1.26566))
                  ((0.776902 0.388461)))
                 ((0.487517 -0.266132)))
                ((n.name is ay_2)
                 ((R:segstate.parent.n.seg_onsetcoda is coda)
                  ((1.23741 0.868107))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                    is
                    0)
                   ((0.364847 -0.155594))
                   ((0.521667 0.176348))))
                 ((R:segstate.parent.name is eh)
                  ((n.name is eh_2)
                   ((R:segstate.parent.R:SylStructure.parent.next_accent
                     <
                     1.4)
                    ((0.106041 -0.674419))
                    ((1.68683 0.0199802)))
                   ((R:segstate.parent.R:SylStructure.parent.position_type
                     is
                     single)
                    ((R:segstate.parent.pp.ph_vrnd is -)
                     ((1.6018 1.10749))
                     ((1.10383 0.454506)))
                    ((n.name is eh_3)
                     ((0.97852 -0.00687433))
                     ((1.04456 0.392898)))))
                  ((n.name is er_3)
                   ((1.44498 0.759621))
                   ((R:segstate.parent.ph_vheight is 1)
                    ((n.name is ih_2)
                     ((R:segstate.parent.n.ph_ctype is f)
                      ((0.299476 -0.354555))
                      ((2.14386e-06 -0.476819)))
                     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                       is
                       0)
                      ((R:segstate.parent.name is iy)
                       ((0.684265 0.0523066))
                       ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                         <
                         4.2)
                        ((1.53118 0.904124))
                        ((0.99057 0.203984))))
                      ((n.name is ih_3)
                       ((0.098901 -0.732038))
                       ((R:segstate.parent.name is iy)
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                          <
                          3.4)
                         ((0.575306 -0.395973))
                         ((0.736679 0.130031)))
                        ((R:segstate.parent.R:SylStructure.parent.asyl_out
                          <
                          1.4)
                         ((0.836897 -0.00218127))
                         ((1.20487 0.821327)))))))
                    ((R:segstate.parent.p.name is t)
                     ((R:segstate.parent.R:SylStructure.parent.last_accent
                       <
                       0.5)
                      ((0.881247 -0.582897))
                      ((0.638873 -0.703696)))
                     ((R:segstate.parent.lisp_coda_glide is 0)
                      ((n.name is ay_3)
                       ((R:segstate.parent.nn.ph_vc is +)
                        ((R:segstate.parent.n.ph_cvox is +)
                         ((0.63233 -0.267424))
                         ((0.45444 -0.657398)))
                        ((0.281328 -0.745713)))
                       ((R:segstate.parent.ph_vlng is l)
                        ((1.08108 0.374918))
                        ((n.name is ey_3)
                         ((R:segstate.parent.n.ph_cplace is a)
                          ((0.875046 0.690131))
                          ((0.563235 -0.488945)))
                         ((R:segstate.parent.n.ph_cvox is +)
                          ((R:segstate.parent.p.seg_onsetcoda is coda)
                           ((R:segstate.parent.R:SylStructure.parent.position_type
                             is
                             single)
                            ((0.889447 -0.272996))
                            ((1.49756 0.350984)))
                           ((R:segstate.parent.name is ay)
                            ((0.792115 -0.740951))
                            ((R:segstate.parent.nn.ph_cplace is a)
                             ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                               <
                               2.2)
                              ((0.255955 -0.565632))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
                                is
                                0)
                               ((0.287575 -0.371574))
                               ((0.124127 -0.486194))))
                             ((R:segstate.parent.pp.ph_vlng is a)
                              ((1.00506 -0.0495392))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                is
                                0)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                 is
                                 0)
                                ((R:segstate.parent.nn.ph_vc is +)
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.lisp_cg_break
                                   <
                                   0.1)
                                  ((0.551279 0.0189631))
                                  ((0.345172 -0.32652)))
                                 ((0.3109 -0.404737)))
                                ((0.341449 -0.426187)))
                               ((0.271424 -0.471701)))))))
                          ((R:segstate.parent.ph_vrnd is -)
                           ((n.name is ah_2)
                            ((0.182496 -0.43726))
                            ((R:segstate.parent.n.ph_cplace is p)
                             ((0.497649 0.452395))
                             ((n.name is ax_3)
                              ((0.963063 0.428615))
                              ((R:segstate.parent.ph_vheight is 2)
                               ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                 <
                                 1.7)
                                ((0.288616 -0.411666))
                                ((R:segstate.parent.R:SylStructure.parent.asyl_out
                                  <
                                  3.2)
                                 ((0.860988 0.12795))
                                 ((0.441386 -0.259822))))
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_out
                                 <
                                 2.6)
                                ((0.568588 0.054224))
                                ((0.714607 0.349825)))))))
                           ((R:segstate.parent.n.ph_ctype is s)
                            ((0.421017 -0.55444))
                            ((R:segstate.parent.R:SylStructure.parent.syl_in
                              <
                              5.4)
                             ((0.685252 -0.357082))
                             ((0.699671 0.00487078)))))))))
                      ((R:segstate.parent.R:SylStructure.parent.position_type
                        is
                        single)
                       ((R:segstate.parent.R:SylStructure.parent.asyl_out
                         <
                         2.5)
                        ((0.530448 -0.647414))
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                          <
                          1.4)
                         ((0.975867 -0.311142))
                         ((0.285592 -0.523228))))
                       ((0.446618 -0.290102))))))))))))))))))))))))
 ((n.name is pau_1)
  ((R:segstate.parent.ph_cvox is -)
   ((R:segstate.parent.pp.ph_cplace is 0)
    ((R:segstate.parent.pp.ph_vheight is 2)
     ((R:segstate.parent.R:SylStructure.parent.syl_in < 7.4)
      ((0.771691 0.0388009))
      ((R:segstate.parent.name is t)
       ((0.357285 -0.298681))
       ((0.430498 -0.499423))))
     ((R:segstate.parent.name is t)
      ((0.356214 -0.396517))
      ((R:segstate.parent.p.ph_ctype is s)
       ((0.158787 -0.765137))
       ((0.348157 -0.696292)))))
    ((R:segstate.parent.R:SylStructure.parent.position_type is final)
     ((R:segstate.parent.p.name is ax)
      ((0.719832 0.0935795))
      ((2.25553 0.913064)))
     ((R:segstate.parent.ph_cplace is a)
      ((R:segstate.parent.p.ph_vheight is 3)
       ((0.263428 -0.506416))
       ((R:segstate.parent.p.ph_vheight is 2)
        ((0.390147 -0.0112197))
        ((0.725846 -0.226149))))
      ((R:segstate.parent.lisp_onset_stop is 0)
       ((R:segstate.parent.pp.ph_cplace is a)
        ((0.671644 0.155202))
        ((1.08814 0.396395)))
       ((0.701549 -0.286021))))))
   ((R:segstate.parent.lisp_coda_stop is 0)
    ((R:segstate.parent.name is ax)
     ((0.209453 -0.407011))
     ((R:segstate.parent.name is l)
      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
        is
        0)
       ((1.21982 3.36162))
       ((1.25734 4.07053)))
      ((R:segstate.parent.ph_vlng is d)
       ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.3)
        ((1.66395 2.13509))
        ((R:segstate.parent.pp.ph_vrnd is -)
         ((0.625397 1.03105))
         ((0.940372 1.4905))))
       ((R:segstate.parent.nn.ph_vc is 0)
        ((R:segstate.parent.name is er)
         ((R:segstate.parent.R:SylStructure.parent.syl_in < 11)
          ((0.828001 3.15244))
          ((0.481644 2.8237)))
         ((R:segstate.parent.p.ph_vlng is a)
          ((R:segstate.parent.pp.ph_cplace is p)
           ((1.05357 3.26302))
           ((R:segstate.parent.lisp_onset_stop is 0)
            ((0.889132 2.53648))
            ((0.784144 2.94112))))
          ((R:segstate.parent.pp.ph_cplace is g)
           ((1.25462 2.84359))
           ((R:segstate.parent.p.ph_ctype is f)
            ((0.621918 2.88714))
            ((R:segstate.parent.p.ph_ctype is n)
             ((0.760165 2.63388))
             ((R:segstate.parent.p.ph_vheight is 2)
              ((R:segstate.parent.ph_cplace is a)
               ((0.936886 2.20419))
               ((1.30066 2.87052)))
              ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 2.4)
               ((1.18357 1.60699))
               ((R:segstate.parent.R:SylStructure.parent.syl_in < 10.4)
                ((R:segstate.parent.pos_in_syl is 2)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                   is
                   0)
                  ((1.32251 2.69068))
                  ((0.79854 2.31931)))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                   is
                   0)
                  ((0.800115 1.76916))
                  ((1.0476 2.36953))))
                ((R:segstate.parent.R:SylStructure.parent.syl_in
                  <
                  13.7)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                   is
                   0)
                  ((0.621716 2.07689))
                  ((0.974279 1.22267)))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((0.998334 2.41236))
                  ((0.754509 2.05621))))))))))))
        ((R:segstate.parent.lisp_onset_glide is 0)
         ((R:segstate.parent.R:SylStructure.parent.next_accent < 1.6)
          ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
            is
            0)
           ((1.24029 2.67757))
           ((1.47454 2.16071)))
          ((1.2066 1.74183)))
         ((1.2906 1.52023)))))))
    ((R:segstate.parent.p.ph_ctype is n)
     ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 4.6)
      ((1.10831 1.62598))
      ((1.67433 2.98422)))
     ((R:segstate.parent.p.ph_ctype is 0)
      ((R:segstate.parent.p.ph_vlng is s)
       ((0.846886 0.0645237))
       ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.2)
        ((1.07188 0.38341))
        ((1.13852 0.810281))))
      ((R:segstate.parent.R:SylStructure.parent.syl_in < 9.5)
       ((1.26189 0.97769))
       ((1.08621 1.4772)))))))
  ((R:segstate.parent.seg_onsetcoda is coda)
   ((R:segstate.parent.name is z)
    ((R:segstate.parent.p.ph_ctype is 0)
     ((n.name is z_3) ((0.532124 -0.213255)) ((0.765331 0.374971)))
     ((R:segstate.parent.pp.ph_vlng is s)
      ((0.761125 -0.567289))
      ((R:segstate.parent.R:SylStructure.parent.sub_phrases < 0.2)
       ((R:segstate.parent.R:SylStructure.parent.syl_in < 8.7)
        ((0.506655 -0.740777))
        ((n.name is z_2) ((0.298692 -1.05361)) ((0.44729 -0.872594))))
       ((0.246472 -1.11289)))))
    ((n.name is n_3)
     ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
       is
       0)
      ((R:segstate.parent.n.ph_cvox is +)
       ((R:segstate.parent.R:SylStructure.parent.syl_in < 10.6)
        ((1.29881 1.2366))
        ((1.5221 -0.521966)))
       ((R:segstate.parent.p.ph_vheight is 1)
        ((0.858339 -0.067307))
        ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.4)
         ((0.924637 -0.419541))
         ((R:segstate.parent.n.name is pau)
          ((R:segstate.parent.R:SylStructure.parent.asyl_in < 3.5)
           ((0.612744 -0.764951))
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
             is
             0)
            ((0.62033 -0.961607))
            ((0.290421 -1.20365))))
          ((0.366353 -1.24298))))))
      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
        is
        0)
       ((1.41439 0.664595))
       ((1.54004 1.06642))))
     ((R:segstate.parent.p.name is ax)
      ((n.name is l_3)
       ((0.132131 -0.622426))
       ((R:segstate.parent.pos_in_syl is 3)
        ((R:segstate.parent.R:SylStructure.parent.syl_in < 10.5)
         ((2.06739 2.52794))
         ((0.936191 1.83372)))
        ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 9.1)
         ((R:segstate.parent.n.name is s)
          ((1.55307 1.67089))
          ((R:segstate.parent.nn.ph_ctype is f)
           ((2.58366 1.73608))
           ((R:segstate.parent.lisp_onset_nasal is 0)
            ((R:segstate.parent.lisp_coda_stop is 0)
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
               <
               0.1)
              ((R:segstate.parent.ph_cplace is a)
               ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 5.8)
                ((1.33051 1.47466))
                ((1.11466 1.05027)))
               ((1.18168 0.694968)))
              ((1.0025 0.220389)))
             ((R:segstate.parent.pp.ph_cplace is a)
              ((R:segstate.parent.nn.ph_vc is 0)
               ((0.957343 0.343786))
               ((1.97671 0.965655)))
              ((0.53248 -0.354203))))
            ((1.59734 1.5549)))))
         ((1.32101 1.84776)))))
      ((R:segstate.parent.R:SylStructure.parent.position_type is final)
       ((n.name is l_2)
        ((0.751217 1.4681))
        ((R:segstate.parent.n.name is l)
         ((n.name is l_1)
          ((R:segstate.parent.name is eh)
           ((0.876605 1.27064))
           ((0.441571 -0.381077)))
          ((n.name is eh_3)
           ((0.092213 -0.805412))
           ((R:segstate.parent.nn.ph_ctype is f)
            ((0.660322 0.0139937))
            ((R:segstate.parent.pp.ph_vheight is 2)
             ((0.361157 -0.407505))
             ((0.351338 -0.682737))))))
         ((R:segstate.parent.R:SylStructure.parent.syl_codasize < 3.1)
          ((n.name is n_1)
           ((R:segstate.parent.ph_vlng is a)
            ((R:segstate.parent.pp.ph_vheight is 2)
             ((R:segstate.parent.p.name is sh)
              ((0.132079 -0.414817))
              ((0.0764798 -0.462654)))
             ((R:segstate.parent.R:SylStructure.parent.syl_in < 7.7)
              ((0.107892 -0.436102))
              ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                <
                2.3)
               ((0.456401 -0.253463))
               ((0.264936 -0.376092)))))
            ((1.28584 0.643513)))
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
             <
             0.3)
            ((n.name is ng_3)
             ((0.80135 -0.940692))
             ((n.name is t_2)
              ((R:segstate.parent.p.name is n)
               ((2.177 2.5168))
               ((1.59426 0.576563)))
              ((R:segstate.parent.p.ph_vheight is 1)
               ((R:segstate.parent.nn.ph_vc is -)
                ((R:segstate.parent.R:SylStructure.parent.last_accent
                  <
                  1.2)
                 ((R:segstate.parent.n.ph_ctype is f)
                  ((0.821442 0.57105))
                  ((0.827401 0.0944673)))
                 ((1.79652 0.939945)))
                ((R:segstate.parent.ph_cplace is v)
                 ((0.711701 0.490489))
                 ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                   <
                   2.2)
                  ((1.31186 0.679373))
                  ((2.12704 1.74971)))))
               ((R:segstate.parent.p.ph_ctype is s)
                ((n.name is d_1)
                 ((0.46251 -0.259707))
                 ((R:segstate.parent.n.name is ng)
                  ((0.491582 -0.236572))
                  ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                    <
                    2.3)
                   ((n.name is ax_3)
                    ((2.68454 1.82662))
                    ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                      <
                      7.2)
                     ((n.name is ax_2)
                      ((1.72923 1.78618))
                      ((R:segstate.parent.pp.ph_vc is +)
                       ((R:segstate.parent.ph_vlng is a)
                        ((1.07237 0.595814))
                        ((1.87573 1.37102)))
                       ((R:segstate.parent.ph_vheight is 2)
                        ((1.22578 0.626121))
                        ((0.741151 0.263738)))))
                     ((R:segstate.parent.p.ph_cvox is +)
                      ((0.971859 0.494795))
                      ((1.11551 0.11465)))))
                   ((R:segstate.parent.name is iy)
                    ((0.524628 0.0759745))
                    ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                      <
                      2)
                     ((R:segstate.parent.pp.ph_vlng is a)
                      ((1.44815 1.00984))
                      ((R:segstate.parent.n.name is d)
                       ((1.44169 0.629779))
                       ((0.838016 0.175041))))
                     ((1.16507 0.161466)))))))
                ((n.name is s_3)
                 ((0.516481 1.59138))
                 ((R:segstate.parent.ph_vheight is 3)
                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                    <
                    2)
                   ((1.09418 0.446311))
                   ((1.51066 1.1556)))
                  ((n.name is iy_2)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((0.855857 0.955962))
                    ((0.683371 0.471636)))
                   ((n.name is ax_3)
                    ((R:segstate.parent.p.ph_ctype is f)
                     ((R:segstate.parent.p.name is sh)
                      ((0.934978 0.056286))
                      ((1.80955 2.23528)))
                     ((R:segstate.parent.n.ph_ctype is n)
                      ((0.257656 -0.410728))
                      ((1.64773 0.447436))))
                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                      <
                      3)
                     ((R:segstate.parent.ph_vlng is l)
                      ((R:segstate.parent.n.name is pau)
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                         is
                         0)
                        ((0.567436 -0.340523))
                        ((0.394988 -0.621303)))
                       ((0.917399 0.140989)))
                      ((n.name is d_2)
                       ((0.307342 -0.379701))
                       ((R:segstate.parent.n.name is d)
                        ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                          <
                          3.9)
                         ((0.787521 0.154709))
                         ((1.57911 0.68655)))
                        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                          is
                          0)
                         ((R:segstate.parent.R:SylStructure.parent.asyl_in
                           <
                           3.6)
                          ((R:segstate.parent.lisp_coda_stop is 0)
                           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                             <
                             1.1)
                            ((R:segstate.parent.R:SylStructure.parent.asyl_in
                              <
                              2)
                             ((R:segstate.parent.p.ph_cvox is +)
                              ((0.616434 -0.0911948))
                              ((0.618256 0.0354063)))
                             ((0.851244 0.200389)))
                            ((1.39242 0.588859)))
                           ((R:segstate.parent.pos_in_syl is 1)
                            ((2.06454 1.41053))
                            ((0.988989 0.165815))))
                          ((R:segstate.parent.pos_in_syl is 1)
                           ((R:segstate.parent.n.ph_cvox is +)
                            ((0.723495 -0.0329912))
                            ((0.50207 -0.332795)))
                           ((R:segstate.parent.lisp_coda_nasal is 0)
                            ((1.13102 0.494036))
                            ((0.841182 -0.0530174)))))
                         ((R:segstate.parent.pos_in_syl is 2)
                          ((R:segstate.parent.R:SylStructure.parent.syl_in
                            <
                            12.5)
                           ((1.65667 0.689377))
                           ((0.999771 -0.168471)))
                          ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                            <
                            1.2)
                           ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                             <
                             6)
                            ((1.5763 0.409813))
                            ((1.0922 0.252936)))
                           ((R:segstate.parent.p.seg_onsetcoda is coda)
                            ((R:segstate.parent.lisp_coda_stop is 0)
                             ((R:segstate.parent.R:SylStructure.parent.asyl_in
                               <
                               2.5)
                              ((0.887986 -0.040137))
                              ((0.373732 -0.346594)))
                             ((0.588385 -0.496052)))
                            ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                              <
                              0.1)
                             ((R:segstate.parent.n.name is n)
                              ((0.438315 -0.417541))
                              ((R:segstate.parent.pp.ph_vlng is s)
                               ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                 <
                                 8)
                                ((0.455526 -0.357234))
                                ((0.722553 -0.0634827)))
                               ((R:segstate.parent.pp.ph_vrnd is -)
                                ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                                  <
                                  8)
                                 ((1.75795 0.486804))
                                 ((0.967107 0.0938014)))
                                ((0.524519 -0.125369)))))
                             ((0.543706 -0.326224))))))))))
                     ((R:segstate.parent.pp.ph_vrnd is -)
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        5.5)
                       ((2.48662 1.20124))
                       ((1.31308 0.741872)))
                      ((R:segstate.parent.lisp_coda_glide is 0)
                       ((R:segstate.parent.p.ph_ctype is 0)
                        ((0.915398 0.494896))
                        ((0.872578 -0.111422)))
                       ((0.621209 -0.30364)))))))))))))
            ((n.name is ax_2)
             ((R:segstate.parent.n.ph_ctype is n)
              ((1.63889 1.48092))
              ((2.64117 2.56695)))
             ((R:segstate.parent.ph_cplace is a)
              ((R:segstate.parent.lisp_onset_stop is 0)
               ((1.01674 -0.392331))
               ((1.31908 0.341216)))
              ((R:segstate.parent.n.ph_cvox is +)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                 is
                 0)
                ((1.32885 0.903149))
                ((R:segstate.parent.pp.ph_vlng is d)
                 ((0.511487 -0.249164))
                 ((n.name is ax_3)
                  ((1.16527 0.747791))
                  ((1.08325 0.0743735)))))
               ((R:segstate.parent.ph_vlng is a)
                ((2.7997 2.09885))
                ((R:segstate.parent.R:SylStructure.parent.asyl_in
                  <
                  0.4)
                 ((1.64953 1.0768))
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                   is
                   0)
                  ((0.919471 0.353352))
                  ((0.769474 0.763985))))))))))
          ((R:segstate.parent.n.ph_ctype is f)
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
             is
             0)
            ((R:segstate.parent.R:SylStructure.parent.asyl_in < 2.4)
             ((R:segstate.parent.R:SylStructure.parent.syl_in < 7.2)
              ((0.529979 -0.349272))
              ((0.526037 -0.541479)))
             ((0.24293 -0.648794)))
            ((1.14261 -0.0728756)))
           ((R:segstate.parent.n.ph_cplace is a)
            ((R:segstate.parent.p.ph_cvox is +)
             ((1.36863 0.316127))
             ((0.993165 0.507843)))
            ((1.00935 -0.314269)))))))
       ((R:segstate.parent.lisp_coda_stop is 0)
        ((n.name is l_3)
         ((0.332779 -0.617248))
         ((n.name is s_3)
          ((1.08157 2.0999))
          ((n.name is ih_2)
           ((R:segstate.parent.n.ph_cplace is a)
            ((2.9966 0.606067))
            ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 5.4)
             ((0.179491 -0.348444))
             ((0.186067 -0.261063))))
           ((n.name is eh_2)
            ((R:segstate.parent.n.ph_ctype is n)
             ((0.456632 -0.356728))
             ((0.772242 0.118592)))
            ((R:segstate.parent.lisp_onset_glide is 0)
             ((n.name is ay_2)
              ((1.6471 2.46785))
              ((R:segstate.parent.R:SylStructure.parent.syl_codasize
                <
                1.3)
               ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 5.5)
                ((R:segstate.parent.R:SylStructure.parent.sub_phrases
                  <
                  0.2)
                 ((R:segstate.parent.pp.ph_cplace is 0)
                  ((1.59926 2.00634))
                  ((1.23149 1.50755)))
                 ((1.32006 1.12872)))
                ((R:segstate.parent.ph_vlng is l)
                 ((1.29414 0.844857))
                 ((1.71337 1.36712))))
               ((n.name is n_1)
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                  is
                  0)
                 ((2.25979 1.07975))
                 ((1.82078 2.42493)))
                ((n.name is m_1)
                 ((R:segstate.parent.ph_vlng is s)
                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                    <
                    1.8)
                   ((0.900483 2.72421))
                   ((0.996647 1.38631)))
                  ((1.54224 0.0505754)))
                 ((R:segstate.parent.n.name is m)
                  ((R:segstate.parent.p.ph_ctype is f)
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((0.149791 -0.76675))
                    ((0.796324 -0.522255)))
                   ((1.74501 0.599314)))
                  ((R:segstate.parent.n.name is l)
                   ((R:segstate.parent.R:SylStructure.parent.asyl_in
                     <
                     1.2)
                    ((1.15439 0.557997))
                    ((R:segstate.parent.ph_vheight is 2)
                     ((1.29669 0.4058))
                     ((0.764867 -0.143037))))
                   ((n.name is ao_3)
                    ((0.338112 -0.340505))
                    ((n.name is r_2)
                     ((1.10904 0.0889805))
                     ((R:segstate.parent.n.name is s)
                      ((n.name is s_1)
                       ((1.14002 1.08851))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                         is
                         0)
                        ((0.534311 -0.0534563))
                        ((0.868423 0.262961))))
                      ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                        <
                        8.4)
                       ((R:segstate.parent.p.name is eh)
                        ((R:segstate.parent.pp.ph_cplace is d)
                         ((1.10698 1.10748))
                         ((1.92258 2.09038)))
                        ((n.name is n_2)
                         ((0.765607 0.111737))
                         ((n.name is r_1)
                          ((1.87314 0.473699))
                          ((R:segstate.parent.ph_vc is +)
                           ((R:segstate.parent.pp.ph_vlng is l)
                            ((1.53903 0.732388))
                            ((R:segstate.parent.R:SylStructure.parent.last_accent
                              <
                              1.4)
                             ((R:segstate.parent.n.name is r)
                              ((1.53891 2.28347))
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                                is
                                0)
                               ((2.39843 1.83889))
                               ((1.55864 1.00757))))
                             ((R:segstate.parent.R:SylStructure.parent.last_accent
                               <
                               3.5)
                              ((R:segstate.parent.R:SylStructure.parent.syl_in
                                <
                                9)
                               ((1.41022 1.02188))
                               ((0.879546 0.767913)))
                              ((R:segstate.parent.p.ph_cvox is +)
                               ((2.07263 1.55554))
                               ((1.51522 1.33055))))))
                           ((n.name is l_2)
                            ((1.35335 1.66007))
                            ((n.name is m_3)
                             ((R:segstate.parent.p.ph_vlng is s)
                              ((0.57891 1.65121))
                              ((1.10257 0.513978)))
                             ((R:segstate.parent.lisp_onset_stop is 0)
                              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.accented
                                is
                                0)
                               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                                 is
                                 0)
                                ((1.34138 1.3866))
                                ((R:segstate.parent.R:SylStructure.parent.last_accent
                                  <
                                  1.4)
                                 ((1.19962 1.15883))
                                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                                   is
                                   0)
                                  ((0.922291 0.452142))
                                  ((0.824254 0.938037)))))
                               ((1.07509 0.199336)))
                              ((0.901915 0.175259)))))))))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_in
                         <
                         3.2)
                        ((1.00847 0.235888))
                        ((1.14666 0.677862)))))))))))))
             ((R:segstate.parent.name is uw)
              ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 3.1)
               ((0.959263 0.0801053))
               ((0.773057 -0.205022)))
              ((R:segstate.parent.nn.ph_vc is -)
               ((R:segstate.parent.lisp_coda_nasal is 0)
                ((R:segstate.parent.R:SylStructure.parent.last_accent
                  <
                  3.5)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                   is
                   0)
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                    <
                    0.3)
                   ((0.863096 -0.0482741))
                   ((1.55634 0.559546)))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                    <
                    0.1)
                   ((0.852579 0.120569))
                   ((0.557791 -0.230092))))
                 ((1.5075 0.793282)))
                ((R:segstate.parent.ph_vheight is 2)
                 ((0.950007 0.212724))
                 ((1.99567 1.29523))))
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.lisp_cg_break
                 <
                 0.3)
                ((1.44907 0.487277))
                ((R:segstate.parent.pos_in_syl is 2)
                 ((2.38533 1.49773))
                 ((1.16151 0.630536)))))))))))
        ((n.name is ih_3)
         ((R:segstate.parent.R:SylStructure.parent.syl_in < 7.8)
          ((1.86702 2.7356))
          ((1.42237 0.875099)))
         ((R:segstate.parent.R:SylStructure.parent.syl_codasize < 2.2)
          ((R:segstate.parent.name is t)
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
             <
             1.2)
            ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 4)
             ((1.03657 -0.221526))
             ((1.79603 1.2262)))
            ((n.name is t_2)
             ((1.5062 -0.215004))
             ((0.0481488 -0.73292))))
           ((n.name is d_2)
            ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.5)
             ((0.346593 -0.341155))
             ((0.89943 0.136493)))
            ((R:segstate.parent.p.name is eh)
             ((1.68978 2.28037))
             ((R:segstate.parent.n.name is k)
              ((R:segstate.parent.R:SylStructure.parent.syl_in < 9)
               ((1.03931 0.429211))
               ((R:segstate.parent.pp.ph_cplace is 0)
                ((1.08698 0.128143))
                ((0.594072 -0.131799))))
              ((n.name is k_3)
               ((0.95013 1.57946))
               ((R:segstate.parent.ph_cplace is v)
                ((1.19243 0.0628985))
                ((R:segstate.parent.p.ph_ctype is r)
                 ((R:segstate.parent.n.name is d)
                  ((1.22903 0.573893))
                  ((0.900924 0.21027)))
                 ((n.name is ih_2)
                  ((0.794766 0.0611298))
                  ((R:segstate.parent.pp.ph_cplace is l)
                   ((1.09064 0.209689))
                   ((R:segstate.parent.n.name is t)
                    ((n.name is t_1)
                     ((R:segstate.parent.p.ph_cvox is -)
                      ((0.860219 0.53559))
                      ((1.58753 1.30479)))
                     ((R:segstate.parent.ph_vlng is d)
                      ((0.681228 0.00555222))
                      ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
                        is
                        0)
                       ((1.4284 1.06332))
                       ((2.01481 0.752581)))))
                    ((R:segstate.parent.p.ph_vheight is 3)
                     ((0.854579 0.592438))
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                       <
                       10)
                      ((R:segstate.parent.p.name is d)
                       ((1.55719 0.664246))
                       ((R:segstate.parent.p.ph_ctype is s)
                        ((1.80505 1.86382))
                        ((R:segstate.parent.pp.ph_cplace is a)
                         ((R:segstate.parent.ph_vrnd is -)
                          ((1.79789 1.4153))
                          ((1.97789 1.81706)))
                         ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                           is
                           0)
                          ((R:segstate.parent.ph_vlng is s)
                           ((2.27143 1.4047))
                           ((1.1928 0.879192)))
                          ((1.22551 0.648658))))))
                      ((1.10222 0.4997))))))))))))))
          ((R:segstate.parent.p.ph_ctype is n)
           ((R:segstate.parent.name is d)
            ((R:segstate.parent.R:SylStructure.parent.syl_in < 10.5)
             ((0.406855 -0.359634))
             ((0.238333 -0.542418)))
            ((R:segstate.parent.pp.ph_vlng is s)
             ((0.672847 -0.436134))
             ((2.26172 0.818575))))
           ((R:segstate.parent.n.name is l)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              1.2)
             ((0.588492 -0.294551))
             ((0.89603 -0.367406)))
            ((R:segstate.parent.R:SylStructure.parent.syl_codasize
              <
              3.1)
             ((R:segstate.parent.n.ph_cplace is a)
              ((R:segstate.parent.ph_vlng is s)
               ((R:segstate.parent.name is eh)
                ((1.08133 0.508904))
                ((1.93838 1.39664)))
               ((n.name is s_1)
                ((0.664491 -0.441649))
                ((R:segstate.parent.lisp_coda_glide is 0)
                 ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                   is
                   0)
                  ((R:segstate.parent.p.ph_cvox is +)
                   ((0.849377 0.301946))
                   ((n.name is d_1)
                    ((1.49969 1.3134))
                    ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                      is
                      0)
                     ((R:segstate.parent.R:SylStructure.parent.asyl_in
                       <
                       2.1)
                      ((1.1008 1.207))
                      ((1.64168 0.74961)))
                     ((R:segstate.parent.pp.ph_cplace is l)
                      ((1.39494 0.899575))
                      ((R:segstate.parent.R:SylStructure.parent.syl_in
                        <
                        10)
                       ((0.785406 0.477594))
                       ((0.664366 0.0257137)))))))
                  ((0.80107 0.073085)))
                 ((R:segstate.parent.n.name is t)
                  ((0.967343 0.647207))
                  ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                    is
                    0)
                   ((1.04346 -0.113954))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                     <
                     1.2)
                    ((0.959782 0.316171))
                    ((1.43009 0.346368))))))))
              ((n.name is t_2)
               ((1.37568 0.813265))
               ((R:segstate.parent.pp.ph_vlng is s)
                ((1.44142 0.612941))
                ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                  is
                  0)
                 ((R:segstate.parent.pp.ph_vrnd is -)
                  ((0.453532 -0.499724))
                  ((0.82189 -0.0104987)))
                 ((R:segstate.parent.ph_cvox is +)
                  ((R:segstate.parent.nn.ph_vc is -)
                   ((1.24449 0.0708487))
                   ((0.536355 -0.300801)))
                  ((R:segstate.parent.n.ph_ctype is s)
                   ((0.692681 0.0517717))
                   ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
                     is
                     0)
                    ((1.30739 0.231843))
                    ((R:segstate.parent.p.ph_cvox is -)
                     ((0.97975 0.59555))
                     ((1.39254 0.418134))))))))))
             ((R:segstate.parent.pos_in_syl is 2)
              ((0.816758 -0.273188))
              ((R:segstate.parent.p.ph_cvox is -)
               ((R:segstate.parent.R:SylStructure.parent.last_accent
                 <
                 2)
                ((1.30144 0.0969014))
                ((0.786977 -0.293263)))
               ((R:segstate.parent.p.seg_onsetcoda is coda)
                ((1.69627 0.735974))
                ((0.90352 0.0940155)))))))))))))))
   ((n.name is r_1)
    ((R:segstate.parent.lisp_coda_stop is 0)
     ((1.29982 2.09081))
     ((R:segstate.parent.R:SylStructure.parent.last_accent < 1.1)
      ((1.64636 1.30354))
      ((0.777255 0.308291))))
    ((R:segstate.parent.p.ph_vlng is 0)
     ((R:segstate.parent.n.name is uw)
      ((R:segstate.parent.R:SylStructure.parent.syl_codasize < 1.2)
       ((R:segstate.parent.lisp_onset_stop is 0)
        ((1.11044 0.580457))
        ((1.05842 1.21742)))
       ((R:segstate.parent.R:SylStructure.parent.position_type
         is
         final)
        ((1.05049 0.426013))
        ((1.11624 -0.0649977))))
      ((n.name is ih_1)
       ((R:segstate.parent.pp.ph_vlng is s)
        ((1.01444 0.0627135))
        ((R:segstate.parent.R:SylStructure.parent.syl_in < 6.6)
         ((0.797518 0.0901936))
         ((1.24002 1.12525))))
       ((R:segstate.parent.ph_cvox is +)
        ((R:segstate.parent.name is w)
         ((R:segstate.parent.nn.ph_cplace is a)
          ((1.31455 0.664832))
          ((1.0971 0.141202)))
         ((n.name is b_2)
          ((1.38509 0.666134))
          ((R:segstate.parent.name is n)
           ((R:segstate.parent.p.ph_ctype is s)
            ((2.06156 0.79682))
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              1.5)
             ((1.90577 0.284631))
             ((0.490436 -0.36363))))
           ((n.name is iy_1)
            ((R:segstate.parent.p.ph_ctype is s)
             ((0.380823 0.408685))
             ((0.589832 0.124371)))
            ((n.name is ey_1)
             ((1.10028 0.286916))
             ((R:segstate.parent.n.name is iy)
              ((n.name is l_2)
               ((0.136396 -0.739447))
               ((R:segstate.parent.R:SylStructure.parent.syl_in < 5)
                ((0.290821 -0.596759))
                ((R:segstate.parent.p.ph_ctype is s)
                 ((0.352808 -0.48841))
                 ((0.46203 -0.273718)))))
              ((R:segstate.parent.p.ph_ctype is n)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.lisp_cg_break
                 <
                 0.1)
                ((R:segstate.parent.n.ph_ctype is 0)
                 ((R:segstate.parent.name is dh)
                  ((0.0291794 -0.644375))
                  ((R:segstate.parent.n.name is ax)
                   ((0.265051 -0.521459))
                   ((0.443845 -0.330526))))
                 ((0.204585 -0.704432)))
                ((0.72331 -0.102195)))
               ((R:segstate.parent.pp.ph_vheight is 3)
                ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 5)
                 ((1.02007 0.337461))
                 ((0.956126 -0.0808106)))
                ((R:segstate.parent.name is g)
                 ((0.5596 -0.662837))
                 ((R:segstate.parent.n.name is ey)
                  ((0.349767 -0.543185))
                  ((R:segstate.parent.lisp_coda_nasal is 0)
                   ((n.name is l_3)
                    ((0.899413 0.0803691))
                    ((R:segstate.parent.R:SylStructure.parent.last_accent
                      <
                      4.5)
                     ((R:segstate.parent.name is r)
                      ((R:segstate.parent.R:SylStructure.parent.position_type
                        is
                        final)
                       ((0.58677 -0.320166))
                       ((R:segstate.parent.R:SylStructure.parent.asyl_in
                         <
                         1.4)
                        ((0.514454 -0.191285))
                        ((0.783658 0.107564))))
                      ((R:segstate.parent.p.name is l)
                       ((0.971008 -0.0254273))
                       ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.syl_codasize
                         <
                         1.2)
                        ((0.790954 -0.142222))
                        ((R:segstate.parent.ph_cplace is l)
                         ((0.657091 -0.249742))
                         ((R:segstate.parent.R:SylStructure.parent.last_accent
                           <
                           2.4)
                          ((R:segstate.parent.nn.ph_ctype is s)
                           ((0.346046 -0.599847))
                           ((R:segstate.parent.lisp_onset_glide is 0)
                            ((0.462866 -0.503181))
                            ((0.794486 -0.282236))))
                          ((0.221407 -0.643329)))))))
                     ((0.722789 0.0500398))))
                   ((R:segstate.parent.name is dh)
                    ((0.700921 -0.182532))
                    ((R:segstate.parent.p.ph_cvox is +)
                     ((R:segstate.parent.pp.ph_vrnd is -)
                      ((0.570982 -0.359795))
                      ((0.562918 -0.15846)))
                     ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                       <
                       6.4)
                      ((R:segstate.parent.R:SylStructure.parent.syl_onsetsize
                        <
                        1.2)
                       ((0.39424 -0.453538))
                       ((0.436764 -0.647156)))
                      ((0.44318 -0.382574))))))))))))))))
        ((n.name is s_2)
         ((R:segstate.parent.p.ph_cvox is +)
          ((0.549831 -0.453745))
          ((0.348446 -0.864289)))
         ((R:segstate.parent.pos_in_syl is 0)
          ((n.name is er_1)
           ((1.36164 0.837578))
           ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 9.6)
            ((R:segstate.parent.R:SylStructure.parent.position_type
              is
              final)
             ((R:segstate.parent.n.ph_ctype is r)
              ((0.778689 -0.585153))
              ((R:segstate.parent.lisp_coda_stop is 0)
               ((R:segstate.parent.pp.ph_vlng is s)
                ((R:segstate.parent.n.name is eh)
                 ((0.404828 0.0826361))
                 ((R:segstate.parent.pp.ph_vheight is 3)
                  ((0.662982 -0.265547))
                  ((R:segstate.parent.ph_cplace is a)
                   ((0.914264 0.013342))
                   ((0.64167 -0.169177)))))
                ((0.957147 0.271864)))
               ((0.54689 -0.288813))))
             ((R:segstate.parent.ph_cplace is a)
              ((R:segstate.parent.lisp_coda_stop is 0)
               ((R:segstate.parent.R:SylStructure.parent.asyl_in < 1.4)
                ((0.928174 0.95148))
                ((0.990054 0.496607)))
               ((0.981958 0.248972)))
              ((R:segstate.parent.nn.ph_ctype is n)
               ((n.name is hh_2)
                ((0.298224 -0.440138))
                ((R:segstate.parent.n.name is ih)
                 ((0.604485 -0.238781))
                 ((0.945605 0.126345))))
               ((n.name is f_2)
                ((0.976147 -0.194318))
                ((R:segstate.parent.R:SylStructure.parent.last_accent
                  <
                  2)
                 ((R:segstate.parent.pp.ph_vheight is 1)
                  ((1.32189 0.801184))
                  ((R:segstate.parent.R:SylStructure.parent.ssyl_in
                    <
                    1.8)
                   ((1.36363 0.504545))
                   ((0.852275 0.16694))))
                 ((R:segstate.parent.p.ph_cvox is +)
                  ((R:segstate.parent.R:SylStructure.parent.last_accent
                    <
                    2.8)
                   ((1.089 0.481704))
                   ((0.943113 0.127722)))
                  ((R:segstate.parent.R:SylStructure.parent.asyl_in
                    <
                    1.5)
                   ((1.03678 0.0080432))
                   ((0.848449 -0.191337)))))))))
            ((R:segstate.parent.pp.ph_vheight is 2)
             ((3.08709 1.16797))
             ((0.885575 0.184583)))))
          ((n.name is t_3)
           ((0.538716 0.319552))
           ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
             is
             0)
            ((R:segstate.parent.pp.ph_vlng is a)
             ((0.64855 -0.396134))
             ((0.606648 -0.589534)))
            ((0.567305 -0.253739)))))))))
     ((R:segstate.parent.R:SylStructure.parent.position_type is final)
      ((n.name is n_2)
       ((3.1312 1.42478))
       ((R:segstate.parent.name is n)
        ((R:segstate.parent.R:SylStructure.parent.R:Syllable.p.stress
          is
          0)
         ((0.550057 -0.275656))
         ((n.name is n_3) ((0.67353 -0.709482)) ((0.192461 -0.505551))))
        ((R:segstate.parent.n.ph_vrnd is -)
         ((R:segstate.parent.pp.ph_cplace is a)
          ((R:segstate.parent.name is t)
           ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 4.7)
            ((0.678194 -0.0294079))
            ((n.name is t_3)
             ((0.117472 -0.662301))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.accented
               is
               0)
              ((0.241692 -0.509474))
              ((0.231279 -0.541979)))))
           ((R:segstate.parent.p.name is ey)
            ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 5.5)
             ((0.529111 -0.422849))
             ((0.407932 -0.181004)))
            ((n.name is ax_1)
             ((0.750793 -0.251251))
             ((R:segstate.parent.ph_cvox is +)
              ((R:segstate.parent.R:SylStructure.parent.asyl_in < 5.6)
               ((R:segstate.parent.R:SylStructure.parent.R:Syllable.n.stress
                 is
                 0)
                ((R:segstate.parent.R:SylStructure.parent.syl_in
                  <
                  14.5)
                 ((R:segstate.parent.lisp_coda_stop is 0)
                  ((R:segstate.parent.R:SylStructure.parent.syl_in
                    <
                    9.2)
                   ((R:segstate.parent.p.ph_vheight is 2)
                    ((0.756744 -0.00796245))
                    ((0.69471 0.156018)))
                   ((R:segstate.parent.ph_cplace is a)
                    ((0.844634 0.0638546))
                    ((0.654807 0.438572))))
                  ((0.984124 -0.223875)))
                 ((0.800351 -0.209114)))
                ((0.730409 -0.230134)))
               ((0.388838 -0.386587)))
              ((R:segstate.parent.p.ph_vheight is 2)
               ((R:segstate.parent.nn.ph_cplace is a)
                ((1.58946 0.510406))
                ((0.773816 0.128061)))
               ((0.41321 -0.0567986)))))))
          ((n.name is ax_1)
           ((R:segstate.parent.lisp_onset_stop is 0)
            ((R:segstate.parent.p.ph_vlng is s)
             ((0.720515 0.204394))
             ((0.52133 -0.207168)))
            ((0.541265 -0.5394)))
           ((R:segstate.parent.p.ph_vlng is d)
            ((R:segstate.parent.R:SylStructure.parent.last_accent
              <
              2.4)
             ((R:segstate.parent.lisp_coda_stop is 0)
              ((R:segstate.parent.R:SylStructure.parent.R:Syllable.pp.stress
                is
                0)
               ((R:segstate.parent.R:SylStructure.parent.asyl_in < 4.5)
                ((0.351124 -0.406135))
                ((0.595096 -0.090659)))
               ((1.01032 0.156646)))
              ((0.56337 -0.507846)))
             ((0.921765 0.519585)))
            ((R:segstate.parent.nn.ph_cplace is l)
             ((0.89353 0.66578))
             ((R:segstate.parent.n.name is aa)
              ((0.595454 0.715958))
              ((R:segstate.parent.name is t)
               ((0.249544 -0.526037))
               ((n.name is l_2)
                ((0.757276 -0.0756109))
                ((R:segstate.parent.R:SylStructure.parent.asyl_in
                  <
                  2.4)
                 ((R:segstate.parent.p.ph_vheight is 3)
                  ((0.806298 -0.128816))
                  ((R:segstate.parent.nn.ph_ctype is n)
                   ((0.59174 -0.0823108))
                   ((R:segstate.parent.p.name is ax)
                    ((0.615548 0.147417))
                    ((R:segstate.parent.pp.ph_vfront is 2)
                     ((0.549475 0.240798))
                     ((1.0155 0.413238))))))
                 ((R:segstate.parent.n.name is iy)
                  ((0.813473 0.199705))
                  ((R:segstate.parent.nn.ph_ctype is n)
                   ((R:segstate.parent.R:SylStructure.parent.parent.word_numsyls
                     <
                     2.2)
                    ((1.08378 0.41065))
                    ((0.804343 0.0698449)))
                   ((R:segstate.parent.pp.ph_cplace is 0)
                    ((0.922763 0.404392))
                    ((R:segstate.parent.p.name is er)
                     ((1.16279 0.4515))
                     ((1.31406 0.770032))))))))))))))
         ((n.name is s_3)
          ((0.69543 -0.170099))
          ((R:segstate.parent.R:SylStructure.parent.syl_codasize < 1.3)
           ((R:segstate.parent.ph_cplace is a)
            ((R:segstate.parent.p.ph_vlng is l)
             ((1.02914 0.238774))
             ((0.606526 0.528295)))
            ((0.762583 0.161753)))
           ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 7.8)
            ((R:segstate.parent.R:SylStructure.parent.sub_phrases
              <
              0.2)
             ((R:segstate.parent.R:SylStructure.parent.ssyl_in < 4.2)
              ((1.09507 0.0985561))
              ((R:segstate.parent.nn.ph_vrnd is -)
               ((1.06863 0.556071))
               ((1.28613 0.346873))))
             ((0.83436 0.592372)))
            ((0.91694 0.77324))))))))
      ((n.name is w_3)
       ((0.786483 1.33773))
       ((R:segstate.parent.name is t)
        ((0.836639 1.2735))
        ((n.name is w_2)
         ((0.457622 -0.27097))
         ((R:segstate.parent.lisp_onset_nasal is 0)
          ((R:segstate.parent.name is k)
           ((1.12261 0.778171))
           ((R:segstate.parent.n.name is ey)
            ((R:segstate.parent.nn.ph_cplace is a)
             ((0.415658 -0.189016))
             ((0.899182 0.388066)))
            ((R:segstate.parent.name is s)
             ((R:segstate.parent.p.name is ax)
              ((0.571847 0.138177))
              ((0.645073 0.273041)))
             ((R:segstate.parent.R:SylStructure.parent.R:Syllable.nn.stress
               is
               0)
              ((R:segstate.parent.ph_cvox is +)
               ((R:segstate.parent.R:SylStructure.parent.asyl_in < 3.5)
                ((R:segstate.parent.nn.ph_ctype is s)
                 ((R:segstate.parent.p.ph_vlng is a)
                  ((1.57446 0.81753))
                  ((1.13535 0.408857)))
                 ((R:segstate.parent.lisp_onset_glide is 0)
                  ((1.25967 0.562161))
                  ((R:segstate.parent.ph_cplace is a)
                   ((1.18317 0.395261))
                   ((0.616778 -0.0827707)))))
                ((R:segstate.parent.lisp_onset_stop is 0)
                 ((0.846121 -0.108596))
                 ((0.639286 0.198636))))
               ((R:segstate.parent.pp.ph_cplace is a)
                ((1.05552 0.46891))
                ((R:segstate.parent.p.ph_vlng is a)
                 ((0.901505 0.500087))
                 ((1.54702 1.0229)))))
              ((1.01494 0.117954))))))
          ((R:segstate.parent.n.ph_vrnd is -)
           ((0.682777 -0.146949))
           ((0.549176 0.209408)))))))))))))
)
)


(provide 'cmu_us_slt_dur)


